//=======================================================================
// gregorien.cc : conversion entre date grgorienne et numro cumulatif
// Copyright (C) 2003-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// ( compiler avec g++ gregorien.cc -o gregorien)
//=======================================================================

#include <iostream>  // dclarer l'entre-sortie standard
#include <iomanip>   // manipulateurs pour formater les flots
#include <string>    // manipuler les chanes de caractres
#include <sstream>   // conversion entre string et flot
using namespace std; // accs direct aux fonctions standard

//-----------------------------------------------------------------------
// La fonction  bissextile(annee)  renvoie 1 si l'anne est bissextile 
// et 0 sinon.  Nous appliquons ici la rgle du calendrier grgorien.
//-----------------------------------------------------------------------
int bissextile( int annee )
{
  if ( annee % 400 == 0 ) return 1;
  if ( annee % 100 == 0 ) return 0;
  if ( annee %   4 == 0 ) return 1;
  return 0;
}

//-----------------------------------------------------------------------
// Pour une date grgorienne donne, la fonction  numero(jour,mois,annee) 
// calcule son numro cumulatif  compter du 01/01/0001.  Nous appliquons
// ici les rgles du calendrier grgorien, bien qu'elles ne prissent 
// effet que le 15 octobre 1582.  Ainsi le 01/01/0001 est le jour numro 1 
// et le 01/01/2001, par exemple, est le jour numro 730486.  La fonction
// suivante ralise cette numrotation cumulative, et en mme temps
// elle teste la validit : pour une date invalide elle renvoie 0.
//-----------------------------------------------------------------------
int numero( int jour, int mois, int annee )
{
  // Attraper quelques erreurs videntes causes par des dates invalides.
  // La vrification pour le jour et le mois sera complte plus bas.
  if ( jour < 1 || mois < 1 || annee < 1 ) return 0;
  if ( jour > 31 || mois > 12 || annee > 9999 ) return 0;
  
  // Dterminer si l'anne courante est bissextile ou non. 
  // Cette rgle raffine est particulire au calendrier grgorien.
  int b= bissextile( annee );
    
  // Calculer le nombre des jours avant le 1er janvier de l'anne courante.
  // Pour ceci il faut  nouveau appliquer la rgle des annes bissextiles.
  int c= 365 * (annee-1) // jours avant le dbut de l'anne courante
    + (annee-1) / 4      // ajouter un jour par anne bissextile
    - (annee-1) / 100    // soustraire un jour par bissextile saut
    + (annee-1) / 400;   // ajouter un jour par bissextile non saut
  
  // Reste  rajouter le nombre des jours couls pendant l'anne courante.
  // Celui-ci dpend du jour et du mois, et peut se raliser comme suit:
  switch( mois )
    {
    case  1: if ( jour <= 31 )   return jour           + c; else return 0;
    case  2: if ( jour <= 28+b ) return jour +  31     + c; else return 0;
    case  3: if ( jour <= 31 )   return jour +  59 + b + c; else return 0;
    case  4: if ( jour <= 30 )   return jour +  90 + b + c; else return 0;
    case  5: if ( jour <= 31 )   return jour + 120 + b + c; else return 0;
    case  6: if ( jour <= 30 )   return jour + 151 + b + c; else return 0;
    case  7: if ( jour <= 31 )   return jour + 181 + b + c; else return 0;
    case  8: if ( jour <= 31 )   return jour + 212 + b + c; else return 0;
    case  9: if ( jour <= 30 )   return jour + 243 + b + c; else return 0;
    case 10: if ( jour <= 31 )   return jour + 273 + b + c; else return 0;
    case 11: if ( jour <= 30 )   return jour + 304 + b + c; else return 0;
    case 12: if ( jour <= 31 )   return jour + 334 + b + c; else return 0;
    default: return 0;
    };
}
     
//-----------------------------------------------------------------------
// La fonction  date(num,jour,mois,annee)  dtermine la date j/m/a
//  partir de son numro cumulatif num  compter du 01/01/0001. Nous
// appliquons  nouveau les rgles du calendrier grgorien, bien qu'elles 
// ne prissent effet que le vendredi 15 octobre 1582. Ainsi le numro 1 
// correspond au 01/01/0001 et le numro 730486 correspond au 01/01/2001.
// Pour information, le numro 3652059 correspond au vendredi 31/12/9999.
//-----------------------------------------------------------------------
void date( int num, int& jour, int& mois, int& annee )
{
  // Refuser un numro cumulatif qui est trop petit ou trop grand
  if ( num < 1 || num > 3652059 ) { jour=0; mois=0; annee=0; return; }

  // Calculer l'anne et le numro du jour dans l'anne courante
  annee= num / 365 + 1;     // calculer l'anne ( corriger plus bas)
  jour = num % 365          // jours restants pour l'anne courante.
    - (annee-1) / 4         // soustraire un jour par anne bissextile
    + (annee-1) / 100       // ajouter un jour par bissextile saute
    - (annee-1) / 400;      // soustraire un jour par bissextile non saute
  while ( jour < 1 )        // reculer un an si ncessaire
    jour+= 365 + bissextile( --annee );
  
  // Reste  trouver le mois et le jour dans l'anne courante.
  int b= bissextile(annee);
  if ( jour <= 31 )   { mois=  1; return; } else jour-= 31; 
  if ( jour <= 28+b ) { mois=  2; return; } else jour-= 28+b;
  if ( jour <= 31 )   { mois=  3; return; } else jour-= 31; 
  if ( jour <= 30 )   { mois=  4; return; } else jour-= 30; 
  if ( jour <= 31 )   { mois=  5; return; } else jour-= 31; 
  if ( jour <= 30 )   { mois=  6; return; } else jour-= 30; 
  if ( jour <= 31 )   { mois=  7; return; } else jour-= 31; 
  if ( jour <= 31 )   { mois=  8; return; } else jour-= 31; 
  if ( jour <= 30 )   { mois=  9; return; } else jour-= 30; 
  if ( jour <= 31 )   { mois= 10; return; } else jour-= 31; 
  if ( jour <= 30 )   { mois= 11; return; } else jour-= 30;
  mois=12; 
  return; 
}

//-----------------------------------------------------------------------
// La fonction  affiche(num,jour,mois,annee)  affiche la date donne.
//-----------------------------------------------------------------------
void affiche( int num, int jour, int mois, int annee )
{
  switch ( num % 7 )
    {
    case 0: cout << "dimanche "; break;
    case 1: cout << "lundi ";    break;
    case 2: cout << "mardi ";    break;
    case 3: cout << "mercredi "; break;
    case 4: cout << "jeudi ";    break;
    case 5: cout << "vendredi "; break;
    case 6: cout << "samedi ";   break;
    default: ;
    };
  cout << setw(2) << setfill('0') << jour << '/'
       << setw(2) << setfill('0') << mois << '/'
       << setw(4) << setfill('0') << annee; 
}

//-----------------------------------------------------------------------
// La fonction  date2num( date )  lit une date d'une chane de caractre
// puis calcule et affiche le numro cumulatif  compter du 01/01/0001.
//-----------------------------------------------------------------------
void date2num( const char* date )
{
  istringstream in( date );
  int jour, mois, annee; 
  char dummy;
  in >> jour >> dummy >> mois >> dummy >> annee;
  
  int num= numero( jour, mois, annee );
  if ( num <= 0 )
    {
      cout << "Dsol, ceci n'est pas une date valable." << endl;
      return;
    };
  
  cout << " compter du 01/01/0001, le ";
  affiche( num, jour, mois, annee );
  cout << " est le jour numro " << num << endl;
}

//-----------------------------------------------------------------------
// La fonction  num2date( numero )  lit un nombre d'une chane de caractre
// et affiche la date j/m/a correspondante avec le jour de la semaine.
//-----------------------------------------------------------------------
void num2date( const char* numero )
{
  istringstream in( numero );
  int num, jour, mois, annee;
  in >> num;
  if ( num < 1)
    {
      cout << "Dsol, ceci n'est pas un numro valable." << endl;
      return;
    };
  
  date( num, jour, mois, annee );
  cout << " compter du 01/01/0001, le jour numro " << num << " est le ";
  affiche( num, jour, mois, annee );
  cout << "." << endl;
}

//-----------------------------------------------------------------------
// La fonction  main()  met  disposition deux conversions de date :
// -  partir d'une date j/m/a elle calcule le numro cumulatif,
// -  partir du numro cumulatif elle retrouve la date j/m/a.
// Pour explication : le logiciel peut tre appel avec des arguments.
// Ainsi  argc  contient le nombres des arguments passs, et  argv[]
// est la liste des arguments, dont chacun est une chane de caractres :
// argv[0] est le 0me argument, qui est toujours le nom du logiciel,
// argv[1] est le 1er argument, ici la date ou le nombre  convertir,
// argv[2],argv[3] etc peuvent contenir d'autres arguments (non utiliss).
//-----------------------------------------------------------------------
int main( int argc, char* argv[] )
{
  // On teste d'abord si le nombre d'arguments vaut 2, comme il faut.
  // Sinon on affiche un message d'aide avec un petit rsum du logiciel.
  if ( argc != 2 )
    {
      cout << "\ngregorien -- conversion entre date grgorienne et numro cumulatif\n"
	   << "(C) 1582--2006 pape Grgoire XIII & Michael Eisermann\n"
	   << "usage : " << argv[0] << " <date>\n"
	   << "        " << argv[0] << " <numro>\n"
	   << "Le premier usage prend une date de la forme jj/mm/aaaa\n"
	   << "et calcule son numro cumulatif  compter du 01/01/0001.\n"
	   << "La deuxime forme effectue la conversion rciproque.\n"
	   << " noter que ce logiciel applique les rgles du calendrier\n"
	   << "grgorien, bien qu'elles ne prissent effet que le vendredi\n"
	   << "15 octobre 1582. Avant ce jour-l, les dates grgoriennes\n"
	   << "diffrent de celles historiquement utilises.\n" << endl;
      return 0;
    };
  
  // Comment dcider quelle conversion devrait tre applique ? 
  // C'est simple : si l'argument contient un des caractres "./-", 
  // alors il s'agit d'une date. Sinon il s'agit d'un simple numro.
  if ( string(argv[1]).find_first_of("./-") == string::npos )
    num2date( argv[1] );
  else
    date2num( argv[1] );
}
