//=======================================================================
// naturel.cc : arithmtique lmentaire des nombres naturels
// Copyright (C) 2004-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Le C++ ne prvoit pas de type primitif pour modliser les grands 
// nombres entiers.  Le programme ci-dessous implmente un tel type 
// en utilisant la classe gnrique vector de la bibliothque STL. 
//
// *** Avertissement: aucun effort d'optimisation n'est fait ici. ***
// L'implmentation propose a pour seul but d'illustrer le principe.
// Dans toute application raliste on privilgiera une implmentation 
// plus complte, comme par exemple celle de la bibliothque GMP.
// ******************************************************************
//
// Inconvnients connus :
// -- On ne traite pas les entiers ngatifs.
// -- Multiplication et division cotent cher pour les grands nombres.
//
// Structure de donnes :
// Dans une base donne, un chiffre est un entier entre 0 et base-1.
// Une reprsentation dans cette base est une suite finie de chiffres
// a[0],a[1],...,a[k], de longueur k+1.  L'entier a ainsi reprsent 
// est a = a[0] + a[1]*base + ... + a[k]*base^k.  Par convention on 
// supprimera systmatiquement d'ventuels zros  la fin, de sorte
// que a[k]>0.  Ainsi la suite vide reprsente le nombre naturel 0.
// Toutes les fonctions suivantes supposent les chiffres dans ce format
// et assurent que le rsultat sera dans le mme format "normalis".
//
// Pourquoi cette convention ?
// On pourra stocker les chiffres dans le sens inverse, mais le stockage
// en poids croissant semble plus commode.  En tout cas il est important
// de fixer une convention et de la respecter dans toute la suite.
//
// Algorithmes lgants mais inefficaces :
// Pour les nombres naturels, l'opration de base est de *compter*
// c'est--dire incrmenter a  a+1.  Toutes les oprations lmentaires 
// (addition, soustraction, multiplication, division euclidienne, ...)
// sont dfinies  partir de l'incrmentation, et ces dfinitions
// se traduisent immdiatement en algorithmes de calcul.   titre 
// d'illustration ces dfinitions primaires sont implmentes 
// littralement.  On obtient ainsi des algorithmes lgants mais 
// terriblement inefficaces.  Ils ont tout leur intrt en mathmatiques
// mais en programmation il faut passer  des algorithmes plus rapides.
//
// Algorithmes scolaires :
// Les algorithmes connus de l'cole en sont dj une amlioration
// importante : en travaillant directement sur les chiffres de la 
// reprsentation, on arrive  un temps de calcul linaire pour
// l'addition/soustraction et quadratique pour la multiplication/
// division euclidienne.  (Ici on mesure la complexit des donnes
// initiales par la longueur dans la prsentation choisie.) 
// Aucune optimisation supplmentaire n'est applique ici.
//
//  noter que la multiplication et la division peuvent tre effectues
// par des algorithmes beaucoup plus efficaces (mais aussi plus pousss).
//=======================================================================

#include <iostream>            // entre-sortie standard
#include <sstream>             // conversion entre string et flot
#include <vector>              // la classe gnrique vector
using namespace std;           // accs direct aux fonctions standard

//=======================================================================
// Types, constantes et fonctions auxiliaires
//=======================================================================
typedef int Indice;            // type pour stocker un indice
typedef short int Chiffre;     // type pour stocker un chiffre (ou deux)
const Chiffre base= 10;        // base entre 2 et 16, ici on choisit 10
const char chiffre[]= "0123456789abcdef";  // chiffres pour l'affichage

//=======================================================================
// Dfinition de la classe Naturel
//=======================================================================
class Naturel
{
public:
  //---------------------------------------------------------------------
  // La suite des chiffres peut tre dclare `public' afin de permettre
  // un accs direct par l'extrieur (voir les fonctions ci-dessous).
  // Dans une implmentation qui se veut professionnelle ces donnes 
  // devraient tre dclares `private' afin d'interdire tout accs 
  // par l'extrieur : ainsi seules le fonctions membres de la classe 
  // et ses fonctions amies y ont un accs direct.  Tout usage passera 
  // par ces fonctions lmentaires, ce qui assure un meilleur contrle 
  // et minimise le risque d'un usage erron.
  //---------------------------------------------------------------------
  vector<Chiffre> chiffres;  // La suite des chiffres dans la base donne 
  
public:
  //---------------------------------------------------------------------
  // Lors de la cration d'une variable le C++ appelle un constructeur :
  // celui-ci initialise la variable.  Cette fonction dpend videmment
  // du type en question (ici la classe Naturel), et peut prendre des 
  // paramtres (par exemple afin de raliser une conversion de type).
  //---------------------------------------------------------------------

  // Constructeur par copie d'un nombre naturel / suite de chiffres
  Naturel( const Naturel& a ) 
    : chiffres( a.chiffres ) {};
  Naturel( const vector<Chiffre>& suite ) 
    : chiffres( suite ) { normaliser(); };
  
  // Constructeur par conversion d'un petit entier en une suite de chiffres
  Naturel( int n= 0 )
    { for( ; n>0; n/=base ) chiffres.push_back( n%base ); };
  Naturel( unsigned int n )
    { for( ; n>0; n/=base ) chiffres.push_back( n%base ); };

  // Constructeur par conversion d'une chane de caractres
  Naturel( const char* s ) 
    { istringstream in(s); in >> *this; };
  Naturel( const string& s ) 
    { istringstream in(s.c_str()); in >> *this; };

  //---------------------------------------------------------------------
  // Manipulations lmentaires (affectation, rinitialisation, etc)
  //---------------------------------------------------------------------
  
  // Affectation (par copie des chiffres)
  Naturel& operator= ( const Naturel& a ) 
    { chiffres= a.chiffres; return *this; };

  // Remettre  zro (= suite vide)
  void clear()
    { chiffres.clear(); };
  
  // Tester si zro (= suite vide)
  bool zero() const
    { return chiffres.empty(); };

  // Dterminer la taille (= longueur de la suite)
  Indice size() const
    { return chiffres.size(); };
  
  // Raccourcir en effaant d'ventuels zros terminaux
  void raccourcir();

  // Normaliser la suite pour n'avoir que de chiffres 0,...,base-1
  void normaliser(); 
    
  // Lire le chiffre  la position i : indexation scurise
  Chiffre operator[] ( Indice i ) const;

  //---------------------------------------------------------------------
  // Les oprations suivantes travaillent directement sur les chiffres. 
  // Or, si l'on dclare ces donnes comme prives, alors les fonctions 
  // extrieures n'y ont accs qu'en lecture, via l'oprateur [] fourni 
  // ci-dessus.  Afin de leur donner un accs privilgi en criture,
  // on les dclare `friend'.
  //---------------------------------------------------------------------
  // Oprateurs d'entre-sortie
  friend ostream& operator<< ( ostream& out, const Naturel& a );
  friend istream& operator>> ( istream& in, Naturel& a );

  // Incrmenter a -> a+1 et decrmenter n -> n-1 (pourvu que n>0)
  friend bool incrementer( Naturel& a );
  friend bool decrementer( Naturel& a );

  // Comparaison (avec rsultat -1 si a<b, et 0 si a=0, et +1 si a>b)
  friend int  comparaison_scolaire( const Naturel& a, const Naturel& b );

  // Oprations arithmtiques : elles renvoient `true' en cas de succs 
  // et `false' en cas d'erreur (nombre ngatif, ou division par zro).
  // Les variantes sous forme d'oprateur sont implmentes plus bas.
  friend bool addition_scolaire( const Naturel& a, const Naturel& b, Naturel& somme );
  friend bool soustraction_scolaire( const Naturel& a, const Naturel& b, Naturel& diff );
  friend bool multiplication_scolaire( const Naturel& a, const Naturel& b, Naturel& produit );
  friend bool division_scolaire( const Naturel& a, const Naturel& b, Naturel& quot, Naturel& reste );

  // D'autres fonctions amies
  friend bool division( const Naturel& a, const unsigned& b, Naturel& quot, unsigned & reste );
  friend void decouper( const Naturel& n, Indice e, Naturel& n1, Naturel& n0 );
  friend bool karatsuba( const Naturel& a, const Naturel& b, Naturel& produit );
};

//=======================================================================
// Implmentation des fonctions membres de la classe Naturel.  
//=======================================================================

//-----------------------------------------------------------------------
// Raccourcir en effaant d'ventuels zros terminaux
//-----------------------------------------------------------------------
void Naturel::raccourcir()
{ 
  while ( !chiffres.empty() && chiffres.back() == Chiffre(0) ) 
    chiffres.pop_back();
};

//-----------------------------------------------------------------------
// Normaliser la suite pour n'avoir que de chiffres 0,...,base-1
//-----------------------------------------------------------------------
void Naturel::normaliser()
{
  Chiffre retenue= 0;
  for( Indice i=0; i<Indice(chiffres.size()); ++i )
  {
    chiffres[i]+= retenue;
    retenue     = chiffres[i]/base;
    chiffres[i]%= base;
  };
  while( retenue > 0 )
  { 
    chiffres.push_back( retenue % base );
    retenue/= base;
  };
  raccourcir();
};

//-----------------------------------------------------------------------
// Lire le chiffre  la position i : indexation scurise
//-----------------------------------------------------------------------
Chiffre Naturel::operator[] ( Indice i ) const
{ 
  if ( i<0 || i>=Indice(chiffres.size()) )
    return Chiffre(0);
  else 
    return chiffres[i];
};

//=======================================================================
// Implmentation des fonctions pour la classe Naturel.  
// En l'occurence il s'agit des fonctions amies qui profitent de leur
// statut privilgi afin de raliser les oprations lmentaires.  
// Tout usage de la classe Naturel passera par ces fonctions.
//=======================================================================

//-----------------------------------------------------------------------
// Conversion valeur -> symbole pour la sortie
//-----------------------------------------------------------------------
char symbole( Chiffre valeur )
{ return ( valeur>=0 && valeur<base ? chiffre[valeur] : '?' ); }

//-----------------------------------------------------------------------
// Oprateur de sortie (afficher la suite des chiffres)
//-----------------------------------------------------------------------
ostream& operator<< ( ostream& out, const Naturel& a )
{
  // Afficher la suite des chiffres, ou bien "0" pour une suite vide
  Indice i= a.chiffres.size();
  if ( i == 0 ) return ( out << 0 ); 
  for( --i; i>=0; --i ) 
    out << symbole( a.chiffres[i] );
  return out;
}

//-----------------------------------------------------------------------
// Conversion symbole -> valeur pour l'entre
//-----------------------------------------------------------------------
Chiffre valeur( char symbole )
{
  Chiffre val;
  switch( symbole )
    {
    case '0': val=  0; break;
    case '1': val=  1; break;
    case '2': val=  2; break;
    case '3': val=  3; break;
    case '4': val=  4; break;
    case '5': val=  5; break;
    case '6': val=  6; break;
    case '7': val=  7; break;
    case '8': val=  8; break;
    case '9': val=  9; break;
    case 'a': 
    case 'A': val= 10; break;
    case 'b': 
    case 'B': val= 11; break;
    case 'c': 
    case 'C': val= 12; break;
    case 'd': 
    case 'D': val= 13; break;
    case 'e': 
    case 'E': val= 14; break;
    case 'f': 
    case 'F': val= 15; break;
    default:  val= -1;
    };
  if ( val >= base ) val= -1;
  return val;
}

//-----------------------------------------------------------------------
// Oprateur d'entre (forcment un peu plus complexe que la sortie)
//-----------------------------------------------------------------------
istream& operator>> ( istream& in, Naturel& a )
{
  in.clear();   // annuler d'ventuelles erreurs prcdentes
  a.clear();    // effacer l'ancienne valeur et initialiser  0
  
  // Trouver le dbut d'une suite de chiffres
  char symbole;
  in >> symbole;
  if ( in.fail() ) return in;
  if ( valeur(symbole) < 0 )
    {
      in.putback( symbole );
      in.setstate( ios::failbit );
      return in;
    };
  
  // Lire et stocker la suite des chiffres
  for( ; in.good() && valeur(symbole) >= 0; in.get(symbole) )
    a.chiffres.push_back( valeur(symbole) );
  if ( in.good() ) in.putback( symbole );
  
  // Inverser la suite,
  for( Indice i=0, j=a.chiffres.size()-1; i<j; ++i, --j ) 
    swap( a.chiffres[i], a.chiffres[j] );
  a.raccourcir();
  return in;
}

//-----------------------------------------------------------------------
// Incrmenter un nombre naturel
// La fonction renvoie toujours `true' (excution sans erreur).
//-----------------------------------------------------------------------
bool incrementer( Naturel& a )
{
  for( Indice i=0; i<a.size(); ++i )
    if ( a.chiffres[i] == Chiffre(base-1) ) { a.chiffres[i]= Chiffre(0); }
    else { ++(a.chiffres[i]); return true; };
  a.chiffres.push_back( Chiffre(1) );
  return true;
}

// Incrmentation sous forme d'oprateur
Naturel& operator++ ( Naturel& a ) 
{ incrementer(a); return a; }

//-----------------------------------------------------------------------
// Decrmenter un nombre naturel
// La fonction renvoie `false' (erreur) si n=0.
//-----------------------------------------------------------------------
bool decrementer( Naturel& a )
{
  for( Indice i=0; i<a.size(); ++i )
    if ( a.chiffres[i] == Chiffre(0) ) { a.chiffres[i]= Chiffre(base-1); }
    else { --(a.chiffres[i]); a.raccourcir(); return true; };
  return false;
}

// Decrmentation sous forme d'oprateur  
// On n'attrape pas le seul erreur possible : on ne peut decrmenter 0.
Naturel& operator-- ( Naturel& a ) 
{ decrementer(a); return a; }

//-----------------------------------------------------------------------
// Comparaison de deux nombres naturels
// Entre : deux nombres naturels a et b
// Sortie : soit 0 si a==b, soit +1 si a>b, soit -1 si a<b.
//-----------------------------------------------------------------------
int comparaison( const Naturel& a, const Naturel& b )
{ return comparaison_scolaire( a, b ); }

// Mthode nave
// On implmente ici la dfinition mme de la relation d'ordre.
// Attention --- cet algorithme est correct mais inefficace.
int comparaison_lente( Naturel a, Naturel b )
{
  while( !a.zero() && !b.zero() ) { decrementer(a); decrementer(b); }
  if ( a.zero() && b.zero() ) return 0;
  if ( a.zero() ) return -1; else return +1;
}

// Mthode scolaire
// On suppose les deux suites de chiffres normalises dans le sens que
// d'ventuels zros terminaux ont t supprims.  Ceci permet de comparer
// d'abord les longueurs, puis chiffre par chiffre le cas chant.
int comparaison_scolaire( const Naturel& a, const Naturel& b )
{
  if ( a.chiffres.size() > b.chiffres.size() ) return +1;
  if ( a.chiffres.size() < b.chiffres.size() ) return -1;
  for( Indice i= a.chiffres.size()-1; i>=0; --i )
    {
      if ( a.chiffres[i] > b.chiffres[i] ) return +1;
      if ( a.chiffres[i] < b.chiffres[i] ) return -1;
    };
  return 0;
}

// Comparaison sous forme d'oprateur
bool operator== ( const Naturel& a, const Naturel& b ) 
{ return ( comparaison_scolaire(a,b) == 0 ); }

bool operator!= ( const Naturel& a, const Naturel& b ) 
{ return ( comparaison_scolaire(a,b) != 0 ); }

bool operator< ( const Naturel& a, const Naturel& b ) 
{ return ( comparaison_scolaire(a,b) < 0 ); }

bool operator<= ( const Naturel& a, const Naturel& b )
{ return ( comparaison_scolaire(a,b) <= 0 ); }

bool operator> ( const Naturel& a, const Naturel& b )
{ return ( comparaison_scolaire(a,b) > 0 ); }

bool operator>= ( const Naturel& a, const Naturel& b )
{ return ( comparaison_scolaire(a,b) >= 0 ); }

//-----------------------------------------------------------------------
// Addition de deux nombres naturels
// Entre : deux nombres naturels a et b
// Sortie : leur somme a+b
//-----------------------------------------------------------------------
bool addition( const Naturel& a, const Naturel& b, Naturel& somme )
{ return addition_scolaire( a, b, somme ); }

// Mthode nave
// On implmente ici la dfinition mme de l'addition.
// Attention --- cet algorithme est correct mais inefficace.
bool addition_lente( const Naturel& a, Naturel b, Naturel& somme )
{ for( somme= a; !b.zero(); incrementer(somme), decrementer(b) ); return true; }

// Mthode scolaire
// On construit la suite des chiffres qui reprsente le rsultat.
bool addition_scolaire( const Naturel& a, const Naturel& b, Naturel& somme )
{
  // Rserver de la mmoire pour recevoir le rsultat
  somme.clear();
  Indice taille= max( a.size(), b.size() );
  somme.chiffres.resize( taille, Chiffre(0) );
  
  // Calculer la somme chiffre par chiffre en tenant compte des retenues
  // NB: a[i] et b[i] utilisent notre indexation scurise (voir plus haut)
  Chiffre retenue= 0;
  for( Indice i=0; i<taille; ++i )
    {
      Chiffre temp= a[i] + b[i] + retenue;
      if ( temp < base ) { somme.chiffres[i]= temp; retenue = 0; }
      else { somme.chiffres[i]= temp - base; retenue = 1; };
    };
  
  // Rajouter ventuellement un chiffre supplmentaire pour la retenue 
  if ( retenue ) somme.chiffres.push_back(retenue);
  return true;
}

// Addition sous forme d'oprateur
Naturel operator+ ( const Naturel& a, const Naturel& b )
{ Naturel somme; addition( a, b, somme ); return somme; }

// Oprateur "en place". Une optimisation serait souhaitable.
Naturel& operator+= ( Naturel& a, const Naturel& b ) { a= a+b; return a; }

//-----------------------------------------------------------------------
// Soustraction de deux nombres naturels a et b vrifiant a >= b.
// Entre : deux nombres naturels a et b
// Sortie : le nombre naturel diff tel que a = b + diff
// La fonction renvoie `true' si une telle solution existe, `false' sinon.
//-----------------------------------------------------------------------
bool soustraction( const Naturel& a, const Naturel& b, Naturel& diff )
{ return soustraction_scolaire( a, b, diff ); }

// Mthode nave
// On implmente ici la dfinition mme de la soustraction.
// Attention --- cet algorithme est correct mais inefficace.
bool soustraction_lente( const Naturel& a, Naturel b, Naturel& diff )
{
  for( diff= a; !diff.zero() && !b.zero(); --diff, --b );
  if ( b.zero() ) { return true; } else { return false; };
}

// Mthode scolaire
// On construit la suite des chiffres qui reprsente le rsultat.
bool soustraction_scolaire( const Naturel& a, const Naturel& b, Naturel& diff )
{
  // On exige a >= b, en particulier a ne peut tre plus court que b
  diff.clear();
  if ( a.size() < b.size() ) return false;

  // Calculer la diffrence chiffre par chiffre en tenant compte des empruntes
  // NB: a[i] et b[i] utilisent notre indexation scurise (voir plus haut)
  diff.chiffres.resize( a.size(), Chiffre(0) );
  Chiffre emprunte= 0;
  for( Indice i=0; i<a.size(); ++i )
    {
      Chiffre temp= a[i] - b[i] - emprunte;
      if ( temp >= 0 ) { diff.chiffres[i]= temp; emprunte= 0; }
      else { diff.chiffres[i]= temp + base; emprunte= 1; };
    };
  
  // Emprunte restant  la fin ? Ceci veut dire que a < b, donc erreur.
  if ( emprunte ) { diff.clear(); return false; };
  diff.raccourcir();
  return true;
}

// Soustraction sous forme d'oprateur
Naturel operator- ( const Naturel& a, const Naturel& b )
{
  Naturel diff;
  if ( !soustraction( a, b, diff ) )
    cerr << "Erreur : la diffrence ngative n'est pas dfinie." << endl;
  return diff;
}

// Oprateur "en place". Une optimisation serait souhaitable.
Naturel& operator-= ( Naturel& a, const Naturel& b ) { a= a-b; return a; }

//-----------------------------------------------------------------------
// Multiplication de deux nombres naturels
// Entre : deux nombres naturels a et b
// Sortie : leur produit a*b 
//-----------------------------------------------------------------------
bool multiplication( const Naturel& a, const Naturel& b, Naturel& produit )
{ return multiplication_scolaire( a, b, produit ); }

// Mthode nave
// On implmente ici la dfinition mme de la multiplication.
// Attention --- cet algorithme est correct mais inefficace.
bool multiplication_lente( const Naturel& a, Naturel b, Naturel& produit )
{ for( produit= 0; !b.zero(); produit+= a, --b ); return true; }

// Mthode scolaire
// On construit la suite des chiffres qui reprsente le rsultat.
bool multiplication_scolaire( const Naturel& a, const Naturel& b, Naturel& produit )
{
  // Attraper le cas particulier d'un facteur nul
  produit.clear();
  if ( a.size() == 0 || b.size() == 0 ) return true;
  
  // Rserver de la mmoire pour recevoir le rsultat
  produit.chiffres.resize( a.size() + b.size(), Chiffre(0) );

  // Calculer le produit par (une variante de) la mthode scolaire
  for( Indice i=0; i<a.size(); ++i )
    {
      Chiffre aux, retenue= 0;
      for( Indice j=0; j<b.size(); ++j )
	{
	  aux= produit.chiffres[i+j] + a.chiffres[i] * b.chiffres[j] + retenue;
	  produit.chiffres[i+j]= aux % base;
	  retenue= aux / base;
	};
      produit.chiffres[i+b.size()]= retenue;
    };
  produit.raccourcir();
  return true;
}

// Multiplication sous forme d'oprateur
Naturel operator* ( const Naturel& a, const Naturel& b )
{ Naturel produit; multiplication( a, b, produit ); return produit; }

// Oprateur "en place". Une optimisation serait souhaitable.
Naturel& operator*= ( Naturel& a, const Naturel& b ) { a= a*b; return a; }

//-----------------------------------------------------------------------
// Division euclidienne par un petit diviseur
// Entre : un nombre naturel a et un petit entier b 
// Sortie : un nombre naturel q et un petit entier r 
//          tels que a = b*q+r et 0 <= r < b
// On exige que le diviseur soit non nul et plus petit que 10^8, disons.
// On assure ainsi que base*diviseur rentre encore dans un unsigned int, 
// ce qui permet d'effectuer les calculs avec ce type primitif du C++.
//-----------------------------------------------------------------------
bool division( const Naturel& a, const unsigned& b, Naturel& quot, unsigned& reste )
{
  // Effacer les anciennes valeurs de quot et reste 
  reste= 0;
  quot.clear();

  // Attraper la division par zro ou un diviseur trop grand
  if ( b == 0 || b > 100000000 ) return false;
  
  // Rserver la mmoire ncessaire pour le quotient
  quot.chiffres.resize( a.size() );

  // Effectuer la division euclidienne
  for( Indice i= a.size()-1; i>=0; --i )
    {
      reste= reste*base + a[i];
      quot.chiffres[i]= reste/b;
      reste= reste%b;
    };
  quot.raccourcir();
  return true;
}

// Quotient sous forme d'oprateur
Naturel operator/ ( const Naturel& a, unsigned b )
{
  Naturel q;
  unsigned r;
  if ( !division(a,b,q,r) )
    cerr << "Erreur : division euclidienne non dfinie." << endl;
  return q;
}

// Reste sous forme d'oprateur
unsigned operator% ( const Naturel& a, unsigned b )
{
  Naturel q;
  unsigned r;
  if ( !division(a,b,q,r) )
    cerr << "Erreur : division euclidienne non dfinie." << endl;
  return r;
}

//-----------------------------------------------------------------------
// Division euclidienne de deux nombres naturels
// Entre : deux nombres naturels a et b, avec b non nul
// Sortie : deux nombres naturels q et r tels que a = b*q+r et 0 <= r < b
// La fonction renvoie `false' si b est nul.
//-----------------------------------------------------------------------
bool division( const Naturel& a, const Naturel& b, Naturel& quot, Naturel& reste )
{ return division_scolaire( a, b, quot, reste ); }

// Mthode nave
// On implmente ici la dfinition mme de la division euclidienne.
// Attention --- cet algorithme est correct mais inefficace.
bool division_lente( const Naturel& a, Naturel b, Naturel& quot, Naturel& reste )
{
  if ( b.zero() ) return false;
  for( quot= 0, reste= a; reste>=b; reste= reste-b, incrementer(quot) );
  return true;
}

// Mthode scolaire
// On construit la suite des chiffres qui reprsente le rsultat.
bool division_scolaire( const Naturel& a, const Naturel& b, Naturel& quot, Naturel& reste )
{
  // Effacer les anciennes valeurs de quot et reste
  quot.clear(); reste.clear();
  
  // Attraper la seule erreur possible : la division par zro
  if ( b.zero() ) return false;
  
  // Attraper le cas trivial o a < b
  if ( b.size() > a.size() ) { reste= a; return true; };

  // Rserver la mmoire ncessaire pour le quotient
  quot.chiffres.resize( a.size() - b.size() + 1, Chiffre(0) );

  // Effectuer la division euclidienne 
  for( Indice i= a.size()-1; i>=0; --i )
    {
      // Calculer reste = base*reste + a[i] puis (peu de) soustractions itres
      reste.chiffres.push_back( Chiffre(0) );
      for( Indice j= reste.size()-1; j>0; --j ) 
	reste.chiffres[j]= reste.chiffres[j-1];
      reste.chiffres[0]= a.chiffres[i];
      reste.raccourcir();
      while( reste>=b ) { reste= reste-b; ++quot.chiffres[i]; };
    };
  quot.raccourcir();
  return true;
}

// Le quotient sous forme d'oprateur
Naturel operator/ ( const Naturel& a, const Naturel& b )
{
  Naturel q,r;
  if ( !division(a,b,q,r) ) 
    cerr << "Erreur : division euclidienne non dfinie." << endl;
  return q;
}

// Le reste sous forme d'oprateur
Naturel operator% ( const Naturel& a, const Naturel& b )
{
  if ( b.zero() ) return a;
  Naturel q,r; division(a,b,q,r); return r;
}

// Oprateurs "en place". Une optimisation serait souhaitable.
Naturel& operator/= ( Naturel& a, const Naturel& b ) { a= a/b; return a; }
Naturel& operator%= ( Naturel& a, const Naturel& b ) { a= a%b; return a; }

//=======================================================================
// Fin de l'implmentation du type Naturel
//=======================================================================
