//=======================================================================
// rabinowitz.cc : calcul de pi d'aprs S.Rabinowitz et S.Wagon
// La traduction en C++ ci-dessous fut crite par Michael Eisermann 
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
// 
// Description :
// Le programme suivant est une traduction en C++ du programme 
// prsent par S.Rabinowitz et S.Wagon  la fin de leur article 
//   "A spigot algorithm for the digits of pi",
//   American Mathematical Monthly 102 (1995), no. 3, 195--203.
// Il a t conu pour calculer jusqu' 1000 chiffres de pi.
// 
// Avertissement :
// Dans le programme de Rabinowitz et Wagon, certaines imprcisions 
// se sont introduites.  Tout d'abord on constate que des chiffres 
// terminaux valant 9 ne sont jamais affichs, car le programme ne 
// peut garantir leur correction.  Il semble donc honnte de ne pas 
// les afficher : au pire, si l'on exige une prcision donne, 
// il faut tout recalculer avec lgrement plus de chiffres.
//
// Plus grave est le choix erron de la variable m, 
// qui reprsente la longueur de la srie approchant pi. 
// La longueur  m = (10*n)/3  mne  une prcision insuffisante.
// La longueur  m = (10*n)/3 + 1  par contre devrait suffire.
//
//  cause de cette erreur, certains chiffres terminaux peuvent tre 
// errons sans que l'utilisateur s'aperoive.  Ceci arrive dj pour 
// n=1, ce qui est trs embarrassant mais assez facile  dtecter. 
// Ensuite pour n=32 le dernier chiffre affich est erron, ce qui 
// est plus embtant : l'utilisateur ne saurait pas, en gnral, 
// que le 32me chiffre devrait valoir 5 et non 4.
//
// Ceci ne diminue en rien le travail de S.Rabinowitz et S.Wagon, 
// mais souligne l'importance d'une implmentation soigneuse.
//=======================================================================

#include <iostream>
#include <vector>
using namespace std;

void affiche_pi( int n )
{
  int m = (10*n)/3;           // choisir une longueur m >= n * log_2(10)
  vector<int> x(m,2);         // initialiser le vecteur avec des valeurs 2
  int nines= 0;               // nombre des chiffres 9 qui attendent en queue
  int predigit= 0;            // le chiffre prcdent la queue des chiffres 9
  for ( ; n>0; --n )          // boucle pour produire n chiffres de pi
    {
      int q= 0;                     // retenue de la normalisation
      for ( int i=m-1; i>=0; --i )  // boucle parcourant le vecteur x 
	{                           //   de droite (i=m-1)  gauche (i=0)
	  int a= 10*x[i] + q*(i+1); // multiplier par 10 et ajouter la retenue
	  x[i] = a % (2*i+1);       // stocker le reste (valeur normalise)
	  q    = a / (2*i+1);       // garder la retenue pour la place i-1
	};
      x[0]= q%10;             // x[0] est la place de poids entier
      q  /= 10;               // q est alors le chiffre  afficher
      switch( q )             // Il faut distinguer plusieurs cas...
	{
	case 9:               // Un chiffre "9" peut encore changer
	  ++nines;            //  cause d'une future propagation de retenue.
	  break;           
	case 10:              // Un chiffre "10" produit une telle retenue.
	  cout << predigit+1; // On l'ajoute et affiche les chiffres dont
	  predigit= 0;        // la valeur ne peut plus changer.
	  for ( ; nines>0; --nines ) cout << 0;
	  break;
	default:              // Tout autre chiffre ne peut pas provoquer 
	  cout << predigit;   // de propagation de retenue, donc on affiche
	  predigit= q;        // maintenant les chiffres dont on est sr.
	  for ( ; nines>0; --nines ) cout << 9;
	};
      cout << flush;          // forcer l'affichage immdiat (vider le tampon)
    };                        // fin de la boucle ; n chiffres ont t affichs
  cout << predigit << endl;   // on affiche le dernier chiffre en stock
}

int main()
{
  cout << endl << "Bonjour et bienvenue au calcul de pi "
       << "d'aprs S.Rabinowitz et S.Wagon.\n"
       << "Cette version fut implmente par Michael Eisermann\n\n"
       << "Entrez le nombre de chiffres (entre 1 et 1000000) : ";
  int chiffres;
  cin >> chiffres;
  if ( chiffres < 1 || chiffres > 1000000 )
    {
      cout << "Vous dsirez calculer " << chiffres
	   << " chiffres. Je ne peux pas effectuer ce calcul.\n";
      return 0;
    };
  affiche_pi(chiffres);
  cout << "Avertissement : quelques chiffres terminaux peuvent tre errons.\n" << endl;
}
