/*=======================================================================*\
 * groupe.cc : implmentation de groupes de permutations
 * Copyright (C) 2005 by Michael Eisermann, all rights reserved
 * Institut Fourier, Universit Joseph Fourier, Grenoble, France
 *-----------------------------------------------------------------------
 * Description :
 *---------------
 * Avec la classe Groupe implmente ci-dessous on peut
 * - construire un groupe  partir d'une famille gnratrice,
 * - dterminer l'ordre du groupe et parcourir tous ses lments,
 * - tester si une permutation appartient au groupe ou non,
 * - utiliser des mthodes commodes d'entre-sortie.
 *
 * Pour rappel :
 * Les permutations d'un ensemble X forment un groupe, appel le groupe  
 * symtrique, not Sym(X). Un groupe de permutations est un sous-groupe
 * de Sym(X), c'est--dire un sous-ensemble G de Sym(X) qui contient 
 * l'identit et tous les produits et inverses de ses lments. 
 * Pour spcifier un groupe de permutations on pourrait bien sr numrer 
 * tous ses lments. Il est souvent plus commode de donner une famille 
 * gnratrice. Pour ceci, rappelons que le groupe <a,b,c,...> engendr 
 * par a,b,c,... est le plus petit groupe G contenant ces lments : 
 * c'est l'ensemble form par a,b,c,... et tous leurs produits. 
 * On appelle a,b,c,... une famille gnratrice de G.
 *
 * Remarque :
 * A premire vue il peut sembler restrictif de regarder ici seulement 
 * les groupes de permutations. Tel n'est pas le cas: tout groupe fini 
 * est isomorphe  un groupe de permutations (thorme de Cayley).
 *-----------------------------------------------------------------------
 * Structure de donnes :
 *------------------------
 * Pour travailler avec un groupe de permutations G = <X> on stocke 
 * la famille gnratrice X donne. Pour numrer les lments de G
 * et pour tester si une permutation donn appartient  G on pourrait
 * engendrer puis stocker tous ses lments, ce qui n'est raisonnable
 * que pour des petits groupes (de moins de 10000 lments disons). 
 * Afin de traiter aussi des grands groupes, on prfre ici travailler
 * avec une base de G, aussi appele une famille gnratrice forte.
 *
 * Chane des stabilisateurs :
 * Pour un sous-groupe G de Sym(n) on considre la chane des
 * sous-groupes  {id} = G(1) < G(2) < ... < G(n-1) > G(n) = G
 * o G(k) est obtenu comme l'intersection de G avec Sym(k).
 * Autrement dit, G(k) est le sous-groupe des lments de G 
 * qui fixent {k+1,k+2,...,n} point par point.
 *
 * Bases :
 * En faisant oprer G(k) sur k on obtient l'orbite  G(k).k  et pour 
 * chaque lment i de cette orbite on peut choisir une transversale
 * t_k^i dans G(k) telle que t_k^i(k) = i. Par exemple, pour i=k,
 * on peut toujours choisir l'identit t_k^k = id, mais pour les 
 * autres points de l'orbite il n'y a pas de choix prfrs.
 * On note B(k) = { t_k^i } la famille de ces transversales, et 
 * on appelle *base* de G la collection B = (B(1),B(2),...,B(n)).
 *
 *   Proposition :
 *   Tout sous-groupe G de Sym(n) admet une base (B(1),B(2),...,B(n)).
 *   Ayant choisi une base, tout lment g de G s'crit de manire 
 *   unique comme  g = b(n) b(n-1) ... b(2) b(1)  avec b(k) dans B(k).
 *
 * Bases rduites :
 * Dans la transversale on peut toujours choisir t_k^k=id, il est donc 
 * inutile de stocker cette permutation. On appelle *base rduite* la 
 * collection (B'(1),B'(2),...,B'(n)) avec B'(k) = { t_k^i pour i<k }.
 * Bases et bases rduites se correspondent de manire vidente.
 * La classe Groupe travaille avec des bases rduites, ce qui conomise 
 * de la mmoire mais complique lgrement l'implmentation des mthodes 
 * car il faut distinguer les cas i=k et i<k.
 *-----------------------------------------------------------------------
 * Algorithmes :
 *---------------
 * Le stockage d'un groupe par une base est trs efficace car on ne stocke 
 * qu'un petit nombre de permutations : au maximum n(n-1)/2 lments !
 * Si l'on dispose d'une base de G on peut facilement dduire l'ordre 
 * de G et numrer tous ses lments. Il est galement possible de 
 * tester si une permutation donne appartient  G. 
 *
 * Le seul point dlicat est d'effectivement *construire* une base
 *  partir d'une famille gnratrice de G. Ce problme est rsolu 
 * par l'algorithme de Schreier-Sims, implment plus bas.
 * (Pour un dveloppement dtaill voir l'excellente introduction 
 * de G.Butler, Fundamental Algorithms for Permutation Groups, 
 * Lecture Notes in Computer Science 559, Springer-Verlag, Berlin 1991.)
\*=======================================================================*/

#include <set>
#include <map>
#include "permut.cc"
using namespace std;

/*-----------------------------------------------------------------------*\
 * Dfinir un type pour stocker les entiers (grands entiers avec GMP)
 * Il se peut que les nombres calculs risquent de dpasser la capacit 
 * des types primitifs du C++, ici 2^63 = 9e18 pour long long int.
 * Dans ce cas il faut prendre une classe de grands entiers, comme
 * mpz_class de la bibliothque GMP ( compiler avec l'option -lgmpxx).
\*-----------------------------------------------------------------------*/
// #include <gmpxx.h>
// typedef mpz_class Integer;
typedef long long int Integer;

/*-----------------------------------------------------------------------*\
 * Dfinition de la classe Groupe
\*-----------------------------------------------------------------------*/
class Groupe
{
private:
  // Quelques types et classes qui sont utiliss dans la suite
  typedef vector<Permut> PGens;           // familles gnratrices
  typedef Permut::Point Point;            // points permuts
  typedef map<Point,Permut> OrbRep;       // reprsentants d'une orbite
  typedef OrbRep::const_iterator OrbIter; // itrateur dans une orbite
  typedef vector<OrbRep> PBase;           // base (voir la prambule)

private:
  // Les donnes (prives) stockes pour un groupe de permutations
  PGens   lesGens;         // la famille gnratrice dfinissant le groupe
  PBase   laBase;          // une base (cd une famille gnratrice forte)
  Integer lOrdre;          // l'ordre du groupe calcul une fois pour toute
  void extendBase( Point n, Permut p );

public:
  // Constructeurs divers 
  Groupe()                    // construire le groupe trivial {id} 
    : lOrdre(1) {};
  Groupe( const Permut& p )   // construire le groupe cyclique <p>
    : lOrdre(1) { insert(p); };
  Groupe( const Groupe& g )   // constructeur par copie
    : lesGens( g.lesGens ), laBase( g.laBase ), lOrdre( g.lOrdre ) {};
      
  // Oprateur d'affectation
  Groupe& operator= ( const Groupe& g ) 
    { lesGens= g.lesGens; laBase= g.laBase; lOrdre= g.lOrdre; return *this; };
  void swap( Groupe& g )
    { lesGens.swap(g.lesGens); laBase.swap(g.laBase); std::swap(lOrdre,g.lOrdre); };
  
  // Effacer toutes les donnes et recommencer par le groupe trivial
  void clear()
    { lesGens.clear(); laBase.clear(); lOrdre= 1; };
  
  // Ajouter un gnrateur ou une liste de gnrateurs
  void insert( const Permut& p );
  void insert( const PGens& g )
    { for( int i=0; i<g.size(); ++i ) insert( g[i] ); };
  
  // Accder aux gnrateurs, la base, et l'ordre du groupe
  const PGens& generateurs() const
    { return lesGens; };
  const PBase& base() const
    { return laBase; };
  const Integer& ordre() const
    { return lOrdre; };

  // Tester l'appartenance d'un lment, convertir entre lment et numro
  bool contient( Permut p ) const;
  Integer operator[] ( Permut perm ) const;
  Permut  operator[] ( Integer num ) const;
  
  // Tester l'inclusion d'un sous-groupe
  bool contient( const Groupe& g ) const
    { 
      for ( int i=0; i<g.lesGens.size(); ++i )
        if ( !contient( g.lesGens[i] ) ) return false;
      return true; 
    };
  
  // Tester l'galit ou ingalit de deux groupes
  friend bool operator == ( const Groupe& g, const Groupe& h )
    { return ( g.contient(h) && h.contient(g) ); };
  friend bool operator != ( const Groupe& g, const Groupe& h )
    { return ( !g.contient(h) || !h.contient(g) ); };
  
  // Tester si le groupe est ablien, construire le centre
  bool estAbelien() const;                                  // exercice !
  Groupe centre() const;                                    // exercice !

  // Tester si le groupe commute avec une permutation p
  bool centralise( const Permut& p ) const;                 // exercice !
  Groupe centralisateur( const Permut& p ) const;           // exercice !

  // Tester si le groupe commute avec un autre groupe g
  bool centralise( const Groupe& g ) const;                 // exercice !
  Groupe centralisateur( const Groupe& g ) const;           // exercice !
  
  // Construire le normalisateur d'un sous-groupe
  bool normalise( const Groupe& g ) const;                  // exercice !
  Groupe normalisateur( const Groupe& g ) const;            // exercice !
  
  // Construire les classes de conjugaison
  set<Permut> conj_class( const Permut& p ) const;          // exercice !
  vector< set<Permut> > conj_classes() const;               // exercice !

  // Tester si un groupe est simple (chercher un sous-groupe distingu)
  bool estSimple( Groupe& dist ) const;                     // exercice !
};

/*-----------------------------------------------------------------------*\
 * Tester si une permutation appartient au groupe 
\*-----------------------------------------------------------------------*/
bool Groupe::contient( Permut perm ) const
{
  Point n= perm.degre();
  if ( n > laBase.size() ) return false;
  for( --n; n>0; --n )
    {
      Point k= perm(n);
      if ( k < n )
	{
	  const OrbRep& rep= laBase[n];
	  OrbIter it= rep.find(k);
	  if ( it == rep.end() ) return false;
	  perm= !(it->second) * perm;
	}
    }
  return true;
}

/*-----------------------------------------------------------------------*\
 * Calculer le numro associ  un lment du groupe.
 *  noter que la numrotation dpend de la base choisie.
\*-----------------------------------------------------------------------*/
Integer Groupe::operator[] ( Permut perm ) const
{
  Point n= perm.degre();
  if ( n > laBase.size() ) return -1;
  Integer number= 0;
  for( --n; n>0; --n )
    {
      const OrbRep& rep= laBase[n];
      number*= Integer( rep.size() + 1 );
      Point k= perm(n);
      if ( k < n )
	{
	  OrbIter it= rep.find(k);
	  if ( it == rep.end() ) return -1;
	  perm= !(it->second) * perm;
	  while( it != rep.end() ) 
	    { ++it; ++number; };
	}
    }
  return number;
}

/*-----------------------------------------------------------------------*\
 * Calculer l'lment du groupe associe  un numro.
 *  noter que la numrotation dpend de la base choisie.
\*-----------------------------------------------------------------------*/
Permut Groupe::operator[] ( Integer num ) const
{
  Permut perm;
  for( Point n=0; n<laBase.size() && num>0; ++n )
    {
      const OrbRep& rep= laBase[n];
      Integer i( num % ( rep.size() + 1 ) );
      if( i > 0 )
	{
	  OrbIter it= rep.end();
	  while( i > 0 ) 
	    { --i; --it; };
	  perm= it->second * perm;
	}
      num/= Integer( rep.size() + 1 );
    }
  return perm;
}

/*-----------------------------------------------------------------------*\
 * Ajouter un gnrateur au groupe : s'il n'appartient pas encore 
 * au groupe, on l'ajoute  la liste des gnrateurs puis on met  jour
 * la base par un appel initial  Schreier-Sims (dbut de la rcursion).
\*-----------------------------------------------------------------------*/
void Groupe::insert( const Permut& perm )
{
  // Attraper le cas trivial o le gnrateur est dj dans le groupe
  if ( contient( perm ) ) return;

  // Sinon, le rajouter  la liste des gnrateurs et tendre la base
  lesGens.push_back( perm );
  Point n= perm.degre();
  if ( laBase.size() < n ) laBase.resize( n );
  extendBase( Point(laBase.size()-1), perm );

  // Recalculer l'ordre du groupe aprs extension
  lOrdre= 1;
  for( Point n=0; n<laBase.size(); ++n )
    lOrdre *= Integer( laBase[n].size() + 1 );
}

/*-----------------------------------------------------------------------*\
 * Construire une base d'un groupe en ajoutant des gnrateurs : Soit G 
 * un groupe donn par une base rduite et  perm  une permutation.
 * La fonction suivante construit une base rduite pour < G, perm >.
 *-----------------------------------------------------------------------
 * On implmente ici l'algorithme rcursif de Schreier-Sims, commenant 
 * par le degr maximum n. On construit l'orbite de n sous l'action
 * de H = < G, perm > en faisant agir chaque gnrateur h de H sur 
 * les points k de l'orbite dj construit. Deux cas se prsentent:
 * - Soit h(k) est un nouvel lment, dans quel cas on le rajoute 
 *   afin de construire l'orbite  Hn  toute entire.
 * - Soit h(k) a dj t construit, dans quel cas on peut en dduire 
 *   un lment du stabilisateur de n, que l'on passe au degr n-1.
 * L'algorithme arrive ainsi de construire une base de H.
\*-----------------------------------------------------------------------*/
void Groupe::extendBase( Point n, Permut perm )
{
  // Tester d'abord si perm est dj contenue dans le groupe
  if( contient( perm ) ) return;

  // Construire les nouveaux lments dans l'orbite de n
  // obtenus en faisant agir la nouvelle permutation perm.
  OrbRep& rep= laBase[n];
  vector<Point> nouveau;
  OrbIter it;
  
  // On commence par l'action de perm sur le point de base n.
  Point im= perm(n);
  if ( im == n ) 
    extendBase( n-1, perm );
  else 
    // if ( rep.find(im) != rep.end() )
    //   extendBase( n-1, !rep[im] * perm );
    if ( (it= rep.find(im)) != rep.end() )
      extendBase( n-1, !(it->second) * perm );
    else
      {
	nouveau.push_back( im );
	rep[im]= perm;
      };
  
  // Puis on fait agir perm sur les autres points de l'orbite.
  for( OrbIter j= rep.begin(); j!=rep.end(); ++j )
    {
      // On considre le point k = t(n) obtenu par l'action de t.
      Point k= j->first;
      const Permut& t= j->second;
      
      // On fait agir le nouveau gnrateur sur le point k : 
      // Si l'on retombe sur n ou un autre point dj construit, 
      // alors on en dduit un lment du stabilisateur de n.
      // Sinon on l'ajoute comme un nouvel lment de l'orbite.
      Point im= perm(k);
      if ( im == n ) 
	extendBase( n-1, perm * t );
      else 
	// if ( rep.find(im) != rep.end() )
	//   extendBase( n-1, !rep[im] * perm * t );
	if ( (it= rep.find(im)) != rep.end() )
	  extendBase( n-1, !(it->second) * perm * t );
	else
	  {
	    nouveau.push_back( im );
	    rep[im]= perm * t;
	  }
    }

  // Complter l'orbite en faisant agir tous les gnrateurs.
  // Chaque fois que l'on cre un nouveau point dans l'orbite
  // on le stocke dans la liste  nouveau  pour un traitement ultrieur.
  for( int l=0; l<nouveau.size(); ++l )
    {
      // On traite le point k = t(n) obtenu par l'action de t.
      Point k= nouveau[l];
      const Permut& t= rep[k];

      // On fait agir le nouveau gnrateur sur le point k.
      // Si l'on retombe sur n ou un autre point dj construit, 
      // alors on en dduit un lment du stabilisateur de n.
      // Sinon on l'ajoute comme un nouvel lment de l'orbite.
      Point im= perm(k);
      if ( im == n ) 
	extendBase( n-1, perm * t );
      else 
	// if ( rep.find(im) != rep.end() )
	//   extendBase( n-1, !rep[im] * perm * t );
	if ( (it= rep.find(im)) != rep.end() )
	  extendBase( n-1, !(it->second) * perm * t );
	else
	  {
	    nouveau.push_back( im );
	    rep[im]= perm * t;
	  }
      
      // Faire agir les permutations de la base sur le point k.
      for( Point i=k; i<=n; ++i )
	{
	  const OrbRep& brep= laBase[i];
	  for( OrbIter j= brep.begin(); j!=brep.end(); ++j )
	    {
	      // On fait agir le gnrateur perm sur le point k.
	      // Si l'on retombe sur n ou un autre point dj construit, 
	      // alors on en dduit un lment du stabilisateur de n.
	      // Sinon on l'ajoute comme un nouvel lment de l'orbite.
	      const Permut& perm= j->second; 
	      Point im= perm(k);
	      if ( im == n ) 
		extendBase( n-1, perm * t );
	      else 
		// if ( rep.find(im) != rep.end() )
		//   extendBase( n-1, !rep[im] * perm * t );
		if ( (it= rep.find(im)) != rep.end() )
		  extendBase( n-1, !(it->second) * perm * t );
		else
		  {
		    nouveau.push_back( im );
		    rep[im]= perm * t;
		  }
	    }
	}
    }
}

/*-----------------------------------------------------------------------*\
 * Oprateur de sortie
 *-----------------------------------------------------------------------
 * On numre la famille gnratrice choisie, sous la forme <a,b,c,...>.
 * Ce format est beaucoup plus conomique qu'numrer tous les lments !
\*-----------------------------------------------------------------------*/
ostream& operator << ( ostream& out, const Groupe& g )
{
  write( out, g.generateurs(), "<,>" );
  return out;
}

/*-----------------------------------------------------------------------*\
 * Oprateur d'entre
 *-----------------------------------------------------------------------
 * Comme toujours, l'entre est symtrique  la sortie : on lit une 
 * famille de permutations sous la forme <a,b,c,...>, puis on en 
 * construit le groupe engendr.
\*-----------------------------------------------------------------------*/
istream& operator >> ( istream& in, Groupe& g )
{
  vector<Permut> generateurs;
  read( in, generateurs, "<,>" );
  g.clear();
  g.insert(generateurs);
  return in;
}

/*-----------------------------------------------------------------------*\
 * Un exemple d'utilisation ( mettre en commentaire aprs des tests)
 *-----------------------------------------------------------------------
 * Quelques groupes de plus en plus grands pour tester le logiciel :
 * - Le groupe didral D(4) d'ordre 8
 *   < ((0,2)), ((0,1,2,3)) >
 * - Le groupe altern Alt(4) d'ordre 12
 *   < ((0,1,2)), ((1,2,3)) >
 * - Le groupe symtrique Sym(4) d'ordre 24
 *   < ((0,1,2,3)), ((2,3)) >
 * - Un groupe d'ordre 24
 *   < ((0,1,2)(3,4,5)), ((2,3)) >
 * - Un groupe d'ordre 72
 *   < ((0,1)(2,3)(4,5)), ((1,2)), ((3,4)) >
 * - Le groupe didral D(10) d'ordre 20
 *   < ((0,1,2,3,4,5,6,7,8,9)), ((9,0)(8,1)(7,2)(6,3)(5,4)) >
 * - Un groupe d'ordre 2^7 = 128 dans Sym(8) d'ordre 2^73^257
 *   < ((0,1)), ((0,2)(1,3)), ((0,4)(1,5)(2,6)(3,7)) >
 * - Le groupe de Mathieu M(11) d'ordre 7920
 *   < ((0,1,2,3,4,5,6,7,8,9,10)), ((0,1,2,4,9,8,10,3,6,7,5)) >  
 * - Le groupe altern Alt(9) d'ordre 9!/2 = 181440
 *   < ((0,1,2)), ((0,1,2,3,4,5,6,7,8)) >
 * - Le groupe symtrique Sym(9) d'ordre 9! = 362880 
 *   < ((0,1)), ((0,1,2,3,4,5,6,7,8)) >
 * - Un groupe d'ordre 3^13 = 1594323 dans S(27) 
 *   < ((0,1,2)), ((0,3,6)(1,4,7)(2,5,8)), ((0,9,18)(1,10,19)(2,11,20)
     (3,12,21)(4,13,22)(5,14,23)(6,15,24)(7,16,25)(8,17,26)) >
 * - Le groupe symtrique Sym(15) d'ordre 15! = 1307674368000
 *   < ((0,1)), ((0,1,2,3,4,5,6,7,8,9,10,11,12,13,14)) >
\*-----------------------------------------------------------------------*/
/*
int main()
{
  Groupe m;
  m.insert( Permut("((0,1,2,3,4,5,6,7,8,9,10))") );
  m.insert( Permut("((0,1,2,4,9,8,10,3,6,7,5))") );
  cout << "\nBonjour et bienvenue aux groupes de permutations !" << endl
       << "Vous pouvez entrer des permutations et regarder quel groupe" << endl
       << "elles engendrent. Voici par exemple un groupe d'ordre " 
       << m.ordre() << " :" << endl << m << endl;

  cout << "Donner un groupe svp : ";
  Groupe g;
  cin >> g; 
  cout << "Vous avez entr " << g << "." << endl;
  cout << "Ce groupe est d'ordre " << g.ordre() << endl;
  cout << "Voici la liste de ses lments : " << endl;
  Integer n= g.ordre(); 
  if ( n>100) n= 100;
  cout << "{ " << g[Integer(0)];
  for( Integer k=1; k<n; ++k ) cout << ", " << g[k];
  if ( n < g.ordre() ) cout << ", ..."; 
  cout << " }" << endl;
  
  for(;;)
    {
      cout << "\nAjouter une permutation svp : ";
      Permut p;
      cin >> p;
      cout << "Vous avez entr " << p << endl;
      if ( p.isIdentity() ) break;

      g.insert(p);
      cout << "On obtient ainsi le groupe " << g << endl;
      cout << "Ce groupe est d'ordre " << g.ordre() << endl;
      cout << "Voici la liste de ses lments : " << endl;
      Integer n= g.ordre(); 
      if ( n>100) n= 100;
      cout << "{ " << g[Integer(0)];
      for( Integer k=1; k<n; ++k ) cout << ", " << g[k];
      if ( n < g.ordre() ) cout << ", ..."; 
      cout << " }" << endl;
    };
  
  cout << "Au revoir et  bientt.\n" << endl;
}
/**/
