/*=======================================================================*\
 * orbite.cc : engendrer une orbite sous l'action d'un groupe
 * Copyright (C) 2005 by Michael Eisermann, all rights reserved
 * Institut Fourier, Universit Joseph Fourier, Grenoble, France
 *-----------------------------------------------------------------------
 * Description :
 *---------------
 * L'implmentation ci-dessus permet d'engendrer l'orbite sous l'action 
 * d'un groupe G=<X> donn par une famille gnratrice X={x_1,...,x_n}.
 * Dans notre implmentation G est un groupe de permutations engendr
 * par les permutations x_1,...,x_n, appeles "les gnrateurs" de G. 
 * L'algorithme reste valable pour tout groupe donn par une famille
 * gnratrice finie, pourvu que l'orbite soit finie.
 *-----------------------------------------------------------------------
 * Algorithme :
 *--------------
 * On procde par la mthode vidente. On commence par l'lment a dont 
 * on cherche  construire l'orbite. On fait agir chaque gnrateur x_k
 * sur a et on rajoute  l'orbite les nouveaux lments ainsi engendrs. 
 * Puis on recommence avec chacun des nouveaux lments : on fait agir
 * les gnrateurs et on rajoute  l'orbite les nouveaux lments, etc.
 * Cet algorithme se termine car l'orbite Ga est fini. Aprs terminaison 
 * on a construit le plus petit ensemble stable par l'action de G et
 * contenant a : c'est l'orbite de a sous l'action de G, comme souhait.
\*=======================================================================*/

#include <set>
#include <map>
using namespace std;

#include "groupe.cc"
typedef Permut::Point Point;

/*-----------------------------------------------------------------------*\
 * Construire l'orbite d'un point a sous l'action d'un groupe G = <X>
 *-----------------------------------------------------------------------
 * Pour l'orbite  construire on choisit ici la structure d'ensemble :
 * en utilisant un ordre (quelconque) sur les lments de l'orbite ceci 
 * permet d'ajouter puis de chercher des lments efficacement. On stocke 
 * les nouveaux lments dans un vecteur, qui nous sert de pile pour le
 * traitement ultrieur. Quand les lments sont des objets relativement 
 * grands, il est utile de ne pas stocker les lments eux-mmes, mais 
 * un pointer vers l'orbite dj construit. Ainsi chaque lment sera 
 * stock seulement une fois.
\*-----------------------------------------------------------------------*/
template <typename Gens, typename Point, typename Action>
void orbiteGauche
( const Gens& lesGens,    // les gnrateurs du groupe G
  const Action& lAction,  // l'action du groupe sur les points
  const Point& lePoint,   // le point a dont on cherche l'orbite
  set<Point>& lOrbite )   // l'orbite Ga  construire
{
  // Quelques types de la biliothques STL qui apparaissent dans la suite
  typedef typename Gens::const_iterator GensIter;
  typedef typename set<Point>::const_iterator OrbIter;
  typedef pair<OrbIter,bool> OrbIns;
  
  // Initialiser l'orbite en commenant par le point a.
  // La fonction insert() renvoie une paire (i,b) o i est
  // un pointer sur l'lment insr et b est un boolen
  // qui vaut  true  si une insertion a eu lieu, et  false
  // si l'lment en question appartenait dj  l'ensemble.
  lOrbite.clear();
  OrbIns ins= lOrbite.insert( lePoint );

  // On stocke le pointeur sur le nouvel lment a, ici 
  // le pointer i = ins.first. (On sait d'ailleurs que 
  // b = ins.second vaut true car l'ensemble tait vide.)
  vector<OrbIter> nouveau;
  nouveau.push_back( ins.first );

  // On fait agir les gnrateurs sur les nouveaux lments.
  // Chaque fois que l'on cre un nouvel lment on stocke
  // un pointeur dans la liste  nouveau  pour un traitement
  // ultrieur. On peut ainsi ritrer le procd.
  while( !nouveau.empty() )
    {
      Point b= *(nouveau.back());
      nouveau.pop_back();
      for( GensIter g= lesGens.begin(); g != lesGens.end(); ++g )
	{
	  Point c= lAction( *g, b );
	  OrbIns ins= lOrbite.insert( c );
	  if ( ins.second ) 
	    {
	      nouveau.push_back( ins.first );
	      cout << b << "->" << c << " par " << (*g) << " : ";
	    }
	}
    }
}

/*-----------------------------------------------------------------------*\
 * Construire l'orbite et le stabilisateur d'un point a 
 *-----------------------------------------------------------------------
 * On construit l'orbite Ga, et en mme temps pour chaque lment b de Ga 
 * un lment t_b du groupe G de sorte que t_b(a) = b. Les paires (b,t_b) 
 * sont stockes comme une application b -> t_b, c'est--dire un ensemble 
 * de paires tries par la premire coordonne, qui sert dans la suite de 
 * clef pour la recherche. L'application b -> t_b est aussi appele une
 * "transversale" ou un "vecteur de Schreier" pour l'orbite Ga.
 *
 * Chaque fois que l'on fait agir un gnrateur g sur un point b de Ga,
 * on sait dj que t_b(a) = b. Deux cas se prsentent alors :
 * - Soit c:= g(b) n'est pas encore dans l'orbite, dans quel cas 
 *   on le rajoute, et on pose t_c := g*t_b pour la transversale.
 * - Soit c:= g(b) est dj dans l'orbite, dans quel cas t_c^(-1)*g*t_b
 *   est un lment du stabilisateur de a. (Exercice : le vrifier.)
 * Le lemme de Schreier dit qu'on obtient ainsi une famille gnratrice 
 * du stabilisateur, ce qui est exploit dans la fonction suivante
\*-----------------------------------------------------------------------*/
template <typename Gens, typename Point, typename Action, typename Groupe>
void orbiteGauche
( const Gens& lesGens,    // les gnrateurs du groupe G
  const Action& lAction,  // l'action du groupe sur les points
  const Point& lePoint,   // le point a dont on cherche l'orbite
  set<Point>& lOrbite,    // l'orbite Ga  construire
  Groupe& stabilisateur ) // reoit les gnrateurs du stabilisateur
{
  // Quelques types de la biliothques STL qui apparaissent dans la suite
  typedef typename Gens::const_iterator GensIter;
  typedef typename Gens::value_type GrpEl;
  typedef pair<Point,GrpEl> PGPair;
  typedef map<Point,GrpEl> OrbRep;
  typedef typename OrbRep::const_iterator OrbIter;
  typedef pair<OrbIter,bool> OrbIns;

  // On construit l'orbite Ga en commenant par a et t_a = id.
  OrbRep orb;
  OrbIns ins= orb.insert( PGPair( lePoint, GrpEl() ) );

  // On stocke le pointeur sur le nouvel lment a
  vector<OrbIter> nouveau;
  nouveau.push_back( ins.first );

  // Il ne nous reste qu' complter l'orbite et le stabilisateur
  stabilisateur.clear();
  while( !nouveau.empty() )
    {
      // On rcupre de la pile le pointeur sur l'lment  traiter.
      // Soit b notre lment et t sa transversale, vrifiant t(a)=b.
      OrbIter it= nouveau.back();
      nouveau.pop_back();
      const Point& b= it->first;
      const GrpEl& t= it->second;

      // On fait agir les gnrateurs sur les nouveaux lments.
      for( GensIter g= lesGens.begin(); g != lesGens.end(); ++g )
	{
	  // On essaie d'insrer c:= g(b) avec transversale s:= g*t
	  Point c= lAction( *g, b );
	  GrpEl s= (*g) * t;
	  cout << b << "->" << c << " par " << (*g) << " : ";

	  OrbIns ins= orb.insert( PGPair( c, s ) );
	  if ( ins.second )
	    {
	      // Si l'lment est nouveau on le rajoute  l'orbite et
	      // stocke un pointeur pour un traitement ultrieur.
	      nouveau.push_back( ins.first );
	      cout << "nouvel lment" << endl;
	    }
	  else
	    {
	      // Si l'on tombe sur un lment de l'orbite dj construit, 
	      // on en dduit un lment du stabilisateur.
	      const GrpEl& u= ins.first->second;
	      stabilisateur.insert( !u * s );
	      cout << "relation " << !u * s << endl;
	    }
	}
    }

  // Copier les points de l'orbite, oublier la transversale 
  lOrbite.clear();
  for( OrbIter it= orb.begin(); it != orb.end(); ++it )
    lOrbite.insert( it->first );
}

/*-----------------------------------------------------------------------*\
 * Action naturelle des permutations sur les points (par "valuation")
\*-----------------------------------------------------------------------*/
struct evaluation
{ 
  Point operator() ( const Permut& g, const Point& p ) const 
    { return g(p); } 
};

/*-----------------------------------------------------------------------*\
 * Actions des permutations sur les permutation par multiplication
\*-----------------------------------------------------------------------*/
struct multGauche 
{ 
  Permut operator() ( const Permut& g, const Permut& h ) const 
    { return g*h; } 
};

struct multDroite
{ 
  Permut operator() ( const Permut& h, const Permut& g ) const 
    { return h*g; } 
};

/*-----------------------------------------------------------------------*\
 * Actions des permutations sur les permutation par conjugaison 
\*-----------------------------------------------------------------------*/
struct conjGauche
{ 
  Permut operator() ( const Permut& g, const Permut& h ) 
    const { return g*h*(!g); } 
};

struct conjDroite
{ 
  Permut operator() ( const Permut& g, const Permut& h ) 
    const { return (!g)*h*g; } 
};

/*-----------------------------------------------------------------------*\
 * Un exemple d'utilisation 
\*-----------------------------------------------------------------------*/
int main()
{
  cout << "\nBienvenue aux groupes et orbites !";
  vector<Permut> generateurs;
  for(;;)
    {
      cout << "\nActuellement notre groupe est donn par la famille gnratrice" << endl;
      write( cout, generateurs, "<,>" );
      cout << "\nQue dsirez-vous faire ?" << endl
	   << "1) redfinir la famille gnratrice," << endl
	   << "2) faire agir sur un point par l'action naturelle," << endl
	   << "3) faire agir sur une permutation par multiplication  gauche," << endl
	   << "4) faire agir sur une permutation par conjugaison  gauche," << endl
	   << "0) terminer le programme." << endl << "Entrez votre choix svp : ";
      int choix;
      cin >> choix;
      switch( choix )
	{
	case 0:
	  {
	    cout << "Au revoir et  bientt.\n" << endl;
	    exit(0);
	  }
	case 1:
	  {
	    cout << "\nEntrez une famille gnratrice svp : " << endl;
	    read( cin, generateurs, "<,>" );
	    break;
	  }
	case 2:
	  {
	    cout << "\nEntrez un point svp : ";
	    Point point;
	    cin >> point;
	    set<Point> orb;
	    Groupe stab;
	    orbiteGauche( generateurs, evaluation(), point, orb, stab );
	    cout << "L'orbite de " << point << " est ";
	    write( cout, orb, "{,}" ); cout << endl;
	    cout << "Le stabilisateur de " << point << " est le groupe "
		 << stab << " d'ordre " << stab.ordre() << endl;
	    break;
	  }
	case 3:
	  {
	    cout << "\nEntrez une permutation svp : ";
	    Permut perm;
	    cin >> perm;
	    set<Permut> classe;
	    orbiteGauche( generateurs, multGauche(), perm, classe );
	    cout << "La classe  droite de " << perm << " est ";
	    write( cout, classe, "{,}" ); cout << endl;
	    break;
	  }
	case 4:
	  {
	    cout << "\nEntrez une permutation svp : ";
	    Permut perm;
	    cin >> perm;
	    set<Permut> classe;
	    Groupe cent;
	    orbiteGauche( generateurs, conjGauche(), perm, classe, cent );
	    cout << "La classe  droite de " << perm << " est ";
	    write( cout, classe, "{,}" ); cout << endl;
	    cout << "Le centralisateur de " << perm << " est le groupe "
		 << cent << " d'ordre " << cent.ordre() << endl;
	    break;
	  }
	default:
	  cout << "Option non reconnue." << endl;
	}
    }   
}
