/*=======================================================================*\
 * permut.cc : implmentation d'une classe modlisant les permutations
 * Copyright (C) 2003-2005 by Michael Eisermann, all rights reserved
 * Institut Fourier, Universit Joseph Fourier, Grenoble, France
 *-----------------------------------------------------------------------
 * Description :
 *---------------
 * La classe Permut implmente ci-dessous permet de calculer avec 
 * des permutations de nombres naturels, en particulier elle permet
 * - de lire et crire des permutations (en produit de cycles),
 * - multiplier et inverser des permutations.
 *
 * Rappels des permutations :
 * Une permutation d'un ensemble X est une bijection p : X -> X,
 * c'est--dire pour tout b il existe un unique a tel que p(a)=b.
 * Ceci permet de dfinir la bijection inverse q en posant q(b)=a,
 * on assure ainsi que la composition vrifie p*q = q*p = identit.
 * Dans la suite l'ensemble X sera fini, disons X = {1,...,n}, ou bien 
 * X = {0,...,n-1} pour accomoder le stockage sur ordinateur.
 *
 * Structure de groupe :
 * On dfinit le produit p*q de deux permutations p et q comme 
 * la composition de fonctions, c'est--dire (p*q)(a) := p(q(a)).
 * Autrement dit, les permutations agissent  gauche sur l'ensemble X.
 * Cette composition est associative, l'identit sert d'lment neutre,
 * et on a dj vu que chaque permutation admet une permutation inverse.
 * Ainsi les permutations d'un ensemble X forment un groupe, appel
 * le groupe symtrique, not Sym(X).
 *-----------------------------------------------------------------------
 * Structure de donnes :
 *------------------------
 * Pour le stockage interne une permutation est considre comme
 * une application de l'ensemble X = {0,...,n-1} dans lui-mme.
 * Ceci permet de stocker toute permutation p : X -> X par le vecteur
 * [p(0),p(1),...,p(n-1)] de longeur n qui numre les images de p. 
 *
 * Convention de minimalit :
 * La longueur n sera toujours choisie minimale : on raccourcit 
 * le vecteur tant que le dernier point est fixe, p(n-1) = n-1. 
 * Ainsi l'identit est reprsent par le vecteur vide [], 
 * la transposition 2 <-> 3 par le vecteur [0,1,3,2], et 
 * le cycle  3 -> 4 -> 5 -> 3  par le vecteur [0,1,2,4,5,3].
 *
 * Degr d'une permutation :
 * Le degr d'une permutation p est le nombre naturel d minimal tel que
 * p(k) = k pour t >= d. Ainsi le degr de l'identit est 0, le degr de 
 * la permutation [1,2,0,3,4] vaut 3 : elle sera raccourcie  [1,2,0].
 *
 * Convention pour l'entre-sortie :
 * Dans l'usage mathmatique les points sont traditionnellement numrots 
 * 1,2,3,...,n au lieu de 0,1,2,...,n-1, utilis pour le stockage interne. 
 * Les oprateurs d'entre-sortie pourraient faire la traduction, mais 
 * nous n'implmentons pas cette approche. Il est toujours possible de
 * travailler avec 1,2,3,...,n seulement, et le gaspillage de mmoire
 * semble ngligeable dans la plupart des applications.
\*=======================================================================*/

#include <iostream>
#include <algorithm>
#include <string>
#include <vector>
using namespace std;

// #include <isstream>
#include <strstream>
typedef istrstream isstream;

/*-----------------------------------------------------------------------*\
 * Dfinir un type pour stocker les entiers (ou grands entiers avec GMP)
\*-----------------------------------------------------------------------*/
// #include <gmpxx.h>
// typedef mpz_class Int;
typedef long long int Int;

/*-----------------------------------------------------------------------*\
 * class Permut : dfinition de la classe
\*-----------------------------------------------------------------------*/
class Permut
{
public:
  // Trois types de donnes auxiliaires utiliss dans la suite
  typedef short int Point;      // type pour les points  permuter
  typedef vector<Point> Cycle;  // type utilis pour un cycle
  typedef vector<Cycle> Cycles; // type utilis pour une liste de cycles
  
private:
  // Reprsentation interne d'une permutation
  vector<Point> leVecteur;      // vecteur pour stocker la permutation
  void raccourcir( void );      // supprimer les points fixes  la fin
  void rallonger( Point deg );  // ajouter des points fixes  la fin
  Point& operator[] ( Point a ) 
    { return leVecteur[a]; };

public:
  // Opration sur les points (sans contrle de domaine)
  Point operator[] ( Point a ) const 
    { return leVecteur[a]; };
  // Opration sur les points (avec contrle de domaine)
  Point operator () ( Point a ) const 
    { return ( a<0 || a>=leVecteur.size() ? a : leVecteur[a] ); };
  // Accs au vecteur qui reprsente la permutation
  const vector<Point>& vecteur( void ) const
    { return leVecteur; };
  
public:
  // Constructeurs divers 
  Permut() 
    {};
  Permut( const Permut& p ) 
    { leVecteur= p.leVecteur; };
  Permut( Point a, Point b ) 
    { setTransposition(a,b); };
  Permut( const Cycle& c ) 
    { setCycle(c); };
  Permut( const Cycles& c ) 
    { setCycles(c); };
  Permut( const char* s ) 
    { isstream in(s); in >> *this; };
  Permut( const string& s ) 
    { isstream in(s.c_str()); in >> *this; };
  
  // Reconnatre l'identit et dterminer le degr 
  bool isIdentity( void ) const 
    { return ( leVecteur.size() == 0 ); };
  Point degre( void ) const 
    { return leVecteur.size(); };
  
  // Oprateurs d'affectation
  Permut& operator= ( const Permut& p )
    { leVecteur= p.leVecteur; return *this; };
  Permut& operator= ( const Cycle& c )
    { setCycle( c ); return *this; };
  Permut& operator= ( const Cycles& c )
    { setCycles( c ); return *this; };
  Permut& operator= ( const char* s )
    { isstream in(s); in >> *this; return *this; };
  Permut& operator= ( const string& s ) 
    { isstream in(s.c_str()); in >> *this; return *this; };
  void swap( Permut& p )
    { leVecteur.swap( p.leVecteur ); };

  // Constructions diverses
  void setIdentity( void ) { leVecteur.clear(); };
  void setTransposition( Point a, Point b );
  void setReversion( Point degre );
  void setRandom( Point degre );
  void setCycle( const Cycle& c );
  void setCycles( const Cycles& c );
  void getCycles( Cycles& cycles ) const;
 
  // Loi de groupe: inversion et multiplication
  friend Permut operator ! ( const Permut& p );
  friend Permut operator * ( const Permut& p, const Permut& q );

  // Oprateurs d'entre-sortie
  friend ostream& operator << ( ostream& out, const Permut& p );
  friend istream& operator >> ( istream& in, Permut& p );

  // Comparaison de deux permutations (par rapport  un ordre choisi)
  friend int compare( const Permut& p, const Permut& q );
  bool operator == (const Permut& p) const { return compare(*this,p)==0; };
  bool operator != (const Permut& p) const { return compare(*this,p)!=0; };
  bool operator <  (const Permut& p) const { return compare(*this,p)< 0; };
  bool operator <= (const Permut& p) const { return compare(*this,p)<=0; };
  bool operator >  (const Permut& p) const { return compare(*this,p)> 0; };
  bool operator >= (const Permut& p) const { return compare(*this,p)>=0; };
  
  // Dterminer le successeur par rapport  l'ordre deglex
  friend Permut& operator ++ ( Permut& p );
};

/*-----------------------------------------------------------------------*\
 * Raccourcir en supprimant d'ventuels points fixes  la fin
 *-----------------------------------------------------------------------
 * Par convention, toute permutation est stocke dans un tableau de 
 * longueur minimale, c'est--dire tous les points fixes qui tranent 
 *  la fin sont supprims. Si une fonction risque de produire un rsultat
 * non conforme  cette convention, il faut rectifier la situation 
 * en sortant de la fonction, par exemple en appelant  raccourcir() .
\*-----------------------------------------------------------------------*/
void Permut::raccourcir( void )
{
  Point p= leVecteur.size()-1;
  while ( p >= 0 && leVecteur[p] == p ) --p;
  leVecteur.resize(p+1);
};

/*-----------------------------------------------------------------------*\
 * Rallonger en ajoutant des points fixes  la fin
 *-----------------------------------------------------------------------
 * Si jamais il y en a besoin, la fonction  rallonger()  agrandit 
 * le tableau et le remplit par des points fixes.
\*-----------------------------------------------------------------------*/
void Permut::rallonger( Point deg )
{
  if ( deg <= leVecteur.size() ) return;
  Point p= leVecteur.size();
  leVecteur.resize( deg );
  for( ; p<deg; ++p ) leVecteur[p]= p;
};

/*-----------------------------------------------------------------------*\
 * Construction d'une transposition (a,b)
\*-----------------------------------------------------------------------*/
void Permut::setTransposition( Point a, Point b )
{ 
  if ( a > b ) std::swap( a, b );
  if ( a < 0 || a == b ) { setIdentity(); return; };
  leVecteur.resize(b+1);
  for( Point i=0; i<=b; ++i ) leVecteur[i]= i;    
  leVecteur[a]= b;
  leVecteur[b]= a;
};

/*-----------------------------------------------------------------------*\
 * Construction de la reversion [n-1,n-2,...,1,0] pour un n donn.
\*-----------------------------------------------------------------------*/
void Permut::setReversion( Point deg )
{ 
  if ( deg < 2 ) { setIdentity(); return; };
  leVecteur.resize( deg );
  for( Point i=0; i<deg; ++i ) leVecteur[i]= deg-1-i;
};

/*-----------------------------------------------------------------------*\
 * Engendrer un nombre (pseudo-)alatoire dans l'intervalle 0,1,...,max
\*-----------------------------------------------------------------------*/
Permut::Point random( Permut::Point max ) 
{ 
  return ( std::random() % (max+1) ); 
};

/*-----------------------------------------------------------------------*\
 * Construire une permutation alatoire
 *-----------------------------------------------------------------------
 * Pour explication: toute permutation de {0,...,n-1} s'crit de manire
 * unique comme produit de transpositions (n-1,b[n-1])...(2,b[2])(1,b[1]) 
 * avec 0 <= b[a] <= a pour tout a = 1,2,...,n-1. (Dans le cas b[a]=a 
 * on interprte (a,b[a]) comme tant l'identit.) La fonction suivante
 * choisit les b[a] de manire alatoire et en construit la permutation 
 * alatoire associe. On obtient ainsi une distribution quiprobable 
 * sur Sym(n) (toujours pourvu que random(k) produise des nombres 
 * alatoires indpendants et quiprobables sur {0,...,k}).
\*-----------------------------------------------------------------------*/
void Permut::setRandom( Point degre )
{ 
  if ( degre < 2 ) { setIdentity(); return; };
  leVecteur.resize( degre );
  leVecteur[0]= 0;
  for ( Point a=1; a<degre; ++a ) 
    { 
      Point b= random(a);
      leVecteur[a]= leVecteur[b];
      leVecteur[b]= a;
    };
  raccourcir();
};

/*-----------------------------------------------------------------------*\
 * Construction d'une permutation  partir d'un cycle (a,b,c,...)
 * Ici on suppose que les points a,b,c,... sont tous distincts ; 
 * la permutation ainsi cre envoie a sur b, puis b sur c, etc.
 * Si la construction choue, on renvoie l'identit.
\*-----------------------------------------------------------------------*/
void Permut::setCycle( const Cycle& c )
{ 
  setIdentity();
  if ( c.size() < 2 ) return;

  // Trouver le maximum de la liste (et vrifier que le minimum est >= 0)
  Point max= -1;
  for ( Point a=0; a<c.size(); ++a )
    if ( c[a] > max ) max= c[a];
    else if ( c[a] < 0 ) return;
  
  // Dfinir une permutation, de degr convenable,  partir du cycle.
  // On vrifie en mme temps qu'il s'agit vraiement d'un cycle, 
  // c'est--dire aucun lment y figure deux fois.
  leVecteur.resize( max+1, -1 );
  Point b= c.front();
  for( int i=c.size()-1; i>=0; --i )
    {
      Point a= c[i];
      if ( leVecteur[a] >= 0 ) 
	{ setIdentity(); return; }; // erreur
      leVecteur[a]= b;
      b= a;
    };

  // Tous les autres points non mentionns sont fixs
  for ( Point a=0; a<=max; ++a )
    if ( leVecteur[a] < 0 ) leVecteur[a]= a;
};

/*-----------------------------------------------------------------------*\
 * Construction d'une permutation comme produit de cycles
\*-----------------------------------------------------------------------*/
void Permut::setCycles( const Cycles& c )
{
  setIdentity();
  for( int i=0; i<c.size(); ++i )
    *this = *this * Permut( c[i] );
};

/*-----------------------------------------------------------------------*\
 * Dcomposer une permutation en produit de cycles disjoints
\*-----------------------------------------------------------------------*/
void Permut::getCycles( Cycles& cycles ) const
{ 
  // Le vecteur suivant indique les points dj visits
  vector<bool> deja_visite( degre(), false );
  
  // Visiter tous les points et parcourir chaque cycle exactement une fois
  cycles.clear();
  for ( Point i=0; i<degre(); ++i )
    {
      // On ne tient compte que des cycles de longueur > 1
      // et on vite bien sr les cycles dj parcourus.
      if ( leVecteur[i]!=i && !deja_visite[i] )
	{
	  // Parcourir ce cycle en visitant ses points un par un
	  Cycle leCycle;
	  leCycle.push_back( i );
	  deja_visite[i]= true;
	  for( Point j= leVecteur[i]; !deja_visite[j]; j= leVecteur[j] )
	    {
	      leCycle.push_back( j );
	      deja_visite[j]= true;
	    };
	  // Ajouter ce cycle  la liste des cycles dj construite
	  cycles.push_back( leCycle );
	};
    };

  // Commes les cycles disjoints commutent, la dcomposition n'est unique
  // qu' l'ordre prs. On prfre ici les trier en longueurs croissantes.
  sort( cycles.begin(), cycles.end() );
};

/*-----------------------------------------------------------------------*\
 * La comparaison suivante permet de trier les cycles par longueur
\*-----------------------------------------------------------------------*/
bool operator < ( const Permut::Cycle& c1, const Permut::Cycle& c2 )
{
  if ( c1.size() < c2.size() ) return true;
  if ( c1.size() > c2.size() ) return false;
  for( int i=0; i<c1.size(); ++i )
    {
      if ( c1[i] < c2[i] ) return true;
      if ( c1[i] > c2[i] ) return false;
    }
  return false;
}     

/*-----------------------------------------------------------------------*\
 * Calculer l'inverse d'une permutation
\*-----------------------------------------------------------------------*/
Permut operator ! ( const Permut& p )
{
  Permut inverse;
  inverse.leVecteur.resize( p.degre() );
  for( Permut::Point a=p.degre()-1; a>=0; --a ) 
    inverse[ p[a] ]= a;
  return inverse;
};

/*-----------------------------------------------------------------------*\
 * Calculer le produit de deux permutations
\*-----------------------------------------------------------------------*/
Permut operator * ( const Permut& p, const Permut& q )
{ 
  // Dterminer le dernier point qui soit boug par le produit p*q
  Permut::Point a= max( p.degre(), q.degre() ) - 1;
  while( a>=0 && p(q(a)) == a ) --a;

  // Calculer le produit p*q (dont le degr vient d'tre dtermin)
  Permut produit;
  produit.leVecteur.resize(a+1);
  for ( ; a>=0; --a )
    produit[a]= p(q(a));
  return produit;
};

/*-----------------------------------------------------------------------*\
 * Calculer la signature d'une permutation
 *-----------------------------------------------------------------------
 * La signature est l'unique homomorphisme de Sym(X) sur {+1,-1} qui
 * envoie chaque transposition sur -1. Un cycle de longueur k s'crit
 * comme produit de k-1 transpositions, sa signature est donc (-1)^(k-1).
 * Ceci permet de dterminer efficacement la signature d'une permutation
 *  partir de sa dcomposition en cycles disjoints : on parcourt chaque 
 * cycle exactement une fois, et dtermine sa signature en mme temps.
\*-----------------------------------------------------------------------*/
int signature( const Permut& p )
{ 
  int sign= +1;
  vector<bool> deja_visite( p.degre(), false );
  for ( Permut::Point i=0; i<p.degre(); ++i )
    if ( p(i)!=i && !deja_visite[i] )
      {
	deja_visite[i]= true;
	for ( Permut::Point j= p(i); !deja_visite[j]; j= p(j) )
	  {
	    deja_visite[j]= true;
	    sign= -sign;
	  };
      };
  return sign;
};

/*-----------------------------------------------------------------------*\
 * Compter le nombre d'inversions d'une permutation
 *-----------------------------------------------------------------------
 * Une inversion d'une permutation p est une paire de points (a,b) 
 * telle que a < b mais p(a) > p(b). Le comptage des inversions de p
 * donne une mthode (moins efficace) de dterminer la signature, car  
 * sign(p) = (-1)^inv  o inv est le nombre d'inversions.
 *-----------------------------------------------------------------------
 * Remarque: toute permutation s'crit comme un produit des transpositions
 * (0,1), (1,2), (2,3), ..., (n-2,n-1); c'est donc une famille gnratrice.
 * Cette criture est loin d'tre unique: tant donn une permutation p, 
 * il existe une infinit de telles critures de plus en plus longues !
 * On appelle *longueur* de p la longueur l minimale d'un tel produit
 * p = t_1 * t_2 * ... * t_l avec des transpositions t_k = (i_k,i_k+1).
 * Comment dterminer efficacement cette longueur minimale ? 
 * C'est le nombre d'inversions de p !
\*-----------------------------------------------------------------------*/
Int inversions( const Permut& p )
{ 
  Int inv= 0;
  for ( Permut::Point a=0; a<p.degre(); ++a )
    for ( Permut::Point b=a+1; b<p.degre(); ++b )
      if ( p(a) > p(b) ) ++inv;
  return inv;
};

/*-----------------------------------------------------------------------*\
 * Dterminer l'ordre d'une permutation
 *-----------------------------------------------------------------------
 * L'ordre d'une permutation p est le plus petit k>0 tel que p^k = id.
 * Pour le calcul, la manire la plus efficace passe par la dcomposition
 * en cycles disjoints : un cycle de longueur k est d'ordre k, et l'ordre
 * d'un produit  supports disjoints est le ppcm des ordres des facteurs.
\*-----------------------------------------------------------------------*/
Int pgcd( Int a, Int b )
{ 
  // pgcd selon l'algorithme d'Euclid
  while ( b != 0 ) { Int r= a%b; a= b; b= r; };
  return a;
};

Int ppcm( Int a, Int b )
{ 
  // ppcm selon la formule  ab = pgcd(a,b)ppcm(a,b)
  return a * ( b / pgcd(a,b) );
};

Int ordre( const Permut& p )
{
  Int ordre= 1;
  vector<bool> deja_visite( p.degre(), false );
  for ( Permut::Point i=0; i<p.degre(); ++i )
    if ( p(i)!=i && !deja_visite[i] )
      {
	deja_visite[i]= true;
	Int longueur= 1;
	for ( Permut::Point j= p(i); !deja_visite[j]; j= p(j) )
	  {
	    deja_visite[j]= true;
	    ++longueur;
	  };
	ordre= ppcm( ordre, longueur );
      };
  return ordre;
}

/*-----------------------------------------------------------------------*\
 * Oprateur de sortie auxiliaire : afficher un conteneur 
 * (par exemple un vecteur, une liste, un ensemble, etc.)
\*-----------------------------------------------------------------------*/
template <typename C>
void write( ostream& out, const C& container, const string& sep= "(,)" )
{
  out << sep[0];
  if ( !container.empty() )
    {
      typename C::const_iterator it= container.begin();
      out << *it;
      for ( ++it; it != container.end(); ++it )
	out << sep[1] << *it;
    };
  out << sep[2];
};

template <typename T>
ostream& operator << ( ostream& out, const vector<T>& vec )
{
  write( out, vec );
  return out;
};

/*-----------------------------------------------------------------------*\
 * Oprateur de sortie pour les permutations
 *-----------------------------------------------------------------------
 * Toute permutation s'crit comme produit de cycles disjoints, et ceci 
 * est la notation utilise ici. Pour viter toute ambigut, on met ce
 * produit en parenthse, comme ((0,1)(2,3)) ou ((0,1)(2,3)(4,5,6)).
\*-----------------------------------------------------------------------*/
ostream& operator << ( ostream& out, const Permut& p )
{ 
  Permut::Cycles lesCycles;
  p.getCycles( lesCycles );
  out << "(";
  for( int i=0; i<lesCycles.size(); ++i )
    {
      Permut::Cycle& leCycle= lesCycles[i];
      out << "(" << leCycle[0];
      for( int j=1; j<leCycle.size(); ++j )
	out << "," << leCycle[j];
      out << ")";
    };
  out << ")";
  return out;
};

/*-----------------------------------------------------------------------*\
 * Oprateur d'entre auxiliaire :
 * Lire un vecteur de type vector<T> sous le format (a,b,c,...)
\*-----------------------------------------------------------------------*/
template <typename T>
void read( istream& in, vector<T>& vec, const string& sep= "(,)" )
{
  // remettre le flot  l'tat normal (si possible)
  in.clear();
  if ( in.bad() ) return;
  
  // lire la parenthse ouvrante (erreur sinon)
  char symbole;
  in >> symbole;
  if ( in.fail() ) return;
  if ( symbole != sep[0] )
    { 
      in.putback( symbole ); 
      in.setstate( ios::failbit ); 
      return; 
    };

  // lire les lments un par un, jusqu' la parenthse fermante
  vec.clear();
  T element;
  for( in >> symbole; symbole!=sep[2]; in >> symbole )
    {
      if ( in.fail() ) return;
      if ( symbole!=sep[1] )
	in.putback( symbole );
      in >> element;
      if ( in.fail() ) return;
      vec.push_back( element );
    };
};

template <typename T>
istream& operator >> ( istream& in, vector<T>& vec )
{
  read( in, vec );
  return in;
};

/*-----------------------------------------------------------------------*\
 * Oprateur d'entre pour les permutations
 *-----------------------------------------------------------------------
 * L'entre s'attend au format dcrit pour la sortie, c'est--dire 
 * un produit de cycles entre deux paranthses. Pour plus de confort 
 * on n'exige pas que ce produit soit disjoint. Si la lecture choue, 
 * par exemple  cause d'une erreur syntactique, on renvoie l'identit.
\*-----------------------------------------------------------------------*/
istream& operator >> ( istream& in, Permut& p )
{ 
  // prparer le flot d'entre et effacer la permutation antrieure
  in.clear();
  p.setIdentity();

  // lire la liste des cycles (en utilisant l'entre de vecteurs)
  Permut::Cycles cycles;
  read( in, cycles, "( )" );
  if ( in.fail() ) return in;

  // traduire la liste des cycles en permutation
  for( int i=0; i<cycles.size(); ++i )
    p = p * Permut( cycles[i] );

  // renvoyer le flot comme il se doit
  return in;
};

/*-----------------------------------------------------------------------*\
 * Comparer deux permutations (par rapport  l'ordre revlex)
 *-----------------------------------------------------------------------
 * L'ordre revlex sur les permutations est dfini comme suit : on pose 
 * p < q s'il existe k tel que p(k) > q(k) et p(l) = q(l) pour tout l>k.
 * C'est un ordre linaire, quivalent  celui des nombres naturels. 
 * Le plus petit lment et l'identit, le plus grand lment du groupe 
 * symtrique Sym(n) est la reversion [n-1,n-2,...,1,0].
\*-----------------------------------------------------------------------*/
/*
int compare( const Permut& p, const Permut& q )
{ 
  if ( p.degre() < q.degre() ) return -1;      // p < q
  if ( p.degre() > q.degre() ) return +1;      // p > q
  for ( Permut::Point i=p.degre()-1; i>0; --i )
    {
      if ( p[i] > q[i] ) return -1;            // p < q
      if ( p[i] < q[i] ) return +1;            // p > q
    };
  return 0;                                    // p == q
};
*/

/*-----------------------------------------------------------------------*\
 * Comparer deux permutations (par rapport  l'ordre deglex)
 *-----------------------------------------------------------------------
 * Les permutations sont ordonnes d'abord par leur degr, puis par 
 * ordre lexicographique. C'est un ordre linaire, quivalent  celui
 * des nombres naturels. Le plus petit lment et l'identit, le plus 
 * grand lment du groupe Sym(n) est la reversion [n-1,n-2,...,1,0].
\*-----------------------------------------------------------------------*/
int compare( const Permut& p, const Permut& q )
{ 
  if ( p.degre() < q.degre() ) return -1;      // p < q
  if ( p.degre() > q.degre() ) return +1;      // p > q
  for ( Permut::Point i=0; i<p.degre(); ++i )
    {
      if ( p[i] < q[i] ) return -1;            // p < q
      if ( p[i] > q[i] ) return +1;            // p > q
    };
  return 0;                                    // p == q
};

/*-----------------------------------------------------------------------*\
 * Dterminer le successeur (par rapport  l'ordre deglex)
 *-----------------------------------------------------------------------
 * Par exemple, le successeur de [0,4,2,5,3,1] est [0,4,3,1,2,5]. 
 * Effectivement, on ne peut pas augmenter 1 ni 3 ni 5 sans perturber 
 * les nombres prcdents. Le nombre  augmenter ici est donc le 2.
 * L'augmentation minimum est ralise en remplacant 2 par 3, puis
 * en rangeant la queue des nombres restants 1,2,5 dans l'ordre croissant.
\*-----------------------------------------------------------------------*/
Permut& operator ++ ( Permut& p )
{
  // le successeur de l'identit est la transposition (0,1)
  if ( p.isIdentity() )
    {
      p.setTransposition(0,1);
      return p;
    };
  
  // dterminer la position a du nombre  augmenter
  Permut::Point deg= p.degre(), a= deg-2;
  while( a>=0 && p[a]>p[a+1] ) --a;

  // cas particulier : le successeur de l'inversion [n-1,n-2,...,1,0]
  if ( a == -1 )
    {
      p.leVecteur.resize( deg+1 );
      for( a=0; a<deg-1; ++a ) p[a]= a;
      p[deg]= deg-1;
      p[deg-1]= deg;
      return p;
    };

  // dterminer la position b du nombre  remplacer
  Permut::Point b= a+1;
  while( b<deg-1 && p[b+1]>p[a] ) ++b;

  // changer p[a] et p[b], puis inverser la queue
  swap( p[a], p[b] );
  for( ++a, b=deg-1; a<b; ++a, --b ) 
    swap( p[a], p[b] );

  // viter le cas d'un point fixe terminal
  if ( p[deg-1] == deg-1 ) 
    swap( p[deg-2], p[deg-1] );
  return p;
};

/*-----------------------------------------------------------------------*\
 * Un exemple d'utilisation ( mettre en commentaire aprs des tests)
\*-----------------------------------------------------------------------*/
/*
int main()
{
  // Conversion d'une chane de caractres en permutation
  string s("((0,1,2)(2,3,4))");
  cout << "\nBonjour et bienvenue aux permutations !\n"
       << "Vous pouvez entrez des permutations comme produits de cycles,\n"
       << "par exemple la chane de caractres \"" << s << "\"\n"
       << "reprsente la permutation " << Permut(s) << "." << endl;

  cout << "Voici les premires permutations dans l'ordre deglex :" << endl;
  Permut p,q;
  int k=0;
  q.setReversion(2);
  for( ; p<=q; ++p, ++k ) 
    cout << p << " ";
  cout << endl << "nombre total : " << k << endl;
  q.setReversion(3);
  for( ; p<=q; ++p, ++k ) 
    cout << p << " ";
  cout << endl << "nombre total : " << k << endl;
  q.setReversion(4);
  for( ; p<=q; ++p, ++k ) 
    cout << p << " ";
  cout << endl << "nombre total : " << k << endl;
  q.setReversion(5);
  for( ; p<=q; ++p, ++k ) 
    cout << p << " ";
  cout << endl << "nombre total : " << k << endl;

  Permut c;
  srandom(12345);
  cout << "Voici quelques permutations alatoires : " << endl;
  for ( int i=0; i<5; ++i )
    {
      c.setRandom(20);
      cout << c << " est d'ordre " << ordre(c) << endl;
    };

  // Un exemple typique d'entre-sortie
  cout << "Entrez deux permutations svp : " << endl;
  cin >> p >> q;
  cout << "Vous avez entr p=" << p << " et q=" << q << endl
       << "Leurs inverses valent " << !p << " et " << !q << endl
       << "Leurs produits valent pq=" << p*q << " et qp=" << q*p << endl;
  cout << "Au revoir et  bientt." << endl;
};
/**/
