/*=======================================================================*\
 * voyageur.cc : une recherche "force brute" d'un plus court chemin
 * Copyright (C) 2005 by Michael Eisermann, all rights reserved
 * Institut Fourier, Universit Joseph Fourier, Grenoble, France
 *-----------------------------------------------------------------------
 * Nous implmentons ici la recherche exhaustive d'un plus court chemin
 * qui joint n villes puis retourne  la ville du dpart. C'est le fameux
 * problme du voyageur de commerce, exemple d'un problme trs complexe.
 * Notre implmentation parcourt une boucle de longueur n!, ce qui n'est
 * praticable pour n trs petit. (10!=3628800, 15!>10^12, 20!>10^18)
\*=======================================================================*/

#include <iomanip>
#include <fstream>
#include <cmath>
#include "permut.cc"

/*-----------------------------------------------------------------------*\
 * Les coordonnes des n villes  visiter sont lu d'un fichier puis 
 * stock dans un vecteur global sous la forme (x1,y1,x2,y2,...)
\*-----------------------------------------------------------------------*/
const char filename[]= "villes.lst";
int nombre_de_villes;
vector<double> coordonnees;

/*-----------------------------------------------------------------------*\
 * La fonction dist(i,j) calcule la distance entre deux villes i,j.
 *  noter qu'en interne nous numrotons les villes par 0,1,...,n-1.
\*-----------------------------------------------------------------------*/
double carre( double r )
{ return r*r; }

double dist( int i, int j )
{
  return sqrt( carre( coordonnees[2*i]   - coordonnees[2*j]   ) + 
	       carre( coordonnees[2*i+1] - coordonnees[2*j+1] ) );
}

/*-----------------------------------------------------------------------*\
 * En voyage est cod par une permutation p des points {0,1,...,n-2}, 
 * ce qui correspond   n-1 -> p(0) -> p(1) -> ... -> p(n-2) -> n-1.
 * Autrement dit, on prend la ville n-1 comme point de dpart, puis visite 
 * toutes les autres villes dans l'ordre prscrite par p, et rentre  n-1.
 * La distance parcourue par ce voyage est la somme des distances entre 
 * deux villes conscutives. (Ne pas oublier de fermer le cycle!)
\*-----------------------------------------------------------------------*/
void affiche_voyage( const Permut& p )
{
  int& n= nombre_de_villes;
  cout << n << " -> ";
  for( int i=0; i<n-1; ++i )
    cout << p(i)+1 << " -> ";
  cout << n;
}

double dist( const Permut& p )
{
  int& n= nombre_de_villes;
  double d= dist( n-1, p(0) );
  for( int i=0; i<n-2; ++i )
    d+= dist( p(i), p(i+1) );
  d+= dist( p(n-2), n-1 );
  return d;
}

/*-----------------------------------------------------------------------*\
 * La fonction principale lit les coordonnes du fichier spcifi,
 * puis cherche un plus court chemin par la mthode "force brute".
\*-----------------------------------------------------------------------*/
int main()
{
  cout << "\nBonjour et bienvenue au problme du voyageur de commerce." << endl;

  // essayer d'ouvrir le fichier  la lecture
  ifstream in( filename );
  if ( !in ) 
    {
      cout << "Je ne peux pas lire le fichier \"" << filename <<"\"." << endl
	   << "Veuillez rendre les donnes accessibles  cet endroit." << endl;
      exit(0);
    }
  
  // lire les coordonnes.
  read( in, coordonnees, "(,)" );
  if( coordonnees.size() % 2 == 1 ) coordonnees.pop_back();
  nombre_de_villes= coordonnees.size() / 2;
  
  // viter quelques cas triviaux
  cout << "J'ai lu les coordonneesonnes de " 
       << nombre_de_villes << " villes :" << endl;
  if ( nombre_de_villes <= 3 ) 
    {
      cout << "Ici le problme du voyageur est trivial." << endl;
      exit(0);
    };

  // afficher les coordonnes des villes
  cout.setf( ios::fixed );
  cout.precision( 5 );
  for( int i=0; i<nombre_de_villes; ++i )
    cout << "( " << setw(10) << coordonnees[2*i] 
	 << ", " << setw(10) << coordonnees[2*i+1] << " )" << endl;

  // parcourir tous les voyages possibles
  Permut pmin, plast;
  plast.setReversion( nombre_de_villes-1 );
  double dmin= dist( pmin );
  for( Permut p; p<=plast; ++p )
    {
      double d= dist( p );
      affiche_voyage( p );
      cout << " : longueur=" << d << endl;
      if ( d < dmin ) { dmin= d; pmin= p; };
    };
  
  // afficher le rsultat : un plus court voyage 
  cout << "Voyage de longueur minimale : " << endl;
  affiche_voyage( pmin );
  cout << " : longueur=" << dmin << endl;
  cout << "Au revoir et bonne route.\n" << endl;  
}
