//=======================================================================
// eval-postfixe.cc : evaluation d'une expressions en notation postfixe
//
// Description :
// La fonction eval_postfixe permet d'valuer un entier donn par une
// expression en notation postfixe.  Elle lit dans un flot d'entre
// (le clavier par exemple) et elle crit un protocole dans un flot de
// sortie (l'cran par exemple).
//
// Utilisation :
// Deux modes sont possibles : ou bien on entre un seul nombre, ou bien 
// on entre toute une liste d'entiers spars d'espaces et dlimite 
// d'une parenthse ouvrante '(' et d'une parenthse fermante ')'.
// Dans le deuxime cas c'est la dernire valeur qui est renvoye.
//
//   Exemples : l'entre "123" ou "( 123 )" produit la valeur 123, 
//   alors que l'entre "( 123 456 678 )" produit la valeur 678.
//
// L'entre d'une liste offre une fonctionnalit encore plus ample : 
// - Le passage  la ligne (return) fait afficher la liste dj entre.
// - La commande 'del' efface le dernier entier (c--d il le dpile).
// - Un oprateur arithmtique (comme +,-,*,/ etc) dpile ses arguments
//   et empile son rsultat. On obtient ainsi une sorte de calculette.
// - La commande '=a' affecte le dernier lment  la variable nomme a.
//   La commande 'a' empile la valeur de a. (Sont disponibles a...z.)
// Ceci permet de commodment entrer ou calculer des grands entiers.
//
//   Exemple : l'entre "( 123 456 789 del + <return>" affiche "( 579 " 
//   de sorte que l'on puisse contrler le rsultat intermdiaire.
//=======================================================================

#include <iostream>        // dclarer l'entre-sortie standard
#include <sstream>         // conversion entre string et flot
#include <string>          // dclarer la classe string 
#include <vector>          // dfinir la classe vector de la STL
#include <cctype>          // dclarer isspace() et isdigit() etc
#include "integer.cc"      // arithmtique des nombres entiers
using namespace std;       // accs direct aux fonctions standard

//-----------------------------------------------------------------------
// Tester si une chane de caractres reprsente un nombre entier.
// (Il nous suffit ici de tester si la chane commence par un chiffre 
// ou par le signe moins suivi d'un chiffre.)
//-----------------------------------------------------------------------
bool est_un_entier( const string& mot )
{
  if ( mot.size()>=1 && isdigit(mot[0]) ) return true;
  if ( mot.size()>=2 && mot[0]=='-' && isdigit(mot[1]) ) return true;
  return false;
}

//-----------------------------------------------------------------------
// Retirer puis effacer le dernier lment d'une pile.
// Si la pile est vide on renvoie la valeur 0 par dfaut.
//-----------------------------------------------------------------------
Integer depiler( vector<Integer>& pile )
{
  if ( pile.empty() ) return 0;
  Integer a= pile.back();
  pile.pop_back();
  return a;
}

//-----------------------------------------------------------------------
// Version calculette (interactive) de la lecture d'un entier
//-----------------------------------------------------------------------
Integer eval_postfixe( istream& in, ostream& out )
{
  // Initialiser le flot et la pile
  in.clear();
  vector<Integer> pile;

  // Lire le premier mot
  string mot;
  in >> mot;
  if ( in.fail() || mot.empty() ) 
    {
      in.setstate( ios::failbit ); 
      return Integer(0);
    }
  
  // Traiter d'abord le cas d'un seul nombre entier
  if ( est_un_entier(mot) ) return Integer(mot);

  // On s'attend maintenant  une liste commenant par une parenthse
  if ( mot != "(" )
    {
      in.setstate( ios::failbit ); 
      return Integer(0);
    }

  // Initialiser la liste des variables (nommes 'a' ... 'z')
  vector<Integer> variable(26,Integer(0));
      
  for(;;)
    {
      // Sauter les espaces mais intercepter un retour  la ligne
      char symbole;
      for( in.get(symbole); isspace(symbole); in.get(symbole) )
	{
	  if ( in.fail() ) return depiler(pile);
	  if ( symbole=='\n' ) 
	    {
	      out << "( ";
	      for( size_t i=0; i<pile.size(); ++i )
		out << pile[i] << " ";
	    }
	}

      // Lire le prochain mot
      in.putback(symbole);
      in >> mot;
      if ( in.fail() || mot.empty() )
	{
	  in.setstate( ios::failbit ); 
	  return depiler(pile);
	}

      // Terminer la lecture par une parenthse fermante
      if ( mot == ")" ) return depiler(pile);

      // Tester si le mot commence comme un nombre, puis valuer l'entier
      if ( est_un_entier(mot) )
	{ 
	  pile.push_back( Integer(mot) ); 
	  continue; 
	}
      
      // Tester si le mot est une seule lettre, puis valuer la variable 
      if ( mot.size()==1 && islower(mot[0]) )
	{ 
	  pile.push_back( variable[mot[0]-'a'] ); 
	  continue; 
	}
      
      // Tester si le mot commence par '=', puis affecter  la variable      
      if ( mot.size()==2 && mot[0]=='=' && islower(mot[1]) )
	{
	  int index= mot[1]-'a';
	  if ( pile.empty() ) variable[index]= 0;
	  else variable[index]= pile.back();
	  continue;
	}
      
      // Effacer le dernier lment sur la pile
      if ( mot == "del" )
	{ 
	  depiler(pile); 
	  continue; 
	}
      
      if ( mot == "!" )         // factorielle : un seul argument
	{
	  Integer a= depiler(pile);
	  pile.push_back( factorielle(a) );
	  continue;
	};
      
      if ( mot == "?" )         // afficher l'aide en ligne
	{
	  out << "Tapez '(' pour commencer, ')' pour terminer l'entre,\n"
	      << "<return> pour afficher la pile, 'del' pour dpiler,\n"
	      << "'=a' pour affecter, 'a' pour lire une variable (a...z),\n"
	      << "'?!' pour plus de dtails." << endl;
	  continue;
	}
      
      if ( mot == "?!" )        // afficher l'aide en ligne
	{
	  out << "Tapez '(' pour commencer, ')' pour terminer l'entre,\n"
	      << "<return> pour afficher la pile, 'del' pour dpiler,\n"
	      << "'=a' pour affecter, 'a' pour lire une variable (a...z).\n"
	      << "Voici les fonctions disponibles:\n" 
	      << "arguments : fonction : commentaire\n"
	      << "      a b   +          addition a+b\n"
	      << "      a b   -          soustraction a-b\n"
	      << "      a b   *          multiplication a*b\n"
	      << "        n   !          factorielle a!\n"
	      << "      a b   /          division de a par b\n"
	      << "      a b   %          reste de a modulo b\n"
	      << "      x e   ^          puissance x^e\n"
	      << "    x e m   ^%         puissance x^e modulo m"
	      << endl;
	  continue;
	}
      
      // Si l'on arrive ici, c'est  cause d'une fonction non reconnue.
      out << "fonction \"" << mot << "\" non reconnue." << endl;
    }
}

//-----------------------------------------------------------------------
// Version muette (non interactive) de la lecture d'un entier
// La sortie n'est pas affiche mais dvie dans un string
//-----------------------------------------------------------------------
Integer eval_postfixe( istream& in )
{
  ostringstream muet;
  return eval_postfixe( in, muet );
}

