//=======================================================================
// anneau.cc : modle abstrait pour implmenter un anneau effectif
// Copyright (C) 2003-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Le modle suivant explicite les lments d'une implmentation minimale
// requise pour un anneau. Ces dclarations ne servent que de modle : 
// aucune implmentation n'est fournie dans ce fichier-ci. Pour une 
// implmentation concrte, il faut remplacer le nom Anneau par le nom de 
// la classe  implmenter, puis dfinir chacune des fonctions ci-dessous.
//=======================================================================

#include <iostream>
using namespace std;

//-----------------------------------------------------------------------
// La classe Anneau (constructeurs et oprations lmentaires)
//-----------------------------------------------------------------------
class Anneau
{
public :
  Anneau( int n=0 );                             // constructeur/conversion
  Anneau( const Anneau& source );                // constructeur par copie
  Anneau& operator=  ( int n );                  // affectation/conversion
  Anneau& operator=  ( const Anneau& a );        // affectation par copie
  bool    operator== ( const Anneau& a ) const;  // test d'galit
  bool    operator!= ( const Anneau& a ) const;  // test de diffrence
  Anneau  operator+  ( const Anneau& a ) const;  // addition
  Anneau& operator+= ( const Anneau& a );        // addition en place
  Anneau  operator-  () const;                   // oppos 
  Anneau  operator-  ( const Anneau& a ) const;  // soustraction
  Anneau& operator-= ( const Anneau& a );        // soustraction en place
  Anneau  operator*  ( const Anneau& a ) const;  // multiplication
  Anneau& operator*= ( const Anneau& a );        // multiplication en place
};

//-----------------------------------------------------------------------
// Oprateurs d'entre-sortie pour la classe Anneau
//-----------------------------------------------------------------------
ostream& operator<< ( ostream& out, const Anneau& a );
istream& operator>> ( istream& in, Anneau& a );

//-----------------------------------------------------------------------
// Fonctions de base requises pour travailler avec un anneau
//-----------------------------------------------------------------------
// Reconnatre les lments 0 et 1
bool    zero       ( const Anneau& a );
bool    one        ( const Anneau& a );

// Dterminer si a est inversible, si oui calculer son inverse
bool    inversible ( const Anneau& a );
bool    inversible ( const Anneau& a, Anneau& inverse );

// Dterminer si a est divisible par b, si oui calculer le quotient q
bool    divisible  ( const Anneau& a, const Anneau& b );
bool    divisible  ( const Anneau& a, const Anneau& b, Anneau& q );

// Dterminer si a et b sont associs, si oui calculer u tel que ua=b
bool    associes   ( const Anneau& a, const Anneau& b );
bool    associes   ( const Anneau& a, const Anneau& b, Anneau& u );

// Choisir un reprsentant prfr r=ua parmi les lments associs  a
Anneau  assopref   ( const Anneau& a );
Anneau  assopref   ( const Anneau& a, Anneau& u );

// NB : Dans certains anneaux, comme les entiers ou les polynmes sur 
// un corps il y a un choix canonique. Il est utile de l'implmenter ici.
// Sinon on peut toujours prendre r=a et u=1 par dfaut.

//-----------------------------------------------------------------------
// Fonction supplmentaire requise pour un anneau factoriel :
// On exige le pgcd (mais pas la factorisation, ni un test de primalit).
//-----------------------------------------------------------------------
Anneau  pgcd( const Anneau& a, const Anneau& b );

//-----------------------------------------------------------------------
// Fonction supplmentaire requise pour un anneau principal :
// On exige le pgcd et des coefficients u,v tels que  pgcd(a,b) = ua+vb.
//-----------------------------------------------------------------------
Anneau  pgcd( const Anneau& a, const Anneau& b, Anneau& u, Anneau& v );
Anneau  pgcd( const Anneau& a, const Anneau& b, Anneau& u );

//-----------------------------------------------------------------------
// Fonctions supplmentaires requises pour un anneau euclidien
//-----------------------------------------------------------------------
// Division euclidienne (celle-ci doit vrifier les axiomes usuels par 
// rapport  un stathme euclidien. Ce dernier n'est pas explicit ici.)
void    eudiv( const Anneau& a, const Anneau& b, Anneau& q, Anneau& r );

// Dterminer uniquement le quotient de la division euclidienne
// (Celle-ci n'est dfinie que pour b != 0, erreur sinon.)
Anneau  eudiv( const Anneau& a, const Anneau& b );

// Dterminer uniquement le reste de la division euclidienne
// (Celui-ci est dfini aussi pour b == 0, dans quel cas il vaut a.)
Anneau  eumod( const Anneau& a, const Anneau& b );
