//=======================================================================
// fraction.hh : corps des fractions d'un anneau de Bezout
// Copyright (C) 2003-2006 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// lments requis : 
// Une classe Anneau qui modlise un anneau intgre et qui fournit :
// - addition, soustraction, multiplication comme d'habitude
// - entre-sortie, affectation, comparaison comme d'habitude
// - une fonction pgcd pour calculer le plus grand commun diviseur
//
// lments fournis :
// Une classe Fraction<Anneau> qui modlise le corps des fractions :
// - addition, soustraction, multiplication comme d'habitude
// - entre-sortie, affectation, comparaison comme d'habitude
//=======================================================================

#ifndef __fraction_hh__
#define __fraction_hh__
#include <iostream>
using namespace std;

template <typename Anneau>
class Fraction
{
private:
  Anneau numer, denom;   // numrateur et dnominateur
  void normaliser(void)  // normaliser numer/denom
    { 
      if ( zero(denom) )
	{ cerr << "Erreur : division par zro !" << endl; exit(0); };
      Anneau d= pgcd(numer,denom); 
      divisible(numer,d,numer); divisible(denom,d,denom);
      Anneau u; denom= assopref(denom,u); numer*= u; 
    }; 
  
public:
  // Constructeurs divers
  Fraction( int num=0, int den=1 )
    : numer(num), denom(den) { normaliser(); }
  Fraction( const Anneau& num, const Anneau& den )
    : numer(num), denom(den) { normaliser(); };
  Fraction( const Fraction& a )
    : numer(a.numer), denom(a.denom) {};

  // Affectation
  void affecter( int num=0, int den=1 )
    { numer= num; denom= den; normaliser(); };
  void affecter( const Anneau& num, const Anneau& den )
    { numer= num; denom= den; normaliser(); };
  Fraction& operator= ( const Fraction& a )
    { numer= a.numer; denom= a.denom; return *this; };

  // Lire le numrateur et le dnominateur
  const Anneau& numerateur() const { return numer; };
  const Anneau& denominateur() const { return denom; };

  // Comparaison
  bool operator== ( const Fraction& a ) const
    { return ( numer*a.denom == denom*a.numer ); };
  bool operator!= ( const Fraction& a ) const
    { return ( numer*a.denom != denom*a.numer ); };

  // Addition
  Fraction operator+ ( const Fraction& a ) const
    { return Fraction( numer*a.denom + denom*a.numer, denom*a.denom ); };
  Fraction& operator+= ( const Fraction& a )
    { affecter( numer*a.denom + denom*a.numer, denom*a.denom ); return *this; };

  // Oppos et soustraction
  Fraction operator- () const
    { return Fraction( -numer, denom ); };
  Fraction operator- ( const Fraction& a ) const
    { return Fraction( numer*a.denom - denom*a.numer, denom*a.denom ); };
  Fraction& operator-= ( const Fraction& a )
    { affecter( numer*a.denom - denom*a.numer, denom*a.denom ); return *this; };

  // Multiplication (correcte mais non optimise)
  Fraction operator* ( const Fraction& a ) const
    { return Fraction( numer*a.numer, denom*a.denom ); };
  Fraction& operator*= ( const Fraction& a )
    { affecter( numer*a.numer, denom*a.denom ); return *this; };

  // Division (correcte mais non optimise)
  Fraction operator/ ( const Fraction& a ) const
    { return Fraction( numer*a.denom, denom*a.numer ); };
  Fraction& operator/= ( const Fraction& a )
    { affecter( numer*a.denom, denom*a.numer ); return *this; };
};

//-----------------------------------------------------------------------
// Oprateurs d'entre-sortie
//-----------------------------------------------------------------------
template <typename Anneau>
ostream& operator<< ( ostream& out, const Fraction<Anneau>& a )
{
  if ( one( a.denominateur() ) ) out << a.numerateur();
  else out << a.numerateur() << "/" << a.denominateur();
  return out;
}

template <typename Anneau>
istream& operator>> ( istream& in, Fraction<Anneau>& a )
{
  in.clear();

  // Lire le numrateur
  Anneau num, den;
  in >> num;
  if ( in.fail() ) return in;

  // Lire le symbole sparateur
  char symbole;
  in >> symbole;
  if ( in.fail() ) return in;
  if ( symbole != '/' )
    {
      in.putback(symbole);
      return in;
    };
  
  // Lire le dnominateur
  in >> den;
  if ( in.fail() ) return in;

  // Affecter en cas de succs
  a.affecter( num, den );
  return in;
}

//-----------------------------------------------------------------------
// Fonctions requises pour travailler avec l'anneau Fraction.
// Exercice : vrifier soigneusement la correction des fonctions suivantes.
//-----------------------------------------------------------------------
// Reconnatre l'lment 0
template <typename Anneau>
bool zero( const Fraction<Anneau>& a )
{ 
  return zero(a.numerateur()); 
}

// Reconnatre l'lment 1
template <typename Anneau>
bool one( const Fraction<Anneau>& a )
{ 
  return ( a.numerateur() == a.denominateur() ); 
}

// Tester si a est inversible
template <typename Anneau>
bool inversible( const Fraction<Anneau>& a )
{ 
  return !zero(a);
}

// Tester si a est inversible, si oui calculer son inverse b
template <typename Anneau>
bool inversible( const Fraction<Anneau>& a, Fraction<Anneau>& b )
{ 
  if ( zero(a) ) return false;
  b.affecter( a.denominateur(), a.numerateur() );
  return true;
}

// Tester si a est divisible par b
template <typename Anneau>
bool divisible( const Fraction<Anneau>& a, const Fraction<Anneau>& b )
{ 
  return zero(a) || !zero(b); 
}

// Tester si a est divisible par b, si oui calculer le quotient q
template <typename Anneau>
bool divisible( const Fraction<Anneau>& a, const Fraction<Anneau>& b, Fraction<Anneau>& q )
{
  if ( zero(a) ) { q.affecter(Anneau(0)); return true; };
  if ( !zero(b) ) { q= a/b; return true; };
  return false;
}

// Tester si a et b sont associs
template <typename Anneau>
bool associes( const Fraction<Anneau>& a, const Fraction<Anneau>& b )
{
  return ( zero(a) && zero(b) ) || ( !zero(a) && !zero(b) ); 
}

// Tester si a et b sont associs, si oui calculer u tel que ua=b
template <typename Anneau>
bool associes( const Fraction<Anneau>& a, const Fraction<Anneau>& b, Fraction<Anneau>& u )
{
  if ( zero(a) && zero(b) ) { u=1; return true; };
  if ( !zero(a) && !zero(b) ) { u= b/a; return true; };
  return false;
}

// Choisir un reprsentant prfr parmi les lments associs  a
template <typename Anneau>
Fraction<Anneau> assopref( const Fraction<Anneau>& a )
{
  if ( zero(a) ) return Fraction<Anneau>(0); 
  else return Fraction<Anneau>(1);
}

// Choisir un reprsentant prfr r=ua parmi les lments associs  a
template <typename Anneau>
Fraction<Anneau> assopref( const Fraction<Anneau>& a, Fraction<Anneau>& u )
{
  if ( zero(a) ) { u.affecter(1); return Fraction<Anneau>(0); }
  else { u= 1/a; return Fraction<Anneau>(1); };
}

// Dterminer le pgcd de a et b (trivial dans un corps)
template <typename Anneau>
Fraction<Anneau> pgcd ( const Fraction<Anneau>& a, const Fraction<Anneau>& b )
{
  if ( zero(a) && zero(b) ) return Fraction<Anneau>(0);
  else return Fraction<Anneau>(1);
}

// Dterminer le pgcd et des coefficients u,v tels que  pgcd(a,b) = ua+vb.
template <typename Anneau>
Fraction<Anneau> pgcd ( const Fraction<Anneau>& a, const Fraction<Anneau>& b, 
			Fraction<Anneau>& u, Fraction<Anneau>& v )
{
  if ( !zero(a) ) { u= 1/a; v= 0; return Fraction<Anneau>(1); };
  if ( !zero(b) ) { u= 0; v= 1/b; return Fraction<Anneau>(1); };
  u= 1; v= 1; return Fraction<Anneau>(0);
}

// Division euclidienne (tout corps est aussi un anneau euclidien)
template <typename Anneau>
void eudiv ( const Fraction<Anneau>& a, const Fraction<Anneau>& b, 
	     Fraction<Anneau>& q, Fraction<Anneau>& r )
{
  q= a/b; r=0; return;
}

// Dterminer le quotient de la division euclidienne
template <typename Anneau>
Fraction<Anneau> eudiv ( const Fraction<Anneau>& a, const Fraction<Anneau>& b )
{
  return a/b;
}

// Dterminer le reste de la division euclidienne
template <typename Anneau>
Fraction<Anneau> eumod ( const Fraction<Anneau>& a, const Fraction<Anneau>& b )
{
  if ( zero(b) ) return a;
  else return Fraction<Anneau>(0);
}

#endif
