//=======================================================================
// quotient.hh : classe pour les quotients d'un anneau euclidien
// Copyright (C) 2003-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// lments requis : 
// Une classe Anneau qui modlise un anneau euclidien :
// - addition, soustraction, multiplication comme d'habitude,
// - entre-sortie, affectation, comparaison comme d'habitude,
// - des fonctions eudiv et eumod ralisant une division euclidienne,
// - une fonction pgcd calculant le pgcd et des coefficients de Bzout.
//
// lments fournis :
// Une classe Quotient<Anneau> qui modlise un anneau quotient :
// - addition, soustraction, multiplication comme d'habitude,
// - entre-sortie, affectation, comparaison comme d'habitude.
//=======================================================================

#ifndef __quotient_hh__
#define __quotient_hh__
#include <iostream>
using namespace std;

template <typename Anneau>
class Quotient
{
private:
  Anneau mod, rep; // le module et le reprsentant 

public:
  // Constructeurs
  Quotient( const Anneau& r=Anneau(0), const Anneau& m=Anneau(0) )
    : mod( assopref(m) ), rep( eumod( r, mod ) ) {};
  Quotient( const Quotient& a )
    : mod(a.mod), rep(a.rep) {};

  // Affectation
  void affecter( const Anneau& r=0, const Anneau& m=0 )
    { mod= assopref(m); rep= eumod( r, mod ); };
  Quotient& operator= ( const Quotient& a )
    { mod= a.mod; rep= a.rep; return *this; };

  // Lire le module et le reprsentant
  const Anneau& module() const { return mod; };
  const Anneau& representant() const { return rep; };

  // Comparaison
  bool operator== ( const Quotient& a ) const
    { return associes( mod, a.mod ) && divisible( rep-a.rep, mod ); }; 
  bool operator!= ( const Quotient& a ) const
    { return !associes( mod, a.mod ) || !divisible( rep-a.rep, mod ); }; 

  // Addition
  Quotient operator+ ( const Quotient& a ) const
    { return Quotient( rep + a.rep, pgcd( mod, a.mod ) ); };
  Quotient& operator+= ( const Quotient& a )
    { affecter( rep + a.rep, pgcd( mod, a.mod ) ); return *this; };

  // Oppos et soustraction
  Quotient operator- () const
    { return Quotient( -rep, mod ); };
  Quotient operator- ( const Quotient& a ) const
    { return Quotient( rep - a.rep, pgcd( mod, a.mod ) ); };
  Quotient& operator-= ( const Quotient& a )
    { affecter( rep - a.rep, pgcd( mod, a.mod ) ); return *this; };

  // Multiplication
  Quotient operator* ( const Quotient& a ) const
    { return Quotient( rep * a.rep, pgcd( mod, a.mod ) ); };
  Quotient& operator*= ( const Quotient& a )
    { affecter( rep * a.rep, pgcd( mod, a.mod ) ); return *this; };
};

//-----------------------------------------------------------------------
// Implmentation d'entre-sortie
//-----------------------------------------------------------------------

template <typename Anneau>
ostream& operator<< ( ostream& out, const Quotient<Anneau>& a )
{
  if ( zero( a.module() ) ) return out << a.representant();
  else return out << a.representant() << "%" << a.module();
}

template <typename Anneau>
istream& operator>> ( istream& in, Quotient<Anneau>& a )
{
  in.clear();

  // Lire le reprsentant
  Anneau rep, mod;
  in >> rep;
  if ( in.fail() ) return in;

  // Lire le symbole sparateur
  char symbole;
  in >> symbole;
  if ( in.fail() ) return in;
  if ( symbole != '%' )
    {
      in.putback(symbole);
      return in;
    };
  
  // Lire le module
  in >> mod;
  if ( in.fail() ) return in;
  
  // Affecter en cas de succs
  a.affecter( rep, mod );
  return in;
}

//-----------------------------------------------------------------------
// Fonctions requises pour travailler avec l'anneau Quotient<Anneau>.
// Exercice : implmenter correctement les fonctions suivantes.
// Attention : dans le quotient, ce sont des oprations non-triviales !
//-----------------------------------------------------------------------

// Reconnatre les lments 0 et 1
template <typename Anneau>
bool zero( const Quotient<Anneau>& a );

template <typename Anneau>
bool one( const Quotient<Anneau>& a );

// Tester si a est inversible, si oui calculer son inverse b
template <typename Anneau>
bool inversible( const Quotient<Anneau>& a );

template <typename Anneau>
bool inversible( const Quotient<Anneau>& a, Quotient<Anneau>& inverse );

// Tester si a est divisible par b, si oui calculer le quotient q
template <typename Anneau>
bool divisible( const Quotient<Anneau>& a, const Quotient<Anneau>& b );

template <typename Anneau>
bool divisible( const Quotient<Anneau>& a, const Quotient<Anneau>& b, Quotient<Anneau>& q );

// Tester si a et b sont associs, si oui calculer u tel que ua=b
template <typename Anneau>
bool associes( const Quotient<Anneau>& a, const Quotient<Anneau>& b );

template <typename Anneau>
bool associes( const Quotient<Anneau>& a, const Quotient<Anneau>& b, Quotient<Anneau>& u );

#endif
