//=======================================================================
// rump.cc : la fonction de Rump est numriquement mchante
// Copyright (C) 2006-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Ce petit programme permet d'valuer la fonction de Rump
// f(x,y) = 1335/4 y^6 + 11/2 y^8 + x/(2y)
//        + x^2 (11 x^2 y^2 - y^6 - 121 y^4 - 2 ) 
// Cette fonction a l'air toute gentille, non ?  Pourtant l'valuation 
// en  x=77617, y=33096  utilisant des nombres  virgule flottante 
// pose de grave problme.  On observe ici une perte totale de chiffres 
// significatifs (explosion d'erreur d'arrondi).
//=======================================================================

#include <iostream>
#include <gmpxx.h>
using namespace std;

// On utilise la classe mpq_class de la bibliothque GMP
// pour calculer avec les nombres rationnels (calcul exact).
// La classe mpf_class permet de le convertir en flottant.
typedef mpq_class Rationnel;
typedef mpf_class BigFloat;

// On dfinit la fonction de Rump comme template.  Ainsi il est
// possible de refaire le mme calcul avec diffrents types.
template <typename Flottant>
Flottant rump( Flottant x, Flottant y )
{
  // Calculer les puissances utilises dans la fonction
  Flottant x2= x * x;
  Flottant y2= y * y;
  Flottant y4= y2 * y2;
  Flottant y6= y4 * y2;
  Flottant y8= y6 * y2;

  // Calculer la fonction de Rump
  Flottant r= Flottant(1335) / Flottant(4) * y6 
    + Flottant(11) / Flottant(2) * y8  +  x / ( Flottant(2) * y )
    + x2 * ( Flottant(11) * x2 * y2 - y6 - Flottant(121) * y4 - Flottant(2) );
  return r;
}

// La fonction main fait le calcul exact puis le calcul numrique
// avec les trois types primitifs du C++ : float, double, long double 
int main()
{
  cout << "\nBonjour et bienvenue  la fonction de Rump dfinie par \n"
       << "f(x,y) = 1335/4 y^6 + 11/2 y^8 + x/(2y) \n"
       << "       + x^2 (11 x^2 y^2 - y^6 - 121 y^4 - 2 ) \n"
       << "Elle est numriquement mchante en x=77617, y=33096. \n"
       << "Comment expliquer ce comportement inattendu ? \n"
       << endl;

  cout << "Entrez deux (petits) entiers x et y svp :" << endl;
  int x, y;
  cout << "x = ";
  cin >> x;
  cout << "y = ";
  cin >> y;

  cout << "Calcul exact :" << endl;
  Rationnel exact= rump( Rationnel(x), Rationnel(y) );
  cout << "  f(x,y) = " << exact << endl;

  cout << "Dveloppement dcimal du rsultat exact :" << endl;
  cout.precision(30);
  cout << "  f(x,y) = " << BigFloat( exact ) << endl;

  cout << "Calcul approch utilisant le type long double :" << endl;
  cout.precision(20);
  cout << "  f(x,y) = " << rump( (long double)(x), (long double)(y) ) << endl;

  cout << "Calcul approch utilisant le type double :" << endl;
  cout.precision(16);
  cout << "  f(x,y) = " << rump( double(x), double(y) ) << endl;
  
  cout << "Calcul approch utilisant le type float :" << endl;
  cout.precision(8);
  cout << "  f(x,y) = " << rump( float(x), float(y) ) << endl;
}
