//=======================================================================
// explog.cc : implmentation des fonctions exp et log
// Copyright (C) <anne> <vos noms>, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description :
// Les fonctions implmentes ci-dessous calculent  exp(x)  et log(x) 
//  une prcision arbitraire tout en garantissant des arrondis fiables.
//
// Avertissement : 
// Il s'agit d'une implmentation exprimentale  titre illustratif.
// Le but principal est d'arriver  des encadrements fiables du rsultat
// exact cherch.  Quant  la performance, on ne fait ici qu'un effort 
// minimal d'optimisation.  En particulier la classe RReal utilise 
// n'est pas optimise.  Pour toute application raliste elle pourrait
// ou devrait tre remplace par une bibliothque plus performante. 
//=======================================================================

#include <iostream>
#include <iomanip>
using namespace std;

#include "interval.cc"
using namespace Numeric;

//-----------------------------------------------------------------------
// Calculer la fonction  exp(x) = 1 + x + x^2/2! + x^3/3! + ...  pour
// x dans l'intervalle [0,1].  Afin d'arriver  un rsultat fiable
// il faut respecter le mode d'arrondi et la prcision choisie.
//
// Stratgie :
// Pour x dans [0,1] les termes  x^k/k!  sont positifs et dcroissent
// vers 0.  Pour tout n le reste  x^(n+1)/(n+1)! + x^(n+2)/(n+2)! + ...
// est major par  2x^(n+1)/(n+1)!  donc par  2/(n+1)!  indpendamment
// de la variable x dans [0,1].  En fonction de la prcision souhaite
// il suffit donc de choisir n convenablement et d'valuer le polynme
// 1 + x + x^2/2! + x^3/3! + ... + x^n/n!  par la mthode de Horner. 
//-----------------------------------------------------------------------
RReal exp1( const RReal& x, Mode mode= stdMode, Len len= stdLen )
{
  //---------------------------------------------------------------------
  // Le choix de l'indice n
  //
  // On sait pour x dans [0,1] que exp(x) est dans [1,3].  Comme 
  // on veut calculer  exp(x)  avec une mantisse de longueur len,
  // l'cart tolr est  eps = 2^(-len).  On choisit n tel que notre 
  // majoration de l'erreur  err = 2/(n+1)!  soit plus petite que eps.
  // Ceci quivaut  choisir n de sorte que  (n+1)! >= 2^(len+1).
  // Ainsi on choisit n suffisamment grand pour que l'approximation 
  // soit suffisamment fine pour une mantisse de la longueur requise.
  // Bien que le choix  n = len  suffise, on peut faire mieux si x est 
  // petit ou len grand ; on trouvera le n optimal par ttonnement. 
  //---------------------------------------------------------------------

  //---------------------------------------------------------------------
  // valuation du polynme
  // 
  // Ayant judicieusement choisi n il ne reste qu' valuer le polynme 
  // s(x) = 1 + x + x^2/2! + x^3/3! + ... + x^n/n!  par la mthode de Horner 
  // s(x) = ((...(( x/n + 1 ) * x/(n-1) + 1 )...) * x/2 + 1 ) * x + 1 .
  // Tous les facteurs  x/k  sont positifs, donc pour obtenir un arrondi
  // RNDU/RNDD/RNDZ/RNDI dans le rsultat final, il suffit d'appliquer
  // ce mme mode d'arrondi  chaque opration lmentaire intermdiaire.
  //---------------------------------------------------------------------

  // Dans le cas d'une majoration il faut encore ajouter une marge d'erreur 
}

//-----------------------------------------------------------------------
// Calculer exp(x) pour x quelconque suivant le mode d'arrondi et 
// la prcision choisie.  On se ramne au cas o x est dans [0,1].
//-----------------------------------------------------------------------
RReal exp( RReal x, Mode mode= stdMode, Len len= stdLen )
{
  //---------------------------------------------------------------------
  // Pour des paramtres  x < 0  on applique exp(x) = 1/exp(-x).
  // On peut ainsi se ramener au cas  x > 0.   noter toutefois
  // qu'il faut changer le mode d'arrondi dans son oppos, cd
  // RNDD <--> RNDU  et  RNDZ <--> RNDI.
  //---------------------------------------------------------------------

  //---------------------------------------------------------------------
  // Pour des paramtres  x > 1 on applique  exp(x) = exp(x/2)^2
  // de manire itre pour se ramener  l'intervalle [-1,0].
  //---------------------------------------------------------------------

}

//-----------------------------------------------------------------------
// Calculer exp(a) pour un intervalle a.  C'est facile car la fonction
// exp : x -> exp(x)  est croissante sur toute la droite relle.
//-----------------------------------------------------------------------
Interval exp( const Interval& a, Len len= stdLen )
{
  
}

//-----------------------------------------------------------------------
// Calculer log(x) pour x dans [1,2].  Afin d'arriver  un rsultat 
// fiable il faut respecter le mode d'arrondi et la prcision choisie.
//
// Stratgie :
// La srie log(1+t) = t - t^2/2 + t^3/3 - t^4/4 + t^5/5  ... converge 
// pour tout t dans ]-1,+1], mais seulement trs lentement proche de 1.
// Pour s'en tirer on passe  une srie qui converge plus rapidement :
// log( (1+t)/(1-t) ) = log(1+t) - log(1-t) = 2t + 2t^3/3 + 2t^5/5 + ...
//-----------------------------------------------------------------------
RReal log1( const RReal& x, Mode mode= stdMode, Len len= stdLen )
{
  // Transformer  x -> t  en respectant le mode d'arrondi

  // Dterminer n tel que  reste < 2^(-len)
  
  // valuer s(t) = 1 + t^2/3 + t^4/5 + ... + t^(2n)/(2n+1) par Horner : 
  // s(t) = ((...( 1/(2n+1) * t^2 + 1/(2n-1) )...) * t^2 + 1/3 ) * t^2 + 1  
  
  // Multiplier la valeur obtenue par 2t : c'est la valeur cherche
  // 2ts(t) = 2t + 2t^3/3 + 2t^5/5 + ... + 2t^(2n+1)/(2n+1) <= log(x)

  // Dans le cas d'une majoration il faut encore ajouter une marge d'erreur 
}

//-----------------------------------------------------------------------
// Calculer log(x) pour x > 0 quelconque. On se ramne au cas x >= 1
// via la formule log(1/x) = -log(x), puis on assure x <= 2 en appliquant
// log(x/2^k) = log(x) - k log(2).
//-----------------------------------------------------------------------
RReal log( RReal x, Mode mode= stdMode, Len len= stdLen )
{
  // Attraper les arguments ngatifs : erreur !
  
  // Attraper le seul cas trivial : log(1) = 0.
  
  // Pour  0 < x < 1  on se ramne au cas  x > 1
  // en appliquant la proprit  log(x) = -log(1/x) .

  // Il ne reste qu' calculer log(x) dans le cas x > 1.
  // La srie utilise converge assez bien pour 1 < x <= 2.
  // Pour  x > 2  on applique log(x) = log(x / 2^k) + k*log(2)
  // afin de ramener x  l'intervalle [1,2].

}

//-----------------------------------------------------------------------
// Calculer log(a) pour un intervalle a.  C'est facile car la fonction
// log : x -> log(x)  est croissante sur toute la demi-droite positive.
//-----------------------------------------------------------------------
Interval log( const Interval& a, Len len= stdLen )
{

}

//-----------------------------------------------------------------------
// Application: la puissance x^y gnralise 
//-----------------------------------------------------------------------
RReal power( const RReal& x, const RReal& y, Mode mode= stdMode, Len len= stdLen )
{

}

Interval power( const Interval& x, const Interval& y, Len len= stdLen )
{

}

//-----------------------------------------------------------------------
// Un exemple d'utilisation
//-----------------------------------------------------------------------
int main()
{
  cout << "\nBonjour et bienvenue au calcul de exp(x) et log(x).\n"
       << "On effectuera un calcul fiable grce aux arrondis dirigs.\n" << endl;
  Len decimales= demander_la_precision();
  
  // Demander la valeur de x
  cout << "Entrer une valeur  x = ";
  RReal x;
  cin >> x;
  
  // L'criture sous forme d'intervalle est trs commode
  cout << "\nJe vais calculer un encadrement fiable de y=exp(x) :" << endl;
  Interval y= exp( Interval(x) );
  Interval z= log( y );

  cout.precision( decimales );
  cout << "x          = " << x << endl;
  cout << "y = exp(x) = " << y << endl;
  cout << "z = log(y) = " << z << endl;

  /*
  // Pour comparaison, le mme calcul directement avec RReal :
  cout << "\nJe vais calculer un encadrement fiable de y=exp(x) :" << endl;
  cout.precision( decimales );
  RReal y1= exp( x, RNDD );
  cout << "minoration fiable = ";
  y1.write_dec( cout, RNDD );
  RReal y2= exp( x, RNDU );
  cout << "\nmajoration fiable = ";
  y2.write_dec( cout, RNDU );
  
  cout << "\n\nPour vrification on applique le logarithme, galement fiable :" << endl;
  RReal z1= log( y1, RNDD );
  cout << "minoration fiable = ";
  z1.write_dec( cout, RNDD );
  RReal z2= log( y2, RNDU );
  cout << "\nmajoration fiable = ";
  z2.write_dec( cout, RNDU );
  */

  cout << "Afin de calculer a^b veuillez entrer deux intervalles svp : " << endl;
  Interval a,b;
  cout << "a = ";
  cin >> a;
  cout << "b = ";
  cin >> b;
  cout << "a = " << a << endl << "b = " << b << endl;
  Interval c= power( a, b );
  cout << "a^b = " << c << endl;
  
  cout << "\n\nAu revoir.\n" << endl;
}
