//=======================================================================
// interval.cc : arithmtique d'intervalles
// Copyright (C) 2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Compilation :
// Compiler avec l'option -lgmpxx pour intgrer la bibliothque gmp. 
//
// Description :
// La classe Interval illustre comment implmenter, avec des outils 
// trs lmentaires, le calcul fiable sous forme d'intervalles.
// Nous utilisons ici la classe RReal faite maison pour contrler 
// explicitement le mode d'arrondi. 
//
// *** Avertissement: aucun effort d'optimisation n'est fait ici. ***
// L'implmentation propose a pour seul but d'illustrer le principe.
// Dans toute application raliste on privilgiera une implmentation 
// plus complte, comme par exemple celle de la bibliothque GMP/MPFI.
// ******************************************************************
//
// Structure des donnes :
// Tout intervalle [mini,maxi] est stock par les deux nombres mini 
// et maxi.  Les bornes sont restreintes aux valeurs reprsentables
// par le type de base, en l'occurrence la classe RReal.
//
// Arithmtique des intervalles :
// On effectue les oprations de base +, -, *, / sur les intervalles,
// de sorte que l'intervalle rsultat englobe toutes les valeurs
// possibles.
//
// Plus prcisment, si l'on note # une des quatre oprations,
// on peut l'tendre  des ensembles A,B de nombres rels en posant  
// A#B = { a#b | a dans A et b dans B }.  Pour l'arithmtique 
// des intervalles on exige que l'intervalle rsultat contienne
// l'ensemble  [a,b]#[c,d] = { x#y | a<=x<=b, c<=y<=d }.   noter  
// que souvent les bornes ne sont pas exactement reprsentables,
// il faut donc arrondir: vers le bas pour la borne infrieure,
// vers le haut pour la borne suprieure.
//=======================================================================

#include <iostream>
#include "rreal.cc"

//-----------------------------------------------------------------------
// On peut considrer Interval comme une petite bibliothque.  On la place
// dans l'espace de noms "Numeric" pour ne pas polluer l'espace "std".
//-----------------------------------------------------------------------
namespace Numeric {

//=======================================================================
// Dfinition de la classe Interval modlisant les intervalles de 
// nombres rels (en prcision arbitraire et avec des arrondis dirigs).
//=======================================================================
class Interval
{
private:
  //---------------------------------------------------------------------
  // Un objet de la classe Interval est donn par une paire (mini,maxi)
  // o mini et maxi sont des nombres rels reprsentables par le type
  // de base, en l'occurrence RReal.  Nous exigeons que mini <= maxi,
  // et pour maintenir cet invariant nous sommes obligs de dclarer 
  // les lments mini et maxi comme tant privs  la classe.  Ceci
  // empche d'accder directement  ces lments -- prudence oblige. 
  //
  //  noter aussi que nous n'autorisons pas l'intervalle vide ici, ni 
  // des intervalles infinis comme  ]-infini,maxi]  ou  [mini,+infini]  
  // voire  ]-infini,+infini[ .  Ces intervalles peuvent tre bien utiles,
  // mais pour simplifier nous ne regardons que des intervalles compacts.
  //---------------------------------------------------------------------

  // Donnes prives: le minimum et le maximum de l'intervalle en question
  RReal mini, maxi;
  
  // Affectation sans contrle ( usage priv uniquement)
  Interval& assign( const RReal& a, const RReal& b )
    { mini= a; maxi= b; return *this; }

  //---------------------------------------------------------------------
  // Constructeurs : initialiser les valeurs mini et maxi. 
  //---------------------------------------------------------------------
public:
  Interval( const Interval& interval )
    : mini( interval.mini ), maxi( interval.maxi ) {}
  
  Interval( const RReal& a, const RReal& b )
    : mini(a), maxi(b) { if( mini > maxi ) mini.swap( maxi ); }
  
  Interval( const RReal& a )
    : mini(a), maxi(a) {}
  
  Interval( double a, double b )
    { if( a > b ) std::swap(a,b); mini= RReal( a, RNDD ); maxi= RReal( b, RNDU ); }

  explicit Interval( double a )
    { mini= RReal( a, RNDD ); maxi= RReal( a, RNDU ); }

  Interval( int a, int b )
    { if( a >= b ) std::swap(a,b); mini= RReal( a, 0, RNDD ); maxi= RReal( b, 0, RNDU ); }

  explicit Interval( int a=0 )
    { mini= RReal( a, 0, RNDD ); maxi= RReal( a, 0, RNDU ); }

  Interval( long a, long b )
    { if( a >= b ) std::swap(a,b); mini= RReal( a, 0l, RNDD ); maxi= RReal( b, 0l, RNDU ); }

  explicit Interval( long a )
    { mini= RReal( a, 0l, RNDD ); maxi= RReal( a, 0l, RNDU ); }

  //---------------------------------------------------------------------
  // change et affectation
  //---------------------------------------------------------------------
  void swap( Interval& a )
    { mini.swap( a.mini ); maxi.swap( a.maxi ); }
  
  Interval& operator= ( const Interval& interval )
    { Interval copie(interval); swap(copie); return *this; }
  
  Interval& set( const Interval& interval )
    { Interval copie(interval); swap(copie); return *this; }

  Interval& set( const RReal& a, const RReal& b )
    { mini= a; maxi= b; if( mini > maxi ) mini.swap( maxi ); return *this; }

  Interval& set( const RReal& a )
    { mini= a; maxi= a; return* this; }
  
  Interval& set( double a, double b )
    { if( a > b ) std::swap(a,b); mini= RReal( a, RNDD ); maxi= RReal( b, RNDU ); return *this; }
  
  Interval& set( double a )
    { mini= RReal( a, RNDD ); maxi= RReal( a, RNDU ); return *this; }

  Interval& set( int a, int b )
    { if( a >= b ) std::swap(a,b); mini= RReal( a, 0, RNDD ); maxi= RReal( b, 0, RNDU ); return *this; }

  Interval& set( int a )
    { mini= RReal( a, 0, RNDD ); maxi= RReal( a, 0, RNDU ); return* this; }

  Interval& set( long a, long b )
    { if( a >= b ) std::swap(a,b); mini= RReal( a, 0l, RNDD ); maxi= RReal( b, 0l, RNDU ); return *this; }

  Interval& set( long a )
    { mini= RReal( a, 0l, RNDD ); maxi= RReal( a, 0l, RNDU ); return* this; }

  //---------------------------------------------------------------------
  // Accs contrl aux informations -- en lecture seulement !
  //---------------------------------------------------------------------
public:
  const RReal& min() const { return mini; }   // lire le minimum
  const RReal& max() const { return maxi; }   // lire le maximum
  RReal len() const                           // calculer la longueur
    { return Numeric::sub(maxi,mini,RNDU); } 

  //---------------------------------------------------------------------
  // Test d'inclusion (pour la comparaison voir plus bas) 
  //---------------------------------------------------------------------
  bool isPoint() const 
    { return ( mini == maxi ); }
  bool contains( const RReal& a ) const
    { return ( mini <= a && a <= maxi ); }
  bool contains( const Interval& a ) const 
    { return ( mini <= a.mini && a.maxi <= maxi ); }

  //---------------------------------------------------------------------
  // Calculer le point au milieu et le rayon (= la marge d'erreur)
  //---------------------------------------------------------------------
  RReal mid() const 
    { 
      if( mini == maxi ) return mini;
      return Numeric::div( Numeric::add(mini,maxi,RNDN), 2, RNDN );
    }
  
  RReal err() const 
    { 
      return Numeric::div( Numeric::sub( maxi, mini, RNDU ), 2, RNDU ); 
    }
  
  void mid_err( RReal& m, RReal& r ) const 
    { 
      m= mid();
      r= std::max( Numeric::abs( Numeric::sub( maxi, m, RNDU ) ) , 
		   Numeric::abs( Numeric::sub( m, mini, RNDU ) ) );
    }

  //---------------------------------------------------------------------
  // Manipuler le signe 
  //---------------------------------------------------------------------
  Interval operator+ () const
    { return *this; }
  Interval operator- () const
    { return Interval( -maxi, -mini ); }

  //---------------------------------------------------------------------
  // Addition de deux intervalles (ou d'un intervalle et d'un rel)
  //---------------------------------------------------------------------
  Interval& add( const Interval& a, const Interval& b, Len len= stdLen )
    { return assign( Numeric::add( a.mini, b.mini, RNDD, len ), 
		     Numeric::add( a.maxi, b.maxi, RNDU, len ) ); }

  Interval& add( const RReal& a, const Interval& b, Len len= stdLen )
    { return assign( Numeric::add( a, b.mini, RNDD, len ), 
		     Numeric::add( a, b.maxi, RNDU, len ) ); }

  Interval& add( const Interval& a, const RReal& b, Len len= stdLen )
    { return assign( Numeric::add( a.mini, b, RNDD, len ), 
		     Numeric::add( a.maxi, b, RNDU, len ) ); }

  Interval& add( const Interval& a, Len len= stdLen )
    { return add( *this, a, len ); };
  
  Interval& add( const RReal& a, Len len= stdLen )
    { return add( *this, a, len ); };
  
  Interval& operator+= ( const Interval& a )
    { return add( a ); }

  Interval& operator+= ( const RReal& a )
    { return add( a ); }
    
  //---------------------------------------------------------------------
  // Soustraction de deux intervalles (ou d'un intervalle et d'un rel)
  //---------------------------------------------------------------------
  Interval& sub( const Interval& a, const Interval& b, Len len= stdLen )
    { return assign( Numeric::sub( a.mini, b.maxi, RNDD, len ), 
		     Numeric::sub( a.maxi, b.mini, RNDU, len ) ); }

  Interval& sub( const RReal& a, const Interval& b, Len len= stdLen )
    { return assign( Numeric::sub( a, b.maxi, RNDD, len ), 
		     Numeric::sub( a, b.mini, RNDU, len ) ); }

  Interval& sub( const Interval& a, const RReal& b, Len len= stdLen )
    { return assign( Numeric::sub( a.mini, b, RNDD, len ), 
		     Numeric::sub( a.maxi, b, RNDU, len ) ); }

  Interval& sub( const Interval& a, Len len= stdLen )
    { return sub( *this, a, len ); };
  
  Interval& sub( const RReal& a, Len len= stdLen )
    { return sub( *this, a, len ); };
  
  Interval& operator-= ( const Interval& a )
    { return sub( a ); }

  Interval& operator-= ( const RReal& a )
    { return sub( a ); }
    
  //---------------------------------------------------------------------
  // Multiplication d'un intervalle par un nombre rel
  //---------------------------------------------------------------------
  Interval& mul( const RReal& a, const Interval& b, Len len= stdLen )
    {
      using Numeric::mul;
      int s= sign(a);
      if ( s > 0 )         // multiplication croissante
	return assign( mul( a, b.mini, RNDD, len ), 
		       mul( a, b.maxi, RNDU, len ) );
      if ( s < 0 )         // multiplication dcroissante
	return assign( mul( a, b.maxi, RNDD, len ), 
		       mul( a, b.mini, RNDU, len ) );
      return assign( a, a );  // multiplication par zro
    }
  
  Interval& mul( const Interval& b, const RReal& a, Len len= stdLen )
    { return mul( a, b, len ); }

  Interval& mul( const RReal& a, Len len= stdLen )
    { return mul( *this, a, len ); };
  
  Interval& operator*= ( const RReal& a )
    { return mul( a ); }
    
  //---------------------------------------------------------------------
  // Multiplication de deux intervalles
  // Parmi les quatre oprations lmentaires sur les intervalles,
  // la multiplication  [a1,a2]*[b1,b2]  est la plus dlicate.
  // Les bornes peuvent tre de diffrents signes suivant trois cas :
  // [a1,a2] :  ou a1 <= a2 <= 0  ou  a1 < 0 < a2  ou  0 <= a1 <= a2 ,
  // [b1,b2] :  ou b1 <= b2 <= 0  ou  b1 < 0 < b2  ou  0 <= b1 <= b2 .
  // Cela fait donc neuf configurations possibles.  L'implmentation
  // suivante ne ncessite que deux multiplications pour chacune de 
  // ces configurations, sauf la dernire qui en ncessite quatre.
  //---------------------------------------------------------------------
  Interval& mul( const Interval& a, const Interval& b, Len len= stdLen )
    { 
      // Premier cas : 0 <= a.mini <= a.maxi
      using Numeric::mul;
      if( sign( a.mini ) >= 0 )
	if( sign( b.mini ) >= 0 )       // 0 <= b.mini <= b.maxi
	  return assign( mul( a.mini, b.mini, RNDD, len ),
			 mul( a.maxi, b.maxi, RNDU, len ) );
	else if( sign( b.maxi ) <= 0 )  // b.mini <= b.maxi <= 0
	  return assign( mul( a.maxi, b.mini, RNDD, len ),
			 mul( a.mini, b.maxi, RNDU, len ) );
	else                            // b.mini < 0 < b.maxi 
	  return assign( mul( a.maxi, b.mini, RNDD, len ),
			 mul( a.maxi, b.maxi, RNDU, len ) );
      
      // Second cas : a.mini <= a.maxi <= 0
      if( sign( a.maxi ) <= 0 )
	if( sign( b.mini ) >= 0 )       // 0 <= b.mini <= b.maxi
	  return assign( mul( a.mini, b.maxi, RNDD, len ),
			 mul( a.maxi, b.mini, RNDU, len ) );
	else if( sign( b.maxi ) <= 0 )  // b.mini <= b.maxi <= 0
	  return assign( mul( a.maxi, b.maxi, RNDD, len ),
			 mul( a.mini, b.mini, RNDU, len ) );
	else                            // b.mini < 0 < b.maxi 
	  return assign( mul( a.mini, b.maxi, RNDD, len ),
			 mul( a.mini, b.mini, RNDU, len ) );
      
      // Finalement le troisime cas: a.mini < 0 < a.maxi
      if( sign( b.mini ) >= 0 )       // 0 <= b.mini <= b.maxi
	return assign( mul( a.mini, b.maxi, RNDD, len ),
		       mul( a.maxi, b.maxi, RNDU, len ) );
      else if( sign( b.maxi ) <= 0 )  // b.mini <= b.maxi <= 0
	return assign( mul( a.maxi, b.mini, RNDD, len ),
		       mul( a.mini, b.mini, RNDU, len ) );
      else                            // b.mini < 0 < b.maxi
	return assign( std::min( mul( a.mini, b.maxi, RNDD, len ), 
				 mul( a.maxi, b.mini, RNDD, len ) ),
		       std::max( mul( a.mini, b.mini, RNDU, len ), 
				 mul( a.maxi, b.maxi, RNDU, len ) ) );
    }
  
  Interval& mul( const Interval& a, Len len= stdLen )
    { return mul( *this, a, len ); };
  
  Interval& operator*= ( const Interval& a )
    { return mul( a ); }
  
  //---------------------------------------------------------------------
  // Compare avec la multiplication, l'inversion  x -> 1/x  est plutt 
  // simple car elle est monotone,  savoir dcroissante.
  //---------------------------------------------------------------------
  Interval& inv( const Interval& a, Len len= stdLen )
    {
      if( sign(a.mini) <= 0 && sign(a.maxi) >= 0 )
	{ cerr << "Interval division by zero!" << endl; exit(1); }
      return assign( Numeric::div( 1, a.maxi, RNDD, len ), 
		     Numeric::div( 1, a.mini, RNDU, len ) );
    }
  
  Interval& inv( const RReal& a, Len len= stdLen )
    { return assign( Numeric::div( 1, a, RNDD, len ), 
		     Numeric::div( 1, a, RNDU, len ) ); }
  
  //---------------------------------------------------------------------
  // La division a/b est ramene  une multiplication  a * (1/b).
  //---------------------------------------------------------------------
  Interval& div( const Interval& a, const Interval& b, Len len= stdLen )
    { return mul( a, inv( b, len ), len ); }
  
  Interval & div( const RReal& a, const Interval& b, Len len= stdLen )
    { return mul( a, inv( b, len ), len ); }
  
  Interval& div( const Interval& a, const RReal& b, Len len= stdLen )
    { return mul( a, inv( b, len ), len ); }

  Interval & div( const Interval& a, Len len= stdLen )
    { return div( *this, a, len ); }
  
  Interval & div( const RReal& a, Len len= stdLen )
    { return div( *this, a, len ); }
  
  Interval& operator/= ( const Interval& a )
    { return div( a ); }

  Interval& operator/= ( const RReal& a )
    { return div( a ); }

  //---------------------------------------------------------------------
  // La comparaison d'intervalles est un sujet dlicat :
  // il n'y pas d'ordre naturel mais plusieurs choix de conventions !
  // On n'implmente que quelques unes des conventions possibles...
  //---------------------------------------------------------------------
  bool surely_less_than( const RReal& a ) const 
    { return ( maxi < a ); }
  bool surely_less_or_equal( const RReal& a ) const 
    { return ( maxi <= a ); }
  bool surely_greater_than( const RReal& a ) const 
    { return ( mini > a ); }
  bool surely_greater_or_equal( const RReal& a ) const 
    { return ( mini >= a ); }

  bool possibly_less_than( const RReal& a ) const 
    { return ( mini < a ); }
  bool possibly_less_or_equal( const RReal& a ) const 
    { return ( mini <= a ); }
  bool possibly_greater_than( const RReal& a ) const 
    { return ( maxi > a ); }
  bool possibly_greater_or_equal( const RReal& a ) const 
    { return ( maxi >= a ); }

  bool surely_less_than( const Interval& a ) const 
    { return ( maxi < a.mini ); }
  bool possibly_less_than( const Interval& a ) const 
    { return ( mini < a.mini ); }
  bool probably_less_than( const Interval& a ) const 
    { return ( mini < a.mini && maxi < a.maxi ); }

  bool surely_greater_than( const Interval& a ) const 
    { return ( mini > a.maxi ); }
  bool possibly_greater_than( const Interval& a ) const 
    { return ( maxi > a.maxi ); }
  bool probably_greater_than( const Interval& a ) const 
    { return ( mini > a.mini && maxi > a.maxi ); }

  //-----------------------------------------------------------------------
  // Notation pour l'affichage d'intervalles
  //
  // Exemples : 
  //   [ 1.56 ; 1.62 ]  affichage du minimum et du maximum (standard)
  //    1.59 +/- 0.03   affichage de la mdiane et de la marge d'erreur
  //    1.59            affichage de la mdiane seulement 
  // L'affichage standard explicite le minimum et le maximum : c'est la
  // notation la plus prcise et elle est conseille pour l'usage gnral.
  // Parfois la second notation est plus commode, mais on perd on peu 
  // d'informations.  L'affichage de la mdiane seulement est la plus
  // concise mais la moins informative : on supprime toute information
  // explicite sur la marge d'erreur --  utiliser avec modration.
  //-----------------------------------------------------------------------  
  enum Notation
    {
      display_min_max,  // display minimum and maximum
      display_mid_err,  // display midpoint and error 
      display_mid_only  // display midpoint only
    };

  static Notation stdNotation;
  
  static Notation notation()
    { return stdNotation; };
  
  static Notation notation( Notation n ) 
    { Notation old= stdNotation; stdNotation= n; return old; }
  
  //-----------------------------------------------------------------------
  // Sortie : affichage dans un format choisi 
  //-----------------------------------------------------------------------  
  void write( ostream& out ) const
    {
      // Notation par dfaut : afficher le minimum et le maximum
      if( stdNotation == display_min_max )
	{
	  out << "[ ";
	  mini.write( out, RNDD );
	  out << " ; ";
	  maxi.write( out, RNDU );
	  out << " ]";
	  return;
	}
      
      //-------------------------------------------------------------------
      // Dterminer le nombre de dcimales significatives.
      // Si la marge d'erreur est non nulle, il sera inutile voire 
      // trompeur d'afficher un nombre trop lev de dcimales !
      //-------------------------------------------------------------------
      RReal mid, err;
      mid_err( mid, err );
      Len decimales= out.precision();
      if( sign(mid) != 0 && sign(err) > 0 )
	{
	  Len bits_significatifs= castExpToLen( mid.magnitude() - err.magnitude() );
	  if( bits_significatifs < 2 ) bits_significatifs= 2;
	  decimales= std::min( decimales, (4+bits_significatifs*3)/10 );
	}

      // Afficher le point au milieu seulement
      if( stdNotation == display_mid_only )
	{
	  mid.write_dec( out, RNDN, decimales );
	  return;
	}
      
      // Afficher le point au milieu et la marge d'erreur
      out << "( ";
      mid.write_dec( out, RNDN, decimales );
      out << " +/- ";
      err.write_dec( out, RNDU, 3 );
      out << " )";
    }
  
  //-----------------------------------------------------------------------
  // Entre : lecture dans le format standard [min;max] ou bien [min,max]
  // On autorise aussi la lecture d'un seul nombre  virgule flottante ici.
  // Dans ce cas ce mme nombre est arrondi vers le bas pour min et arrondi
  // vers le haut pour max, afin d'obtenir un encadrement fiable.
  //-----------------------------------------------------------------------  
  void read( istream& in, Len len= stdLen )
    { 
      // remettre le flot  l'tat normal (si possible)
      in.clear();
      if ( in.bad() ) return;
      
      // lire le crochet ouvrant
      char symbole;
      in >> symbole;
      if ( in.fail() ) return;
      if ( symbole != '[' ) 
	{
	  // Cas particulier : lire un seul rel
	  in.putback( symbole );
	  read_dec( in, len );
	  return;
	}
  
      // lire la borne infrieure
      RReal min;
      min.read( in, RNDD, len );
      if ( in.fail() ) return;
      
      // lire le point-virgule (erreur sinon)
      in >> symbole;
      if ( in.fail() ) return;
      if ( symbole != ';' && symbole !=',' )
	{
	  in.putback( symbole );
	  in.setstate( ios::failbit ); 
	  return;
	}
      
      // lire la borne suprieure
      RReal max;
      max.read( in, RNDU, len );
      if ( in.fail() ) return;
      
      // lire le crochet fermant (erreur sinon)
      in >> symbole;
      if ( in.fail() ) return;
      if ( symbole != ']' )
	{
	  in.putback( symbole );
	  in.setstate( ios::failbit ); 
	  return;
	}
      
      // Affecter si tout s'est bien pass
      set( min, max );
    }

  //-----------------------------------------------------------------------
  // Entre : lecture dans nombre rel.
  // Il faut arrondir deux fois pour en dduire un intervalle.
  //-----------------------------------------------------------------------  
  void read_dec( istream& in, Len len= stdLen )
    {
      // Lire la mantisse et l'exposant
      Man man;
      Exp exp;
      RReal::read_dec( in, man, exp );
      
      // Convertir le nombre en format interne (avec arrondi)
      mini.set( man, 0, RNDD, len );
      maxi.set( man, 0, RNDU, len );
      RReal base(10);
      for( ; exp>0 ; --exp ) 
	{
	  mini.mul( base, RNDD, len );
	  maxi.mul( base, RNDU, len );
	}
      for( ; exp<0 ; ++exp ) 
	{
	  mini.div( base, RNDD, len );
	  maxi.div( base, RNDU, len );
	}
     }
};
  
//-----------------------------------------------------------------------
// Dfinir les variables statiques (= globales pour Interval)
//-----------------------------------------------------------------------
Interval::Notation Interval::stdNotation= display_min_max; 

//-----------------------------------------------------------------------
// Oprations arithmtiques crant un objet temporaire : il ne s'agit 
// que de rformulations commodes des oprations dj implmentes.
//-----------------------------------------------------------------------
Interval add( const Interval& a, const Interval& b, Len len= stdLen )
{ Interval c; c.add( a, b, len ); return c; }

Interval add( const Interval& a, const RReal& b, Len len= stdLen )
{ Interval c; c.add( a, b, len ); return c; }

Interval add( const RReal& a, const Interval& b, Len len= stdLen )
{ Interval c; c.add( a, b, len ); return c; }

Interval operator+ ( const Interval& a, const Interval& b )
{ Interval c; c.add( a, b ); return c; }

Interval operator+ ( const Interval& a, const RReal& b )
{ Interval c; c.add( a, b ); return c; }

Interval operator+ ( const RReal& a, const Interval& b )
{ Interval c; c.add( a, b ); return c; }


Interval sub( const Interval& a, const Interval& b, Len len= stdLen )
{ Interval c; c.sub( a, b, len ); return c; }

Interval sub( const Interval& a, const RReal& b, Len len= stdLen )
{ Interval c; c.sub( a, b, len ); return c; }

Interval sub( const RReal& a, const Interval& b, Len len= stdLen )
{ Interval c; c.sub( a, b, len ); return c; }
  
Interval operator- ( const Interval& a, const Interval& b )
{ Interval c; c.sub( a, b ); return c; }

Interval operator- ( const Interval& a, const RReal& b )
{ Interval c; c.sub( a, b ); return c; }

Interval operator- ( const RReal& a, const Interval& b )
{ Interval c; c.sub( a, b ); return c; }
  

Interval mul( const Interval& a, const Interval& b, Len len= stdLen )
{ Interval c; c.mul( a, b, len ); return c; }

Interval mul( const Interval& a, const RReal& b, Len len= stdLen )
{ Interval c; c.mul( a, b, len ); return c; }

Interval mul( const RReal& a, const Interval& b, Len len= stdLen )
{ Interval c; c.mul( a, b, len ); return c; }
  
Interval operator* ( const Interval& a, const Interval& b )
{ Interval c; c.mul( a, b ); return c; }

Interval operator* ( const Interval& a, const RReal& b )
{ Interval c; c.mul( a, b ); return c; }

Interval operator* ( const RReal& a, const Interval& b )
{ Interval c; c.mul( a, b ); return c; }
  

Interval div( const Interval& a, const Interval& b, Len len= stdLen )
{ Interval c; c.div( a, b, len ); return c; }

Interval div( const Interval& a, const RReal& b, Len len= stdLen )
{ Interval c; c.div( a, b, len ); return c; }

Interval div( const RReal& a, const Interval& b, Len len= stdLen )
{ Interval c; c.div( a, b, len ); return c; }

Interval operator/ ( const Interval& a, const Interval& b )
{ Interval c; c.div( a, b ); return c; }

Interval operator/ ( const Interval& a, const RReal& b )
{ Interval c; c.div( a, b ); return c; }

Interval operator/ ( const RReal& a, const Interval& b )
{ Interval c; c.div( a, b ); return c; }


Interval inv( const Interval& a, Len len= stdLen  )
{ Interval b; b.inv( a, len ); return b; }

Interval inv( const RReal& a, Len len= stdLen  )
{ Interval b; b.inv( a, len ); return b; }
  
//-----------------------------------------------------------------------
// Dterminer le signe : on s'accorde sur la convention indique
//-----------------------------------------------------------------------
int sign( const Interval& a ) 
{
  if( sign( a.min() ) > 0 ) return +1;  // 0 < min <= max
  if( sign( a.max() ) < 0 ) return -1;  // min <= max < 0
  return 0;                             // min <= 0 <= max
}

//-----------------------------------------------------------------------
// Valeur absolue : la fonction  abs : x -> |x|  est 
// croissante pour  x >= 0  mais dcroissante pour  x <= 0 . 
//-----------------------------------------------------------------------
Interval abs( const Interval& a )
{ 
  if( sign( a.min() ) >= 0 ) return a;
  if( sign( a.max() ) <= 0 ) return -a;
  return Interval( RReal(0), max( -a.min(), a.max() ) );
}

//-----------------------------------------------------------------------
// Maximum et minimum de deux intervalles
//-----------------------------------------------------------------------
Interval max( const Interval& a, const Interval& b )
{ return Interval( std::max( a.min(), b.min() ), 
		   std::max( a.max(), b.max() ) ); }

Interval min( const Interval& a, const Interval& b )
{ return Interval( std::min( a.min(), b.min() ), 
		   std::min( a.max(), b.max() ) ); }

//-----------------------------------------------------------------------
// Calculer le carr 
// Le carr  A d'un intervalle diffre en gnral du produit A*A.  
// Exemple : [-2,1] = [0,4] mais [-2,1]*[-2,1] = [-2,4] !
//-----------------------------------------------------------------------
Interval sqr( const Interval& a, Len len= stdLen )
{
  // La fonction  x -> x^2  est croissante pour x >= 0.  
  if( a.min() >= 0 ) 
    return Interval( sqr( a.min(), RNDD, len ), sqr( a.max(), RNDU, len ) );
  
  // La fonction  x -> x^2  est dcroissante pour x <= 0.
  if( a.max() <= 0 )
    return Interval( sqr( a.max(), RNDD, len ), sqr( a.min(), RNDU, len ) );
  
  // Finalement, il faut traiter le cas o min < 0 < max.
  return Interval( RReal(0), std::max( sqr( a.min(), RNDU, len ), 
				       sqr( a.max(), RNDU, len ) ) );
}

//-----------------------------------------------------------------------
// Calculer la racine carre : la fonction sqrt dclenche 
// une erreur si l'on l'applique aux rels ngatifs.
//-----------------------------------------------------------------------
Interval sqrt( const Interval& a, Len len= stdLen )
{
  return Interval( sqrt( a.min(), RNDD, len ), sqrt( a.max(), RNDU, len ) );
}

//-----------------------------------------------------------------------
// Calculer la puissance n-ime d'un intervalle.
//  noter que pour un intervalle  A=[a,b]  et un exposant n pair
// la puissance  A^n diffre en gnral du produit itr  A*...*A.
//-----------------------------------------------------------------------
Interval power( const Interval& a, long n, Len len= stdLen )
{
  if( n == 1 ) return a;
  if( n == 0 ) return Interval(1);
  if( n < 0 )  return power( inv(a), -n, len );
  
  // Pour n impair la fonction  x -> x^n  est croissante,
  // et pour n pair elle est croissante pour x >= 0.  
  if( n%2 != 0 || a.min() >= 0 ) 
    return Interval( power( a.min(), n, RNDD, len ),
		     power( a.max(), n, RNDU, len ) );
  
  // Pour n pair la fonction  x -> x^n  est dcroissante pour x <= 0.
  if( a.max() <= 0 )
    return Interval( power( a.max(), n, RNDD, len ),
		     power( a.min(), n, RNDU, len ) );
  
  // Finalement, il faut traiter le cas o n est pair et min < 0 < max.
  return Interval( RReal(0), std::max( power( a.min(), n, RNDU, len ),
				       power( a.max(), n, RNDU, len ) ) );
}

//-----------------------------------------------------------------------
// Calculer la racine n-ime d'un intervalle. 
// Pour n impair la fonction x -> x^(1/n) est dfinie sur toute la droite 
// relle, et elle est croissante.  Pour n pair elle n'est dfinie que 
// sur la demi-droite  x >= 0 , mais elle y est galement croissante.
// Par consquent la fonction root dclenche une erreur si l'on 
// l'applique aux nombres rels ngatifs et n pair.
//-----------------------------------------------------------------------
Interval root( const Interval& a, long n, Len len= stdLen )
{
  return Interval( root( a.min(), n, RNDD, len ), root( a.max(), n, RNDU, len ) );
}

//-----------------------------------------------------------------------
// L'galit et la diffrence de deux intervalles est facile  dfinir.
// Par commodit on implmente aussi la comparaison avec un point.
//-----------------------------------------------------------------------
bool operator== ( const Interval& a, const Interval& b )
{ return ( a.min() == b.min() && a.max() == b.max() ); }

bool operator== ( const RReal& a, const Interval& b )
{ return ( a == b.min() && a == b.max() ); }

bool operator== ( const Interval& a, const RReal& b )
{ return ( a.min() == b && a.max() == b ); }


bool operator!= ( const Interval& a, const Interval& b )
{ return ( a.min() != b.min() || a.max() != b.max() ); }

bool operator!= ( const RReal& a, const Interval& b )
{ return ( a != b.min() || a != b.max() ); }

bool operator!= ( const Interval& a, const RReal& b )
{ return ( a.min() != b || a.max() != b ); }

//-----------------------------------------------------------------------
// Comparaison (1re variante) --  utiliser avec prudence.
//
//  noter que la comparaison d'intervalles est un sujet dlicat:
// selon l'application plusieurs conventions sont possibles !
//-----------------------------------------------------------------------
bool operator<= ( const Interval& a, const Interval& b )
{ return ( a.min() <= b.min() ) && ( a.max() <= b.max() ); }

bool operator<= ( const RReal& a, const Interval& b )
{ return ( a <= b.min() ); }

bool operator<= ( const Interval& a, const RReal& b )
{ return ( a.max() <= b ); }


bool operator>= ( const Interval& a, const Interval& b )
{ return ( a.min() >= b.min() ) && ( a.max() >= b.max() ); }

bool operator>= ( const RReal& a, const Interval& b )
{ return ( a >= b.max() ); }

bool operator>= ( const Interval& a, const RReal& b )
{ return ( a.min() >= b ); }

//-----------------------------------------------------------------------
// Comparaison lexicographique --  utiliser avec prudence.
//
//  noter que la comparaison d'intervalles est un sujet dlicat:
// selon l'application plusieurs conventions sont possibles !
//-----------------------------------------------------------------------
bool operator< ( const Interval& a, const Interval& b )
{
  if ( a.min() < b.min() ) return true;
  if ( a.min() > b.min() ) return false;
  return ( a.max() < b.max() );
}

bool operator< ( const RReal& a, const Interval& b )
{ return ( a < b.min() ); }

bool operator< ( const Interval& a, const RReal& b )
{ return ( a.max() < b ); }


bool operator> ( const Interval& a, const Interval& b )
{
  if ( a.min() > b.min() ) return true;
  if ( a.min() < b.min() ) return false;
  return ( a.max() > b.max() );
}

bool operator> ( const RReal& a, const Interval& b )
{ return ( a > b.max() ); }

bool operator> ( const Interval& a, const RReal& b )
{ return ( a.min() > b ); }

//-----------------------------------------------------------------------
// Entre-sortie sous forme d'oprateurs
//-----------------------------------------------------------------------  
ostream& operator<< ( ostream& out, const Interval& a )
{ a.write( out ); return out; }

istream& operator>> ( istream& in, Interval& a )
{ a.read( in ); return in; }

//-----------------------------------------------------------------------
} // fin de l'espace de noms "Numeric"
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Spcialiser la fonction  swap(Interval,Interval)  dans l'espace "std"
//-----------------------------------------------------------------------
namespace std
{
  template<>
  void swap<Numeric::Interval>( Numeric::Interval& a, Numeric::Interval& b ) 
  { a.swap(b); }
}
