//=======================================================================
// polynome.cc : implmentation basique de polynomes complexes
// Copyright (C) 2005-2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
// 
// Structure des donnes :
// Un polynme  P = a[0] + ... + a[n]X^n  de degr n est stock par 
// un vecteur de coefficients ( a[0], a[1], ..., a[n] ). Pour un usage
// commode on implmente une trs simple classe  Polynome  qui permet de
// - lire et crire la liste de coefficients,
// - effectuer quelques oprations de base, notamment
// - valuer le polynme par p(x) en un point x.
//=======================================================================

#include <vector>    // vecteurs gnriques de la STL
using namespace std;

//-----------------------------------------------------------------------
// Dfinir le type C des nombres complexes
//-----------------------------------------------------------------------
#include <complex>       // nombres complexes gnriques
typedef complex<Flo> C;  // nombres complexes utilisant le type Flo

//-----------------------------------------------------------------------
// La classe Polynome
//-----------------------------------------------------------------------
class Polynome
{
public:
  // Les coefficients du polynme stocks dans un vecteur
  vector<C> coeffs;

  // Constante statique = globale  la classe Polynome
  static const C zero; 

  // Constructeurs divers
  Polynome( int deg= -1 )
    { if (deg>=0) coeffs.resize( deg+1, zero ); };
  Polynome( const C& a )
    { coeffs.resize( 1 ); coeffs[0]= a; };
  Polynome( const C& a, const C& b )
    { coeffs.resize( 2 ); coeffs[0]= a; coeffs[1]= b; };
  Polynome( const vector<C>& liste )
    : coeffs( liste ) {};
  
  // Accs en lecture (indpendemment des limites du vecteur)
  const C& operator[] ( size_t indice ) const;

  // Accs en criture (indpendemment des limites du vecteur)
  C& operator[] ( size_t indice );

  // valuer le polynme en un point x
  C operator() ( const C& x ) const;
};

//-----------------------------------------------------------------------
// La constante zero. Elle est utilise comme coefficient par dfaut.
//-----------------------------------------------------------------------
const C Polynome::zero( Flo(0) );

//-----------------------------------------------------------------------
// On peut maintenant dfinir la variable X comme le polynome ( 0 1 )
//-----------------------------------------------------------------------
const Polynome X( Flo(0), Flo(1) );

//-----------------------------------------------------------------------
// Dterminer le degr d'un polynme (pour avoir une criture commode)
// Attention: le polynme nul est reprsent par un vecteur vide, son
// degr vaut donc -1. Les polynmes constants non nuls ont degr 0.
//-----------------------------------------------------------------------
int degre( const Polynome& a )
{
  return a.coeffs.size() - 1;
}

//-----------------------------------------------------------------------
// Lire un coefficient indpendemment des limites du vecteur. Ceci est 
// souvent plus commode car notre oprateur [] teste automatiquement 
// si l'indice tombe dans le domaine du vecteur. Sinon on renvoie 0.
// Attention : on ne renvoie pas la valeur du coefficient mais une
// rfrence constante, ce qui permet la lecture mais pas l'criture.
//-----------------------------------------------------------------------
const C& Polynome::operator[] ( size_t indice ) const
{
  if ( indice >= coeffs.size() ) return zero;
  return coeffs[indice];
}

//-----------------------------------------------------------------------
// crire un coefficient indpendemment des limites du vecteur. 
// Si l'indice est en dhors du vecteur dfini jusqu'ici, alors on
// le rallonge par des coefficients nuls.  On renvoie une rfrence 
// qui permet la lecture *et* l'criture du coefficient ainsi index.
//-----------------------------------------------------------------------
C& Polynome::operator[] ( size_t indice )
{
  if ( indice >= coeffs.size() ) 
    { coeffs.resize( indice+1, zero ); }
  return coeffs[indice];
}

//-----------------------------------------------------------------------
// valuer un polynme en un point x par la mthode de Horner
//-----------------------------------------------------------------------
C Polynome::operator() ( const C& x ) const 
{
  if ( coeffs.empty() ) return zero;
  C valeur= coeffs.back();
  for( int k= coeffs.size()-2; k>=0; --k )
    valeur= valeur*x + coeffs[k];
  return valeur;
}

//-----------------------------------------------------------------------
// Addition de deux polynmes
//-----------------------------------------------------------------------
Polynome operator+ ( const Polynome& a, const Polynome& b )
{
  int deg= max( degre(a), degre(b) );
  Polynome somme( deg );
  for( int k=0; k<=deg; ++k )
    somme.coeffs[k]= a[k] + b[k];
  return somme;
}

Polynome& operator+= ( Polynome& a, const Polynome& b )
{ return ( a = a+b ); }

//-----------------------------------------------------------------------
// Soustraction de deux polynmes
//-----------------------------------------------------------------------
Polynome operator- ( const Polynome& a, const Polynome& b )
{
  int deg= max( degre(a), degre(b) );
  Polynome somme( deg );
  for( int k=0; k<=deg; ++k )
    somme.coeffs[k]= a[k] - b[k];
  return somme;
}

Polynome& operator-= ( Polynome& a, const Polynome& b )
{ return ( a = a-b ); }

//-----------------------------------------------------------------------
// Multiplication d'un polynme par un scalaire
//-----------------------------------------------------------------------
Polynome operator* ( const C& a, const Polynome& b )
{
  if ( a==Polynome::zero || degre(b)<0 ) return Polynome();
  Polynome produit( degre(b) );
  for( int j=0; j<=degre(b); ++j )
    produit.coeffs[j]= a * b.coeffs[j];
  return produit;
}

Polynome operator* ( const Polynome& a, const C& b )
{ return b*a; }

Polynome& operator*= ( Polynome& a, const C& b )
{ return ( a = b*a ); }

//-----------------------------------------------------------------------
// Multiplication de deux polynmes
//-----------------------------------------------------------------------
Polynome operator* ( const Polynome& a, const Polynome& b )
{
  if ( degre(a)<0 || degre(b)<0 ) return Polynome();
  Polynome produit( degre(a) + degre(b) );
  for( int i=0; i<=degre(a); ++i )
    for( int j=0; j<=degre(b); ++j )
      produit.coeffs[i+j] += a.coeffs[i] * b.coeffs[j];
  return produit;
}  

Polynome& operator*= ( Polynome& a, const Polynome& b )
{ return ( a = a*b ); }

//-----------------------------------------------------------------------
// Oprateur de sortie
//
// Le format choisi ici est le plus simple qui soit : on liste les
// coefficients entre des paranthses "(" et ")", spars d'espaces.
// Ainsi X^3 - 5X + 2 sera crit ( 2, -5, 0, 1 ).  
//-----------------------------------------------------------------------
ostream& operator << ( ostream& out, const Polynome& lePolynome )
{
  out << "( ";
  for ( size_t i=0; i<lePolynome.coeffs.size(); ++i )
    out << lePolynome.coeffs[i] << " ";
  out << ")";
  return out;
}

//-----------------------------------------------------------------------
// Oprateur d'entre
//
// L'entre accepte le format produit par la sortie. Comme toujours 
// l'implmentation est plus dlicate parce qu'il faut tenir compte
// de diffrents problmes, notamment d'ventuelles erreurs de syntaxe.
//-----------------------------------------------------------------------
istream& operator >> ( istream& in, Polynome& lePolynome )
{
  // Tout d'abord on efface d'ventuels anciens coefficients
  lePolynome.coeffs.clear();
  
  // Remettre le flot  l'tat normal (si possible)
  in.clear();
  if ( in.bad() ) return in;
  
  // Lire la parenthse ouvrante (erreur sinon)
  char symbole;
  in >> symbole;
  if ( in.fail() ) return in;
  if ( symbole != '(' )
    {
      in.putback( symbole );
      in.setstate( ios::failbit ); 
      return in;
    };
  
  // Lire les lments un par un, jusqu' la parenthse fermante
  C coeff;
  for( in >> symbole; symbole!=')'; in >> symbole )
    {
      if ( in.fail() ) return in;
      in.putback( symbole );
      in >> coeff;
      if ( in.fail() ) return in;
      lePolynome.coeffs.push_back( coeff );
    };
  
  // Si tout s'est bien pass, on renvoie le flot comme il se doit.
  return in;
}

