//=======================================================================
// rreal.cc : rounded real -- nombres flottants avec arrondis dirigs
// Copyright (C) 2004-2007 Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Compilation :
// Compiler avec l'option -lgmpxx pour intgrer la bibliothque gmp. 
//
// Description :
// La classe RReal illustre comment implmenter, avec des outils trs
// lmentaires, des nombres  virgule flottante en prcision arbitraire.
// Nous en profitant pour contrler explicitement les modes d'arrondi.
//
// *** Avertissement: aucun effort d'optimisation n'est fait ici. ***
// L'implmentation propose a pour seul but d'illustrer le principe.
// Dans toute application raliste on privilgiera une implmentation 
// plus complte, comme par exemple celle de la bibliothque GMP.
// ******************************************************************
//
// Structure des donnes :
// La reprsentation des nombres  virgule flottante est base sur 
// les nombres entiers : les flottants sont stocks sous la forme 
//  m2^e  avec deux entiers m (la mantisse) et e (l'exposant). 
// Afin de pouvoir choisir la longueur de la mantisse (ventuellement 
// trs grande) il nous faut une implmentation de *grands* entiers.
// Nous nous servirons de la classe mpz_class de la bibliothque GMP.
//
// Arithmtique arrondie :
// On effectue les oprations de base +, -, *, / de manire exacte, puis
// on raccourcit la mantisse lors qu'elle dpasse la longueur autorise.
// Pour cela on effectue une division euclidienne par 2^s o s est le 
// nombre des chiffres  supprimer, et "arrondir" veut dire "oublier 
// le reste".  (En compensation, bien sr, on ajoute s  l'exposant.)
// Cette manipulation garantit que la mantisse ne dpasse pas la taille
// autorise.  Le seul prix  payer est une (petite) erreur d'arrondi.
//
// Mode d'arrondi :
// La variable globale  stdMode  permet de choisir un mode d'arrondi 
// global pour les calculs.  En gnral durant un calcul on introduit
// plusieurs erreurs d'arrondi, le rsultat final ne sera donc pas exact,
// et pire encore on ne saura ni la taille ni la direction de l'erreur.
// En choisissant judicieusement le mode d'arrondi on peut au moins 
// garantir une ingalit.  Avec un double calcul on peut ainsi majorer
// puis minorer le rsultat exact, ce qui permet de contrler les erreurs 
// d'arrondi sous forme d'un encadrement fiable.  (Cette ide est  
// la base de l'arithmtique des intervalles.)
//
// Pourquoi travailler en base 2 ?
// On travaille sur un ordinateur binaire, donc il vaut mieux stocker les
// entiers en base 2.  Multiplication et division par 2 ne sont que des
// dcalages, il est donc utile de travailler en base 2 aussi pour les
// nombres flottants.  Le choix de la base 10 serait possible mais moins
// efficace.  Nanmoins, on convertira en base 10 pour l'entre-sortie.
//=======================================================================

#include <iostream>
#include <sstream>
#include <cmath>
using namespace std;

#include <gmpxx.h>

//-----------------------------------------------------------------------
// On peut considrer RReal comme une petite bibliothque.  On la place
// dans l'espace de noms "Numeric" pour ne pas polluer l'espace "std".
//-----------------------------------------------------------------------
namespace Numeric {

//=======================================================================
// Le type Mode spcifie le mode d'arrondi : quand un rsultat dpasse
// la longueur prescrite on raccourcit la mantisse, cd on arrondit.  
//  ce moment-l il faut dcider dans quel sens on veut arrondir. 
//=======================================================================
enum Mode
{
  RNDD,  // round down (towards -infinity)
  RNDU,  // round up (towards +infinity)
  RNDZ,  // round towards zero (round down if >0, round up if <0)
  RNDI,  // round towards infinity (round up if >0, round down if <0)
  RNDN,  // round to nearest (if ambiguous prefer even)
  RNDO   // round to nearest (if ambiguous prefer odd)
};

//-----------------------------------------------------------------------
// Changer la direction de l'arrondi lors du passage  l'oppos x -> -x
//-----------------------------------------------------------------------
Mode operator- ( Mode mode )
{
  switch( mode )
    {
    case RNDD: return RNDU;
    case RNDU: return RNDD;
    case RNDZ: return RNDZ;
    case RNDI: return RNDI;
    default: return mode;
    };
}  

Mode operator* ( int sign, Mode mode )
{ if ( sign < 0 ) return -mode; else return mode; }

//-----------------------------------------------------------------------
// Changer la direction de l'arrondi lors du passage  l'inverse x -> 1/x
//-----------------------------------------------------------------------
Mode operator! ( Mode mode )
{
  switch( mode )
    {
    case RNDD: return RNDU;
    case RNDU: return RNDD;
    case RNDZ: return RNDI;
    case RNDI: return RNDZ;
    default: return mode;
    };
}  

Mode operator^ ( Mode mode, int exposant )
{ if ( exposant < 0 ) return !mode; else return mode; }

//-----------------------------------------------------------------------
// Afficher le mode d'arrondi en texte
//-----------------------------------------------------------------------
ostream& operator<< ( ostream& out, Mode mode )
{
  switch( mode )
    {
    case RNDD: return out << "RNDD";
    case RNDU: return out << "RNDU";
    case RNDZ: return out << "RNDZ";
    case RNDI: return out << "RNDI";
    case RNDN: return out << "RNDN";
    case RNDO: return out << "RNDO";
    };
  return out;
}

//=======================================================================
// Dfinir un type pour la mantisse et l'exposant (grands entiers). 
//=======================================================================
typedef mpz_class Man;   // type pour la mantisse (grand entier)
typedef mpz_class Exp;   // type pour l'exposant  (grand entier)
typedef long int Len;    // type pour la longueur (petit entier)

// Conversion de Exp en Len
Len castExpToLen( const Exp& exp )
{
  if( ! exp.fits_slong_p() )
    { cerr << "Exponent out of range!" << endl; exit(1); }
  return exp.get_si();
} 

//-----------------------------------------------------------------------
// Dterminer la longueur d'un entier en binaire (= nombre de bits).
// Puisque les entiers sont dj stocks sous dveloppement binaires,
// cette opration est presque gratuite (au moins trs peu coteuse).
//
// Spcification :
// On pose  length(0)==0  puis pour m non nul on a length(m)==len 
// si et seulement si 2^(len-1) <= abs(m) < 2^len.  Quelques exemples :
// length=1 :  1dec=1bin
// length=2 :  2dec=10bin, 3dec=11bin
// length=3 :  4dec=100bin, 5dec=101bin, 6dec=110bin, 7dec=111bin
// length=4 :  8dec=1000bin, 9dec=1001bin, 10dec=1010bin, ...
// length=5 : 16dec=10000bin, 17dec=10001bin, 18dec=10010bin, ...
// length=6 : 64dec=100000bin, 65dec=100001bin,...
//-----------------------------------------------------------------------
Len length( const Man& m ) 
{ if ( m==0 ) return 0; else return mpz_sizeinbase( m.get_mpz_t(), 2l ); }

//-----------------------------------------------------------------------
// Dterminer le signe et la parit.  (On fait appel aux fonctions GMP.)
//-----------------------------------------------------------------------
int sign( const Man& m )
{ return mpz_sgn( m.get_mpz_t() ); }

bool odd( const Man& m )
{ return mpz_odd_p( m.get_mpz_t() ); }

bool even( const Man& m )
{ return mpz_even_p( m.get_mpz_t() ); }

bool even( const Man& m, Len exp )
{ return mpz_divisible_2exp_p( m.get_mpz_t(), exp ); }

//-----------------------------------------------------------------------
// Multiplication d'un entier a par 2^e.  Ceci revient  rajouter des
// zros au dveloppement binaire de a.  C'est simple et efficace, mais
// consomme tout de mme de la mmoire.   utiliser avec modration !
//-----------------------------------------------------------------------
Man operator<< ( const Man& a, const Exp& e )
{ return ( a << castExpToLen(e) ); }

//-----------------------------------------------------------------------
// On a besoin d'une division euclidienne respectant le mode d'arrondi.
// On se sert ici de diffrentes variantes de la division euclidienne
// fournies par la bibliothque GMP, en les adaptant le cas chant.
//-----------------------------------------------------------------------
Man eudiv( const Man& a, const Man& b, Mode mode )
{
  Man q;
  switch( mode )
    {
    case RNDD: // round down = floor
      mpz_fdiv_q( q.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() );
      break;
    case RNDU: // round up = ceil
      mpz_cdiv_q( q.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() ); 
      break;
    case RNDZ: // round towards zero = truncate
      mpz_tdiv_q( q.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() );
      break;
    case RNDI: // round towards infinity
      {
	Man r;
	mpz_tdiv_qr( q.get_mpz_t(), r.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() );
	q+= sign( r );
      }
      break;
    case RNDN: // round to nearest (if ambiguous prefer even)
      {
	Man r;
	mpz_fdiv_qr( q.get_mpz_t(), r.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() ); 
	r*= 2;
	if ( b > 0 ) { if( r>b || ( r==b && odd(q) ) ) ++q; }
	else         { if( r>-b || ( r==-b && odd(q) ) ) --q; };
      }
      break;
    case RNDO: // round to nearest (if ambiguous prefer odd)
      {
	Man r;
	mpz_fdiv_qr( q.get_mpz_t(), r.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t() ); 
	r*= 2;
	if ( b > 0 ) { if( r>b || ( r==b && even(q) ) ) ++q; }
	else         { if( r>-b || ( r==-b && even(q) ) ) --q; };
      }
      break;
    }
  return q;
}

//-----------------------------------------------------------------------
// La division euclidienne par 2^e est particulire car il ne s'agit 
// que d'un dcalage des bits vers la droite, le reste tant les bits 
// carts.  (Comme avant il faut respecter le mode d'arrondi.)  
// On se sert ici de diffrentes variantes de la division euclidienne 
// fournies par la bibliothque GMP, en les adaptant le cas chant.
//-----------------------------------------------------------------------
Man eudiv2( const Man& a, Len exp, Mode mode )
{
  Man q;
  switch( mode )
    {
    case RNDD: // round down = floor
      mpz_fdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
      break;
    case RNDU: // round up = ceil
      mpz_cdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
      break;
    case RNDZ: // round towards zero = truncate
      mpz_tdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
      break;
    case RNDI: // round towards infinity
      {
	Man r;
	mpz_tdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
	mpz_tdiv_r_2exp( r.get_mpz_t(), a.get_mpz_t(), exp );
	q+= sign( r );
      }
      break;
    case RNDN: // round to nearest (if ambiguous prefer even)
      {
	Man r;
	mpz_fdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
	mpz_fdiv_r_2exp( r.get_mpz_t(), a.get_mpz_t(), exp );
	if ( length(r)==exp && ( !even(r,exp-1) || odd(q) ) ) ++q;
      }
      break;
    case RNDO: // round to nearest (if ambiguous prefer odd)
      {
	Man r;
	mpz_fdiv_q_2exp( q.get_mpz_t(), a.get_mpz_t(), exp );
	mpz_fdiv_r_2exp( r.get_mpz_t(), a.get_mpz_t(), exp );
	if ( length(r)==exp && ( !even(r,exp-1) || even(q) ) ) ++q;
      }
      break;
    }
  return q;
}

//=======================================================================
// Dfinition de la classe RReal modlisant les nombres flottants 
// en prcision arbitraire et avec des arrondis dirigs.
//=======================================================================
class RReal
{
public:         
  //---------------------------------------------------------------------
  // Un objet de la classe RReal est un nombre rel de la forme  m2^e 
  // avec deux grands entiers  m  (la mantisse) et  e  (l'exposant).
  //
  // Ces lments sont-ils publics ou plutt privs  la classe RReal ?
  // C'est peu orthodoxe, mais je ne vois pas de mal de les dclarer 
  // publics ici.  Il n'y a pas d'invariant  garantir, donc on peut 
  // tout aussi bien autoriser l'accs direct par l'extrieur.
  //---------------------------------------------------------------------
  Man mantisse;
  Exp exposant;

  //---------------------------------------------------------------------
  // Dterminer la longueur de la mantisse = nombre de bits.
  // (Pour une spcification voir Numeric::length() dfinie plus haut.)
  //---------------------------------------------------------------------
  Len length() const
    { return Numeric::length(mantisse); };

  //---------------------------------------------------------------------
  // La magnitude  mag  d'un rel  r!=0  est l'unique entier tel que  
  // 2^mag <= abs(r) < 2^(mag+1).  Autrement dit, c'est la partie entire 
  // de  log2( abs(r) ) .  Heureusement on n'a pas besoin de calculer
  // le logarithme.  La magnitude peut tre dtermine de manire efficace 
  // en additionant l'exposant et la longueur de la mantisse (moins 1).
  //
  // Quelques exemples :
  // mag=-2 : 0.25 <= r < 0.5
  // mag=-1 : 0.5  <= r < 1
  // mag=0  : 1    <= r < 2
  // mag=1  : 2    <= r < 4
  // mag=2  : 4    <= r < 8
  //---------------------------------------------------------------------
  Exp magnitude() const
    { return exposant + length() - 1; };
  
  //---------------------------------------------------------------------
  // Prcision par dfaut (variable statique, cd globale pour RReal)
  //---------------------------------------------------------------------
  static Len stdLen;
  static Len len() { return stdLen; };
  static Len len( Len l ) { Len old= stdLen; stdLen= max(l,2l); return old; }
  
  //---------------------------------------------------------------------
  // Mode d'arrondi par dfaut (variable statique, cd globale pour RReal)
  //---------------------------------------------------------------------
  static Mode stdMode;
  static Mode mode() { return stdMode; }
  static Mode mode( Mode m ) { Mode old= stdMode; stdMode= m; return old; }

  //---------------------------------------------------------------------
  // Arrondir : on raccourcit la mantisse  la longueur prescrite.
  // Dans toutes les oprations qui suivent, on effectue le calcul 
  // d'abord de manire exacte, puis on arrondit, c'est--dire on 
  // raccourcit la mantisse lors qu'elle dpasse la longueur prescrite. 
  //  noter que cet arrondi doit respecter le mode d'arrondi choisi.
  //---------------------------------------------------------------------
  void arrondir( Mode mode= stdMode, Len len= stdLen )
    {
      if ( mantisse == 0 ) { exposant= 0; return; }; // cas particulier
      Len exces= length() - max( len, 2l );          // bits de trop
      if ( exces <= 0 ) return;                      // rien  raccourcir
      mantisse= eudiv2( mantisse, exces, mode );     // supprimer l'excs
      exposant+= exces;                              // compenser l'exposant
    }

  void arrondir( Len len, Mode mode= stdMode )
    { arrondir( mode, len ); }

  //---------------------------------------------------------------------
  // Constructeurs sans arrondi : initialiser la mantisse et l'exposant
  // par les valeurs donnes sans pour autant effectuer un arrondi.
  //---------------------------------------------------------------------
  RReal( const RReal& r )
    : mantisse(r.mantisse), exposant(r.exposant) {}
  
  RReal( Man man, Exp exp=0 )
    : mantisse(man), exposant(exp) {}

  RReal( long man=0, long exp=0 )
    : mantisse(man), exposant(exp) {}

  RReal( int man, long exp=0 )
    : mantisse(man), exposant(exp) {}

  //---------------------------------------------------------------------
  // Constructeurs suivi d'arrondi : la spcification explicite du mode 
  // d'arrondi ou de la prcision (ou des deux) dclenche un arrondi.
  //---------------------------------------------------------------------
  RReal( const RReal& r, Mode mode, Len len= stdLen )
    : mantisse(r.mantisse), exposant(r.exposant) { arrondir( mode, len ); }

  RReal( const RReal& r, Len len, Mode mode= stdMode )
    : mantisse(r.mantisse), exposant(r.exposant) { arrondir( mode, len ); }

  RReal( Man man, Exp exp, Mode mode, Len len= stdLen )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  RReal( Man man, Exp exp, Len len, Mode mode= stdMode )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  RReal( int man, int exp, Mode mode, Len len= stdLen )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  RReal( int man, int exp, Len len, Mode mode= stdMode )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  RReal( long man, long exp, Mode mode, Len len= stdLen )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  RReal( long man, long exp, Len len, Mode mode= stdMode )
    : mantisse(man), exposant(exp) { arrondir( mode, len ); }

  //---------------------------------------------------------------------
  // Constructeurs par conversion (toujours suivi d'arrondi)
  //---------------------------------------------------------------------
  RReal( const string& s, Mode mode= stdMode, Len len= stdLen )
    { istringstream in(s.c_str()); read( in, mode, len ); };

  RReal( const string& s, Len len, Mode mode= stdMode )
    { istringstream in(s.c_str()); read( in, mode, len ); };

  RReal( const char* s, Mode mode= stdMode, Len len= stdLen )
    { istringstream in(s); read( in, mode, len ); };

  RReal( const char* s, Len len, Mode mode= stdMode )
    { istringstream in(s); read( in, mode, len ); };

  RReal( double a, Mode mode= stdMode, Len len= stdLen )
    {
      // Attraper le cas exceptionnel a==0
      if ( a == 0 ) { mantisse= 0; exposant= 0; return; }
      
      // Dterminer le signe
      int signe= 1;
      if ( a < 0 ) { a= -a; signe= -1; };
      
      // Dterminer l'exposant 
      exposant= 0;
      Exp exp= Exp( log(a) / log(2.0) );
      for( ; exp > 53 ; --exp, ++exposant ) a/= 2.0;
      for( ; exp < 53 ; ++exp, --exposant ) a*= 2.0;
      
      // En dduire la mantisse 
      mantisse= signe * Man( a );
      arrondir( mode, len );
    }
  
  //---------------------------------------------------------------------
  // change et affectation sans arrondi
  //---------------------------------------------------------------------

  // change efficace et sr, sans allocation de mmoire
  void swap( RReal& r )
    { using std::swap; swap( mantisse, r.mantisse ); swap( exposant, r.exposant ); }
  
  // Affectation ralise par une copie et un swap
  RReal& operator= ( const RReal& r )
    { RReal copie(r); swap(copie); return *this; }
  
  RReal& set( const RReal& r )
    { RReal copie(r); swap(copie); return *this; }

  //---------------------------------------------------------------------
  // Affectation suivi d'un arrondi : la fonction set affecte et 
  // effectue un arrondi suivant le mode et la prcision spcifis.
  //---------------------------------------------------------------------
  RReal& set( const RReal& r, Mode mode, Len len= stdLen )
    { RReal copie(r); swap(copie); arrondir( mode, len ); return *this; }
  
  RReal& set( const RReal& r, Len len, Mode mode= stdMode )
    { RReal copie(r); swap(copie); arrondir( mode, len ); return *this; }
  
  RReal& set( Man man=0, Exp exp=0, Mode mode= stdMode, Len len= stdLen )
    { mantisse= man; exposant= exp; arrondir( mode, len ); return *this; }
  
  RReal& set( Man man, Exp exp, Len len, Mode mode= stdMode )
    { mantisse= man; exposant= exp; arrondir( mode, len ); return *this; }
  
  //---------------------------------------------------------------------
  // La fonction  rallonger(len)  rajoute len bits (zros)  la mantisse, 
  // et  remplir(len)  assure que la mantisse soit de longueur >= len. 
  // Ces manipulations sont compenses par une soustraction dans 
  // l'exposant ; ainsi le nombre rel reprsent ne change pas.
  //---------------------------------------------------------------------
  void rallonger( Len len )
    { if ( len > 0 ) { mantisse <<= len; exposant -= len; }; } 
  
  void rallonger( Exp len )
    { rallonger( castExpToLen(len) ); }
  
  void remplir( Len len= stdLen )
    { rallonger( len - length() ); } 

  //---------------------------------------------------------------------
  // Multiplication : multiplier les mantisses et additionner les exposants
  //---------------------------------------------------------------------
  RReal& mul( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
    { return set( a.mantisse * b.mantisse, a.exposant + b.exposant, mode, len ); }
  
  RReal& mul( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
    { return mul( a, b, mode, len ); }

  RReal& mul( const RReal& a, Mode mode= stdMode, Len len= stdLen )
    { return mul( *this, a, mode, len ); }

  RReal& mul( const RReal& a, Len len, Mode mode= stdMode )
    { return mul( *this, a, mode, len ); }

  RReal& operator*= ( const RReal& a )
    { return mul( *this, a ); }

  //---------------------------------------------------------------------
  // Division : diviser les mantisses et soustraire les exposants
  // Lors de la division il faut arrondir suivant le mode choisi.
  //---------------------------------------------------------------------
  RReal& div( const RReal& a, Mode mode= stdMode, Len len= stdLen )
    { 
      remplir( len + a.length() );
      return set( eudiv( mantisse, a.mantisse, mode ), exposant - a.exposant, mode, len );
    }

  RReal& div( const RReal& a, Len len, Mode mode= stdMode )
    { return div( *this, a, mode, len ); }
  
  RReal& div( RReal a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
    { return *this= a.div( b, mode, len ); }
  
  RReal& div( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
    { return div( a, b, mode, len ); }
  
  RReal& operator/= ( const RReal& a )
    { return div( a ); }

  //---------------------------------------------------------------------
  // Addition : galiser les exposants puis additionner les mantisses.
  // Attention : galiser les exposants est coteux s'ils diffrent
  // de beaucoup, par consquent on ne le fait qu'en dernier recours.
  //---------------------------------------------------------------------
  RReal& add( RReal a, RReal b, Mode mode= stdMode, Len len= stdLen )
    {
      // Attraper d'abord les cas triviaux.
      if( a.mantisse == 0 ) return set( b, mode, len );
      if( b.mantisse == 0 ) return set( a, mode, len );

      // Optimisation : on traite d'abord le cas o les exposants diffrent 
      //  tel point que les mantisses occupent des plages disjointes.
      if( a.magnitude() < b.magnitude() ) a.swap( b );
      Len extra= 1 + len - a.length();
      if ( extra < 0 ) extra= 0;
      if( a.exposant - extra > b.magnitude() )
	{ 
	  a.rallonger( extra+1 ); 
	  a.mantisse+= sign( b.mantisse ); 
	  return set( a, mode, len ); 
	}

      // Cas gnrique : on galise les exposants afin d'additionner.
      Exp minexp= min( a.exposant, b.exposant );
      a.rallonger( a.exposant - minexp );
      b.rallonger( b.exposant - minexp );
      return set( a.mantisse + b.mantisse, minexp, mode, len );
    }
  
  RReal& add( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
    { return add( a, b, mode, len ); }

  RReal& add( const RReal& a, Mode mode= stdMode, Len len= stdLen )
    { return add( *this, a, mode, len ); }

  RReal& add( const RReal& a, Len len, Mode mode= stdMode )
    { return add( *this, a, mode, len ); }
  
  RReal& operator+= ( const RReal& a )
    { return add( *this, a ); }
  
  //---------------------------------------------------------------------
  // Changer le signe. (C'est une opration sans arrondi.)
  //---------------------------------------------------------------------
  RReal operator+ () const
    { return *this; }
  
  RReal operator- () const
    { return RReal( -mantisse, exposant ); }
  
  //---------------------------------------------------------------------
  // Soustraction : pour simplifier on implmente  a - b  par  a + (-b)
  //---------------------------------------------------------------------
  RReal& sub( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
    { return add( a, -b, mode, len ); }
  
  RReal& sub( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
    { return add( a, -b, mode, len ); }
  
  RReal& sub( const RReal& a, Mode mode= stdMode, Len len= stdLen )
    { return add( *this, -a, mode, len ); }
  
  RReal& sub( const RReal& a, Len len, Mode mode= stdMode )
    { return add( *this, -a, mode, len ); }
  
  RReal& operator-= ( const RReal& a )
    { return add( *this, -a ); }
  
  //---------------------------------------------------------------------
  // Arrondir vers un entier (suivant le mode d'arrondi choisi)
  //---------------------------------------------------------------------
  Man entier( Mode mode= stdMode ) const
    { 
      if( exposant >= 0 ) return mantisse << exposant;
      return eudiv2( mantisse, castExpToLen(-exposant), mode );
    }

  bool isInteger() const
    {
      if( exposant >= 0 ) return true;
      return even( mantisse, castExpToLen(-exposant) );
    }
  
  //---------------------------------------------------------------------
  // Affichage exact en base 2 : toute la mantisse bit par bit 
  //---------------------------------------------------------------------
  void write_bin( ostream& out ) const
    {
      // Attraper le cas particulier du nombre 0
      if ( mantisse == 0 ) { out << "0"; return; };
      
      // Afficher le signe 
      if ( mantisse > 0 ) out << "0."; else out << "-0.";
      
      // Afficher la mantisse bit par bit
      out.unsetf( ios::showpos );
      Exp e( exposant + length() );
      for( Len bit= length(); bit>0; --bit )
	out << mpz_tstbit( mantisse.get_mpz_t(), bit-1 );

      // Afficher l'exposant par rapport  la base 2
      out << "b" << e;
    }

  //---------------------------------------------------------------------
  // Notation scientifique en base 10 (encore peu optimise)
  // Attention : cet affichage doit tenir compte du mode d'arrondi, car 
  // on effectue un changement de base 2 (interne) en base 10 (externe).
  //---------------------------------------------------------------------
  void write_dec( ostream& out, Mode mode= stdMode, Len decimales= 0 ) const
    {
      // Attraper le cas particulier du nombre 0
      if ( mantisse == 0 ) { out << "0"; return; };

      // Prparer le nombre a par dcalage en base 10 de sorte que 
      // son dveloppement dcimal commence par 0,c... avec un chiffre c!=0
      Exp e(0);
      int base= 10;
      RReal a(*this), b;
      if ( decimales <= 0 ) decimales= out.precision();
      Len bits= 10 + (2+10*decimales)/3;
      while( b.mul( a, base, mode, bits ).magnitude() < 0  ) { a.set(b); --e; }
      while( a.magnitude() >= 0  ) { a.div( base, mode, bits ); ++e; }
      
      // Extraire la mantisse comme un grand entier : multiplier par 10^decimales
      Man puissance= 1;
      for( Len k= decimales; k>0; --k ) 
	{ a.mul( base, mode, bits ); puissance*= base; } 
      Man man= a.entier( mode );
      
      // Sparer le signe et corriger un ventuel dbordement 
      int sgn= sign( man );
      if( sgn < 0 ) man= -man;
      while( man >= puissance ) { man/= base; ++e; }
      while( man * base < puissance ) { man*= base; --e; }
      
      // Afficher la mantisse et l'exposant en base 10
      out.unsetf( ios::showpos );
      if ( sgn < 0 ) out << '-';
      out << "0." << man << "e" << e;
    }

  //---------------------------------------------------------------------
  // crire un nombre rel dans une chane de caractres
  //---------------------------------------------------------------------
  string dec( Mode mode= stdMode, Len decimales= 0 ) const
  {
    ostringstream s;
    if( decimales <= 0 ) decimales= (stdLen*3)/10;
    write_dec( s, mode, decimales );
    return s.str();
  }
  
  //-----------------------------------------------------------------------
  // Affichage d'un entier exact ou affichage scientifique arrondi 
  //-----------------------------------------------------------------------
  void write( ostream& out, Mode mode= stdMode ) const
    {
      // Traiter le cas particulier d'un petit entier
      Exp mag= magnitude();
      if( mag >= 0 && mag <= 32 && exposant.fits_slong_p() )
	{
	  Len exp= castExpToLen( exposant );
	  if( exp >= 0 )
	    { out << ( mantisse << exp ); return; }
	  if( even( mantisse, -exp ) )
	    { out << ( mantisse >> -exp ); return; }
	}

      // Retomber sur l'affiche en notation scientifique
      write_dec( out, mode );
    }

  //-----------------------------------------------------------------------
  // Lecture d'un nombre flottant en entre (encore peu optimise)
  // C'est forcment un peu complexe car on accepte plusieurs formats :
  //   123    123.456    0.456    123.    .456  
  //   123e0  123.456e0  0.456e0  123.e0  .456e0
  //-----------------------------------------------------------------------
  void read( istream& in, Mode mode= stdMode, Len len= stdLen )
    {
      // Lire la mantisse et l'exposant
      Man man;
      Exp exp;
      read_dec( in, man, exp );
      
      // Convertir le nombre en format interne (avec arrondi)
      set( man, 0, mode, len );
      RReal base(10);
      for( ; exp>0 ; --exp ) mul( base, mode, len );
      for( ; exp<0 ; ++exp ) div( base, mode, len );
    }

  //-----------------------------------------------------------------------
  // Cette fonction auxiliaire lit la mantisse  man  et l'exposant  exp  
  // par rapport  la base 10 (sans arrondi = opration exacte).
  //-----------------------------------------------------------------------
  static void read_dec( istream& in, Man& man, Exp& exp )
    {
      // Initialiser le signe, la mantisse et l'exposant
      int sign= 1;
      man= 0;
      exp= 0;
      
      // remettre le flot  l'tat normal (si possible)
      in.clear();
      if( in.bad() ) return;
      
      // Dterminer le signe
      char c;
      in >> c;
      if ( c == '+' || c == '-' )
	{
	  if ( c == '-' ) sign= -1;
	  in.get(c);
	}
      
      // Dterminer la partie entire
      while( !in.eof() && isdigit(c) )
	{
	  man*= 10;
	  man+= int(c-'0');
	  in.get(c);
	}
      
      // Ajouter la partie fractionnaire
      if ( c == '.' )
	{
	  in.get(c);      
	  while( !in.eof() && isdigit(c) )
	    {
	      man*= 10;
	      man+= int(c-'0');
	      --exp;
	      in.get(c);
	    }
	}
      
      // Multiplier la mantisse par le signe
      if( sign < 0 ) man= -man;
      
      // Dterminer l'exposant
      if ( c == 'e' ) 
	{
	  Exp e;
	  in >> e;
	  exp+= e;
	}
      else
	{
	  in.putback(c);
	}
    }
};

//-----------------------------------------------------------------------
// Dfinir et initialiser les variables statiques (= globales pour RReal)
//-----------------------------------------------------------------------
Len RReal::stdLen= 53; 
Mode RReal::stdMode= RNDN; 

//-----------------------------------------------------------------------
// Introduire des synonymes plus commodes
//-----------------------------------------------------------------------
const Len& stdLen= RReal::stdLen;
const Mode& stdMode= RReal::stdMode;

//-----------------------------------------------------------------------
// Demander la prcision souhaite  l'utilisateur
//-----------------------------------------------------------------------
Len demander_la_precision()
{
  cout << "Combien de dcimales dsirez-vous faire calculer ? ";
  Len decimales;
  cin >> decimales;
  if( decimales < 3 ) decimales= 3;
  if( decimales > 1000 ) decimales= 1000;
  RReal::len( (10*decimales)/3 ); 
  cout << "Pour " << decimales << " dcimales je prvois "
       << "des mantisses de " << RReal::len() << " bits." << endl;
  return decimales;
}

//-----------------------------------------------------------------------
// Oprations arithmtiques crant un objet temporaire
//-----------------------------------------------------------------------
RReal add( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
{ RReal c; c.add( a, b, mode, len ); return c; }

RReal add( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
{ RReal c; c.add( a, b, mode, len ); return c; }

RReal operator+ ( const RReal& a, const RReal& b )
{ return add( a, b ); }


RReal sub( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
{ RReal c; c.sub( a, b, mode, len ); return c; }

RReal sub( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
{ RReal c; c.sub( a, b, mode, len ); return c; }

RReal operator- ( const RReal& a, const RReal& b )
{ return sub( a, b ); }


RReal mul( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
{ RReal c; c.mul( a, b, mode, len ); return c; }

RReal mul( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
{ RReal c; c.mul( a, b, mode, len ); return c; }

RReal operator* ( const RReal& a, const RReal& b )
{ return mul( a, b ); }


RReal div( const RReal& a, const RReal& b, Mode mode= stdMode, Len len= stdLen )
{ RReal c(a); c.div( b, mode, len ); return c; }

RReal div( const RReal& a, const RReal& b, Len len, Mode mode= stdMode )
{ RReal c(a); c.div( b, mode, len ); return c; }

RReal operator/ ( const RReal& a, const RReal& b )
{ return div( a, b ); }

//-----------------------------------------------------------------------
// Dterminer le signe, la valeur absolue, et la magnitude
//-----------------------------------------------------------------------
int sign( const RReal& a )
{ return sign( a.mantisse ); }

RReal abs( const RReal& a )
{ if ( sign( a.mantisse ) >= 0 ) return a; else return -a; }

Exp magnitude( const RReal& a )
{ return a.magnitude(); }

//-----------------------------------------------------------------------
// Comparaison sous forme d'une fonction qui calcule sign(a-b).
// Attention : galiser les exposants est coteux s'ils diffrent
// de beaucoup, par consquent on ne le fait qu'en dernier recours.
//-----------------------------------------------------------------------
int compare( const RReal& a, const RReal& b )
{
  // Attraper d'abord les cas faciles  signes distincts.
  int a_sign= sign( a.mantisse ), b_sign= sign( b.mantisse );
  if( a_sign > b_sign ) return +1; 
  if( a_sign < b_sign ) return -1;
  if( a_sign==0 && b_sign==0 ) return 0;
  
  // On sait maintenant que a et b sont tous les deux positifs
  // ou tous les deux ngatifs.  On compare leurs magnitudes.
  Exp a_mag= a.magnitude(), b_mag= b.magnitude();
  if( a_mag > b_mag ) return a_sign;
  if( a_mag < b_mag ) return -a_sign;
  
  // Quand on arrive ici, finalement, a et b ont mme signe et mme
  // ordre de grandeur.  Dans ce cas il est raisonnable d'galiser 
  // les exposants afin de comparer les mantisses.
  Exp minexp= min( a.exposant, b.exposant );
  Man aman( a.mantisse << ( a.exposant - minexp ) );
  Man bman( b.mantisse << ( b.exposant - minexp ) );
  if ( aman > bman ) return +1;
  if ( aman < bman ) return -1;
  return 0;
}

//-----------------------------------------------------------------------
// Comparaison sous forme d'oprateur
//-----------------------------------------------------------------------
bool operator== ( const RReal&a, const RReal& b ) { return compare(a,b) == 0; }
bool operator!= ( const RReal&a, const RReal& b ) { return compare(a,b) != 0; }
bool operator<= ( const RReal&a, const RReal& b ) { return compare(a,b) <= 0; }
bool operator<  ( const RReal&a, const RReal& b ) { return compare(a,b) <  0; }
bool operator>= ( const RReal&a, const RReal& b ) { return compare(a,b) >= 0; }
bool operator>  ( const RReal&a, const RReal& b ) { return compare(a,b) >  0; }
  
//-----------------------------------------------------------------------
// Oprateurs d'entre-sortie
//-----------------------------------------------------------------------
istream& operator>> ( istream& in, RReal& a )
{ a.read( in ); return in; }

ostream& operator<< ( ostream& out, const RReal& a )
{ a.write( out ); return out; }

//-----------------------------------------------------------------------
// Calculer le carr 
//-----------------------------------------------------------------------
RReal sqr( RReal a, Mode mode= stdMode, Len len= stdLen )
{
  return mul( a, a, mode, len );
}

//-----------------------------------------------------------------------
// Calculer la racine carre 
//-----------------------------------------------------------------------
RReal sqrt( RReal a, Mode mode= stdMode, Len len= stdLen )
{
  // Attraper le seul cas d'erreur
  if( sign(a) < 0 ) { cerr << "The square root of a negative number is undefined!" << endl; exit(1); }
  if( sign(a) == 0 ) return a;
 
  // On utilise d'abord le mode d'arrondi vers le haut partout.
  // On choisit une valeur initiale plus grande que la racine cherche,
  // puis on applique l'itration de Newton tant que le rsultat s'amliore.
  // On garantit ainsi une approximation du haut, de plus en plus proche.
  RReal u0, u1= (1+a)/2;
  do 
    { 
      // L'itration de Newton-Hron  u -> ( u + a/u ) / 2
      u1.swap( u0 );
      u1.div( a, u0, RNDU, len );
      u1.add( u0, RNDU, len );
      u1.div( 2, RNDU, len );
    } 
  while ( u1 < u0 );
  
  // On adapte la valeur finale suivant le mode d'arrondi spcifi
  if( mode==RNDD || mode==RNDZ || mode==RNDN )
    u1.div( a, u1, mode, len );
  return u1;
}

//-----------------------------------------------------------------------
// Calculer la puissance n-ime (par la mthode dichotomique)
// Entre : un nombre rel a et un petit entier n
// Sortie : la puissance  a^n  arrondie comme spcifie
//-----------------------------------------------------------------------
RReal power( RReal a, long n, Mode mode= stdMode, Len len= stdLen )
{
  // Attraper quelques cas triviaux
  if( n < 0 ) { n= -n; a.div( 1, a, mode, len ); }
  if( n == 0 ) return RReal(1); 
  if( n == 1 ) return a;

  // Dans le cas a<0 les signes changent et il faut adapter l'arrondi.
  // Contrairement aux modes RNDD, RNDU, les modes RNDZ, RNDI sont 
  // invariants par changement de signe.  Il vaut donc mieux se ramener 
  //  ces cas-l.  Les modes RNDN et RNDO restent tels quels.
  if ( sign(a) < 0 )
    {
      if ( n%2 == 0 )  // Le rsultat final sera positif.
	{
	  if ( mode == RNDD ) mode= RNDZ;
	  if ( mode == RNDU ) mode= RNDI;
	}
      else             // Le rsultat final sera ngatif.
	{
	  if ( mode == RNDD ) mode= RNDI;
	  if ( mode == RNDU ) mode= RNDZ;
	}
    }

  // Calcul de la puissance par la mthode dichotomique
  RReal p(1);
  while( n > 0 )
    {
      while( n%2 == 0 )
	{
	  a.mul( a, mode, len );
	  n/= 2;
	}
      p.mul( a, mode, len );
      --n;
    }
  return p;
}

//-----------------------------------------------------------------------
// Calculer la racine n-ime
// Entre : un nombre rel positif a et un petit entier n >= 2
// Sortie : un nombre positif u proche de la racine tel que  
//          soit  x <= u^n  pour les modes RNDU et RNDI
//          soit  u^n <= x  pour les modes RNDD et RNDZ
//          soit un nombre u proche de la racine (RNDN,RNDO)
//
// Convention :
// Pour n impair on tend aux rels ngatifs par symtrie.
//
// Problmes connus :
// - Le choix de la valeur initiale n'est pas encore optimis.
// - Mme si la racine est exacte on ne trouve qu'une approximation.
//-----------------------------------------------------------------------
RReal root( const RReal& a, long n, Mode mode= stdMode, Len len= stdLen )
{
  // Attraper d'abord les divers cas particuliers et erreurs
  if ( n == 0 ) { cerr << "The 0th root is undefined!" << endl; exit(1); }
  if ( n == 1 ) return a;
  if ( n < 0 ) return root( div(1,a,mode,len), -n, mode, len );
  
  if ( sign(a) == 0 ) return a;
  if ( sign(a) < 0 )
    if ( n%2 != 0 ) return -root( -a, n, -mode, len ); 
    else { cerr << "The square root of a negative number is undefined!" << endl; exit(1); }
  
  // On sait maintenant que l'on est dans le cas  a > 0  et  n >= 2 .
  // On applique l'itration de Newton-Hron tant que le rsultat s'amliore.
  // L'arrondi assure une approximation du haut, de plus en plus proche.
  RReal aux, u0, u1= (n-1+a)/n; // valeur initiale plus grande que la racine 
  do 
    { 
      // L'itration de Newton-Hron  u -> ( (n-1)*u + a/u^(n-1) ) / n
      u1.swap( u0 );
      u1.mul( n-1, u0, RNDU, len );
      aux= power( u0, n-1, RNDD, len );
      aux.div( a, aux, RNDU, len );
      u1.add( u1, aux, RNDU, len );
      u1.div( u1, n, RNDU, len );
    } 
  while ( u1 < u0 );
  
  // On adapte la valeur finale suivant le mode d'arrondi spcifi
  if( mode==RNDD || mode==RNDZ || mode==RNDN || mode==RNDO )
    {
      aux= power( u1, n-1, -mode, len );
      u1.div( a, aux, mode, len );
    };
  return u1;
}

//-----------------------------------------------------------------------
// Encadrer la racine n-ime
// Entre : un nombre rel positif a et un petit entier n >= 2
// Sortie : deux nombres positifs inf et sup tels que  
//          inf^n <= x <= sup^n  avec inf et sup proches
//-----------------------------------------------------------------------
void root( const RReal& a, long n, RReal& inf, RReal& sup, Len len= stdLen )
{
  sup= root( a, n, RNDU, len );
  RReal aux= power( sup, n-1, RNDU, len );
  inf.div( a, aux, RNDD, len );
}

//-----------------------------------------------------------------------
} // fin de l'espace de noms "Numeric"
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Spcialiser la fonction  swap(RReal,RReal)  dans l'espace "std"
//-----------------------------------------------------------------------
namespace std
{
  template<>
  void swap<Numeric::RReal>( Numeric::RReal& a, Numeric::RReal& b ) 
  { a.swap(b); }
}
