//=======================================================================
// strassen.cc : multiplication de matrices d'aprs Strassen
// Copyright (C) 2007 by Michael Eisermann, all rights reserved
// Institut Fourier, Universit Joseph Fourier, Grenoble, France
//
// Description : 
// On implmente ici la multiplication de matrices d'aprs Strassen
// pour la comparer empiriquement avec la multiplication usuelle.
//=======================================================================

#include "matrix.cc"
#include "timer.hh"
using namespace std;

//-----------------------------------------------------------------------
// Multiplication de matrices d'aprs Strassen (fonction principale)
//-----------------------------------------------------------------------
template<typename T>
Matrix<T> strassen( const Matrix<T>& a, const Matrix<T>& b )
{
  // Attraper la seule erreur possible
  if( a.coldim() != b.rowdim() )
    { cerr << "Erreur : multiplication de deux matrices incongrues !" << endl; exit(1); }

  // Les petites multiplications sont dlgues  la multiplication usuelle
  // if( a.rowdim() < 50 || b.coldim() < 50 ) return a*b;
  
  // Allouer la mmoire ncessaire pour le rsultat
  Matrix<T> c( a.rowdim(), b.coldim() );
  
  // La mthode de Strassen s'applique  la multiplication 
  // de matrices carres.  Sinon, on complte pas des zros.
  Dim adim= max( a.rowdim(), a.coldim() );
  Dim bdim= max( b.rowdim(), b.coldim() );
  Dim n= max( adim, bdim );
  strassen( n, a, b, c );
  return c;
}

//-----------------------------------------------------------------------
// Multiplication de matrices d'aprs Strassen (fonction rcursive)
// On suppose que a et b sont des matrices carres de mme taille n x n
// et que la matrice rsultat c a dj alloue la mmoire ncessaire.
//-----------------------------------------------------------------------
template<typename T>
void strassen( Dim n, const Matrix<T>& a, const Matrix<T>& b, Matrix<T>& c )
{
  // Les petites multiplications sont dlgues  la multiplication usuelle
  if( n < 50 ) { c= a*b; return; }

  // La dimension n doit tre paire, sinon on complte pas des zros.
  if( n%2 ) ++n;
  Dim m= n/2;

  // On dcompose a en quatre sous-matrices
  Matrix<T> a11(m,m), a12(m,m), a21(m,m), a22(m,m);
  a.get_submatrix(   1,   1, a11 );
  a.get_submatrix(   1, m+1, a12 );
  a.get_submatrix( m+1,   1, a21 );
  a.get_submatrix( m+1, m+1, a22 );

  // On dcompose b en quatre sous-matrices
  Matrix<T> b11(m,m), b12(m,m), b21(m,m), b22(m,m); 
  b.get_submatrix(   1,   1, b11 );
  b.get_submatrix(   1, m+1, b12 );
  b.get_submatrix( m+1,   1, b21 );
  b.get_submatrix( m+1, m+1, b22 );

  // On calcule sept produits auxiliaires
  Matrix<T> p1(m,m), p2(m,m), p3(m,m), p4(m,m), p5(m,m), p6(m,m), p7(m,m);
  strassen( m, a11+a22, b11+b22, p1 );
  strassen( m, a21+a22, b11    , p2 );
  strassen( m, a11    , b12-b22, p3 );
  strassen( m, a22    , b21-b11, p4 );
  strassen( m, a11+a12, b22    , p5 );
  strassen( m, a21-a11, b11+b12, p6 );
  strassen( m, a12-a22, b21+b22, p7 );

  // On applique les formules de Strassen
  c.set_submatrix(   1,   1, p1+p4-p5+p7 );
  c.set_submatrix(   1, m+1, p3+p5 );
  c.set_submatrix( m+1,   1, p2+p4 );
  c.set_submatrix( m+1, m+1, p1-p2+p3+p6 );
}

//-----------------------------------------------------------------------
// Un exemple d'utilisation et un premier test.
//-----------------------------------------------------------------------
int main()
{
  cout << "\nBonjour et bienvenue  la multiplication de matrices d'aprs Strassen !" << endl;
  Timer timer;

  for( Dim n=10; n<=1000; n+=10 )
    {
      cout << "Construction de deux matrices alatoires de taille "
	   << n << "x" << n << "... " << flush;
      Matrix<int> a(n,n), b(n,n);
      for( Dim i=1; i<=n; ++i )
	for( Dim j=1; j<=n; ++j )
	  {
	    a(i,j)= rand();
	    b(i,j)= rand();
	  }
      cout << "ok" << endl;

      cout << "Multiplication cubique usuelle  : " << flush;
      Matrix<int> c1;
      timer.start();
      c1= a*b;
      timer.stop();
      cout << timer << endl;

      cout << "Multiplication d'aprs Strassen : " << flush;
      Matrix<int> c2;
      timer.start();
      c2= strassen( a, b );
      timer.stop();
      cout << timer << endl;

      if( c1 != c2 ) cout << "Erreur : les rsultats ne concident pas !" << endl;
      cout << endl;
    }

  cout << "Au revoir et  bientt." << endl;
}
 
